/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): Adriana Danes
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.common;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;
import javax.naming.Context;


/**
 * This class manages reading of the resource properties used by the JOnAS Server. 
 * This class adopts the singleton design-pattern. No need to instanciate it, 
 * this is done by the private constructor called in the static reference.
 * Contributors:
 * - ../03/2003 Adriana Danes
 *   Manage unique instances for the resources properties file
 *   Change 'config' en 'conf'
 *   Change initial configuration policy : read properties file in a sole location : JONAS_BASE/CONFIG_DIR/
 *   Change method name getFilesEnv() to getConfigFileEnv()
 */
public class JProp {

    public static final String JONASPREFIX = "jonas";

    // configuration properties names
    static private final String JONAS_NAME = "jonas.name";
    static private final String JONAS_DEF_NAME = "jonas";
    static private final String JONAS_BASE = "jonas.base";
    
    // configuration directory name (changed from 'config' to 'conf' !!)
    static private final String CONFIG_DIR = "conf";

    // some environment properties
    static private Properties systEnv   = System.getProperties();  
    static private String jonasBase     = systEnv.getProperty(JONAS_BASE);
    static private String installRoot   = systEnv.getProperty("install.root");

    static private String fileSeparator = systEnv.getProperty("file.separator");
    static private String homeUser      = systEnv.getProperty("user.home");

    private Properties configFileEnv  = new Properties();
    private Properties allEnv         = null;
    private String propFilesName      = null;

    // The JProp singleton for jonas.properties file
    private static JProp unique = null;

    // The multiple JProp objects for the resources configuration files (resource.properties)
    // the keys are the configuration file names and the values, the JProp objects
    private static Hashtable multiple = new Hashtable();

    /**
     * private constructor to initialize and read properties files.
     */
    private JProp(String fileName) throws Exception {
	readPropFiles(fileName);
    }

    /**
     * private constructor to initialize and read JONAS properties files.
     */
    private JProp() throws Exception {
	readPropFiles(JONASPREFIX);
    }
    /**
     * Get the unique instance.
     * Create it at first call.
     */
    public static JProp getInstance() throws Exception {
	if (unique == null) {
	    unique = new JProp();
	}
	return unique;
    }

    /**
     * Get one of the multiple instances.
     * Create it at first call with a given configuration file name
     * @param fileName the name of the configuration file
     */
    public static JProp getInstance(String fileName) throws Exception {
	if (!multiple.contains(fileName)) {
	    multiple.put(fileName, new JProp(fileName));
	}
	return (JProp)multiple.get(fileName);
    }

    /**
     * Read initial configuration properties in file getProperty(JONAS_BASE)/conf/fileName
     * These properties may be overridden by system properties, provided on the java command line.
     */
    private void readPropFiles(String fileName) throws Exception {

	// Check the JONAS_BASE environment property
	if (jonasBase == null) {
	    throw new Exception("JOnAS configuration error: environment property jonas.base not set!");
	}
	jonasBase = jonasBase.trim();

	// JONAS_BASE/conf/fileName
	propFilesName = jonasBase + fileSeparator+ CONFIG_DIR + fileSeparator + fileName + ".properties";

	try {
	    FileInputStream is = new FileInputStream(propFilesName);
	    configFileEnv.load(is);
	} catch (FileNotFoundException e) {
	    throw new FileNotFoundException("Cannot find properties for "+fileName);
	} catch (IOException e) {
	    System.err.println(e);
	}

	allEnv = (Properties) configFileEnv.clone(); 
	// Overriddes with syst properties
	if (fileName.equals("jonas")) {
	    for (Enumeration e = systEnv.keys(); e.hasMoreElements();) {
		Object key   = e.nextElement();
		String value = ((String)systEnv.get(key)).trim();
		allEnv.put(key, (Object)value);
	    }
	    if (!systEnv.containsKey(JONAS_NAME)) 
		allEnv.put(JONAS_NAME, JONAS_DEF_NAME);
	}
    }

    /**
     * Static method which return the JOnAS install root value.
     */
    public static String getInstallRoot() {
	// Check install.root system property
	if (installRoot == null) {
	    throw new RuntimeException("JOnAS configuration error: System property install.root not set!");
	}
	return(installRoot);
    }

    /**
     * Static method which return the jonas.base proprty
     */
    public static String getJonasBase() {
	return jonasBase;
    }

    /**
     * Returns properties files name
     *
     * @return JOnAS properties files name
     */
    public String getPropFile() {
	return propFilesName;
    }

    /**
     * Returns JOnAS environment as configured with configuration file properties content and
     * system properties.
     * @return JOnAS properties
     */
    public Properties getEnv() {
	return allEnv;
    }


    /**
     * Returns JOnAS environment as configured with files properties only.
     *
     * @return JOnAS properties
     */
    public Properties getConfigFileEnv() {
	return configFileEnv;
    }

    /**
     * Returns the value of the related property. With default values.
     *
     * @return property value
     */
    public String getValue(String key, String defaultVal) {
	
	String retProperty = allEnv.getProperty(key, defaultVal);
	return retProperty.trim();
    }

    /**
     * Returns the value of the related property.
     * The method returns null if the property is not found.
     *
     * @return property value, null if not exist
     */
    public String getValue(String key) {
	
	String retProperty = allEnv.getProperty(key);
	if (retProperty != null) {
	    retProperty = retProperty.trim();
	}
	return retProperty;
    }

    /**
     * Returns the value of the related property as boolean.
     * @return property value, true or false.
     */
    public boolean getValueAsBoolean(String key, boolean def) {
	boolean ret = def;
	String value = this.getValue(key);
	if (value != null && value.equalsIgnoreCase("true")) {
	    ret = true;
	}
	return ret;
    }

    /**
     * Returns the value of the related property as String [].
     * The method returns null if the property is not found.
     *
     * @return property value, null if not exist
     */
    public String[] getValueAsArray(String key) {
	
	String [] res = null;
	String value = this.getValue(key);
	if (value != null) {
	    StringTokenizer st = new StringTokenizer(value, ",");
	    res = new String [st.countTokens()];
	    int i = 0;
	    while (st.hasMoreTokens()) {
		res[i++] = st.nextToken().trim();
	    }
	}
	return res;
    }

    /**
     * String representation of the object for trace purpose
     * @return String representation of this object
     */
    public String toString() {
	String s = new String();
	for (Enumeration e = this.configFileEnv.keys(); e.hasMoreElements();) {
	    Object key   = e.nextElement();
	    Object value = this.configFileEnv.get(key);
	    s = s.concat("   " + key + " = " + value + "\n");
	}
	if (s.length()>0) {
	    // take of the last '\n'
	    s = s.substring(0, s.length()-1);
	}
	return(s);
    }

    /**
     * Bind all the properties found in file properties in a naming context
     * the naming context must be allocated by the caller
     */
    public void env2Ctx(Context ctx) throws Exception {
	Enumeration e = configFileEnv.propertyNames();
	String key = null;
	while ( e.hasMoreElements() ) {
	    key = (String) e.nextElement();
	    ctx.bind( key, configFileEnv.getProperty( key, "") );
	}
    }

    /**
     * Displays the JOnAS properties values, as they are set by the 
     *  different property files.
     */
    static public void main(String args[]) {
	
	JProp jonasProperties = null;
	try {
	    jonasProperties = JProp.getInstance();
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(2);
	}
	for (Enumeration e = jonasProperties.configFileEnv.keys(); e.hasMoreElements();) {
	    Object key   = e.nextElement();
	    Object value = jonasProperties.configFileEnv.get(key);
	    System.out.println(key.toString() + "=" + value.toString());
	}	
    }


}
