/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.io.PrintWriter;
import java.io.FileNotFoundException;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Set;
import java.util.Iterator;
import java.rmi.RemoteException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.naming.NamingException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.XADataSource;
import javax.transaction.Transaction;
import org.enhydra.jdbc.pool.StandardPoolDataSource;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * DatabaseService acts as a factory for the DataSource objects.
 * Its goal is to create such objects and to register them in JNDI
 *
 * @author Philippe Durieux
 *
 * Contributor(s):
 *
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          New implementation of unbindDataSources for closing correctly 
 *          all connections at server shutdown.
 * 03/01/14 Adriana Danes
 *          Highlight Configuration properties
 *          Change createDataSource() signature : take additional argument, the datasource name
 *          Manage binded datasources (mapping of jndi name to datasource name)
 *          Use datasource name for naming MBeans (instead of jndi name)
 *          Modify MBean methods to take into account the previous points.
 */
public class DataBaseServiceImpl extends AbsServiceImpl implements DataBaseService, DataBaseServiceImplMBean {

    static private Logger logger = null;

    // List of DataSource in use.
    // Better to use only 1 list ?
    private Vector cmList = new Vector();	// ConnectionManager
    private Vector ldsList = new Vector(); 	// StandardPoolDataSource

    // List of JonasXAPoolDataSource (used by JNDI lookup)
    private static Hashtable poolList = new Hashtable();
    
    // Binded Datasource (jndi name -> datasource name)
    private Hashtable bindedDatasources = new Hashtable();

    // Names of the DataSource to create when starting the DataBase service
    private Vector dataSourceNames = new Vector();

    // Transaction Service reference
    private TransactionService transactionService = null;

    // Initial Context for Naming
    private Context ictx = null;
    
    // The DataBase service configuration parameters
    static final String DATASOURCES = "jonas.service.dbm.datasources";
    static final String CLASS = "jonas.service.dbm.class";

    // The Datasource configuration parameters
    static final String NAME = "datasource.name"; // JNDI name of the datasource
    static final String CLASSNAME = "datasource.classname";
    static final String DEF_CLASSNAME = "no class name";
    static final String URL = "datasource.url";
    static final String DEF_URL = "no url";
    static final String DESCRIPTION = "datasource.description";
    static final String DEF_DESCRIPTION = "no desc";
    static final String USERNAME = "datasource.username";
    static final String DEF_USERNAME = "";
    static final String PASSWORD = "datasource.password";
    static final String DEF_PASSWORD = "";
    static final String FACTORY = "datasource.factory";
    static final String DEF_FACTORY = "none";
    static final String XADATANAME = "datasource.xadataname";
    static final String DEF_XADATANAME = "none";
    static final String ISOLATIONLEVEL = "datasource.isolationlevel";
    static final String DEF_ISOLATIONLEVEL = "";
    static final String MAPPERNAME = "datasource.mapper";
    static final String DEF_MAPPERNAME = "rdb";
    static final String MINCON = "datasource.mincon";
    static final String DEF_MINCON = "5";
    static final String MAXCON = "datasource.maxcon";
    static final String DEF_MAXCON = "10";

    // The ConnectionManager configuration parameters
    static final String CONNCHECKLEVEL = "jdbc.connchecklevel";
    static final String DEF_CONNCHECKLEVEL = "1";
    static final String CONNMAXAGE = "jdbc.connmaxage";
    static final String DEF_CONNMAXAGE = "30";
    static final String CONNTESTSTMT = "jdbc.connteststmt";
    static final String DEF_CONNTESTSTMT = "SELECT 1";
    static final String MINCONPOOL = "jdbc.minconpool";
    static final String DEF_MINCONPOOL = "10";
    static final String MAXCONPOOL = "jdbc.maxconpool";
    static final String DEF_MAXCONPOOL = "20";
    static final String SLEEPTIMEPOOL = "jdbc.sleeptimepool";
    static final String DEF_SLEEPTIMEPOOL = "300000";
    static final String GCPOOL = "jdbc.gcpool";
    static final String DEF_GCPOOL = "false";
    static final String GENERATIONPOOL = "jdbc.generationpool";
    static final String DEF_GENERATIONPOOL = "1";
    static final String DEADLOCKPOOL = "jdbc.deadlockpool";
    static final String DEF_DEADLOCKPOOL = "300000";
    static final String LOOPWAITPOOL = "jdbc.loopwaitpool";
    static final String DEF_LOOPWAITPOOL = "10000";
    
    
    // -------------------------------------------------------------------
    // Service Implementation
    // -------------------------------------------------------------------

    /**
     * Initialization of the DataBase service.
     * Configuration information is passed thru a Context object.
     */
    public void doInit(Context ctx) throws ServiceException {
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
	// Avoid using NamingManager here: performance is not a goal here.
	try {
	    ictx = new InitialContext();
	} catch (NamingException e) {
	    logger.log(BasicLevel.ERROR, "Cannot create initial context when DataBase service initializing");
	    throw new ServiceException("Cannot create initial context when DataBase service initializing", e);
	}

	// Get the list of the datasource names
	String ds = null;
	try {
	    ds = (String)ctx.lookup(DATASOURCES);
	} catch (NamingException e) {
	    // No problem if there is no value for 'datasources'
	}
	if (ds != null) {
	    StringTokenizer st = new StringTokenizer(ds, ",");
	    while (st.hasMoreTokens()) {
		dataSourceNames.add(st.nextToken().trim());
	    }
	}

	// Get un reference to the Transaction service
	try {
	    transactionService =
		(TransactionService) ServiceManager.getInstance().getTransactionService();
	} catch (ServiceException se) {
	    logger.log(BasicLevel.ERROR, "Cannot get the Transaction service: "+se);
	    throw se;
	} catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "Cannot get the Transaction service: "+e);
	    throw new ServiceException("Cannot get the Transaction service: ", e);
	}

	logger.log(BasicLevel.DEBUG, "DataBaseService initialized");
	
    }

    /**
     * Starting DataBase service
     * Initialization of the service is already done.
     */
    public void doStart() throws ServiceException {

	// creates connection managers for each datasource
	String dsName = null;
	for (int i = 0; i < dataSourceNames.size(); i++) {
	    dsName = (String) dataSourceNames.elementAt(i);
	    try {
		JProp prop = JProp.getInstance(dsName);
		logger.log(BasicLevel.DEBUG, "Creating Datasource "+dsName);
		createDataSource(dsName, prop.getConfigFileEnv());
	    } catch (Exception e) {
		logger.log(BasicLevel.ERROR, "JOnAS: Cannot create datasource: "+dsName+" "+e);
		logger.log(BasicLevel.ERROR, "Please check if "+dsName+".properties is available");
	    }
	}

	try {
	    // Register DataBaseService MBean : JmxDataBaseService 
	    ((JmxService)ServiceManager.getJmxService()).getJmxServer().registerMBean(this, JonasObjectName.databaseService());
	} catch (ServiceException se) {
	    // Jmx Service not available, do nothing
	} catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "DataBaseService: Cannot start the DataBase service:\n"+e);
	    throw new ServiceException("DataBaseService: Cannot start the DataBase service", e);
	}

	logger.log(BasicLevel.DEBUG, "DataBaseService started");
    }

    /**
     * Stopping DataBase service
     * Unbind DataSources
     */
    public void doStop() throws ServiceException {
	try {
	    unbindDataSources();
	} catch (NamingException e) {
	    logger.log(BasicLevel.ERROR, "Cannot unbind datasources " + e);
	    throw new ServiceException("Cannot unbind datasources ", e);
	}

	try {
	    MBeanServer mbeanServer = ((JmxService)ServiceManager.getJmxService()).getJmxServer();
	    // unregister Database Service MBean : DatabaseServiceImplMBean 
	    mbeanServer.unregisterMBean(JonasObjectName.databaseService());
	    // unregister all Datasource
	    Set objectNames = mbeanServer.queryNames(JonasObjectName.allDatasources(), null) ;
	    Iterator it  = objectNames.iterator();
	    while (it.hasNext()) {
		mbeanServer.unregisterMBean((ObjectName)it.next());
	    }
	} catch (ServiceException se) {
	    // Jmx Service not available, do nothing
	} catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "EJBService: Cannot stop the DataBase service:\n"+e);
	    throw new ServiceException("DataBaseService: Cannot stop the DataBase service",e);
	}

	logger.log(BasicLevel.DEBUG, "DataBaseService stopped");
    }
    
    // -------------------------------------------------------------------
    // DataBaseService Implementation
    // -------------------------------------------------------------------

    /**
     * Creates a ConnectionManager (implementation of sql.dataSource).
     * @param datasourceName - datasource name
     * @param dsd - a set of properties that describes a dataSource and the ConnectionPool
     *  <p>If datasource.factory is set, we create a JonasXAPoolDataSource.
     *  If not set, we create an old fashioned ConnectionManager</p>
     */
    public void createDataSource(String datasourceName, Properties dsd) throws Exception {
	logger.log(BasicLevel.DEBUG, "entering");
	// Get properties common to all DataSource types
	String dsName = dsd.getProperty(NAME).trim(); // JDNI name of the datasource
	String className = dsd.getProperty(CLASSNAME, DEF_CLASSNAME).trim();
	String url = dsd.getProperty(URL, DEF_URL).trim();
	String description = dsd.getProperty(DESCRIPTION, DEF_DESCRIPTION).trim();
	String user = dsd.getProperty(USERNAME, DEF_USERNAME).trim();
	String password = dsd.getProperty(PASSWORD, DEF_PASSWORD).trim();
	String xaFactory = dsd.getProperty(FACTORY, DEF_FACTORY).trim();
	String xaName = dsd.getProperty(XADATANAME, DEF_XADATANAME).trim();
	String connCheckLevel = dsd.getProperty(CONNCHECKLEVEL, DEF_CONNCHECKLEVEL).trim();
	String connMaxAge = dsd.getProperty(CONNMAXAGE, DEF_CONNMAXAGE).trim();
	String defaultStatement = dsd.getProperty(CONNTESTSTMT, DEF_CONNTESTSTMT).trim();

	logger.log(BasicLevel.DEBUG, "Create data source " + datasourceName + " having JNDI name " + dsName);

	if (xaFactory.equals("none")) {
	    // Create ConnectionManager (JOnAS DataSource)
	    logger.log(BasicLevel.DEBUG, "create JOnAS ConnectionManager");
	    ConnectionManager ds = new ConnectionManager();

	    // Initialize ConnectionManager
	    ds.setDatasourceName(datasourceName); // set datasource name
	    //logger.log(BasicLevel.DEBUG, "datasourceName set");
	    ds._setDSName(dsName); // set jndi name
	    //logger.log(BasicLevel.DEBUG, "dsName set");
	    ds.setUrl(url);
	    //logger.log(BasicLevel.DEBUG, "url set");
	    ds.setClassName(className);
	    //logger.log(BasicLevel.DEBUG, "className set");
	    ds.setUserName(user);
	    //logger.log(BasicLevel.DEBUG, "user set");
	    ds.setPassword(password);
	    //logger.log(BasicLevel.DEBUG, "pwd set");
	    ds.setTransactionIsolation(dsd.getProperty(ISOLATIONLEVEL, DEF_ISOLATIONLEVEL).trim());
	    //logger.log(BasicLevel.DEBUG, "isolev set");
	    ds.setMapperName(dsd.getProperty(MAPPERNAME, DEF_MAPPERNAME).trim());
	    ds.setDataSourceXADataName(xaName);
	    //logger.log(BasicLevel.DEBUG, "xaName set");
	    ds.setDataSourceDescription(description);
	    //logger.log(BasicLevel.DEBUG, "desc set");
	    ds.poolConfigure(connCheckLevel, connMaxAge, defaultStatement);
	    //logger.log(BasicLevel.DEBUG, "pool Configured");

	    cmList.addElement(ds);
	    ictx.rebind(dsName, ds);
	    // allows for getting the data source name from the jndi name
	    bindedDatasources.put(dsName, datasourceName);
	    logger.log(BasicLevel.WARN, "Mapping ConnectionManager "+url+" on "+dsName);
	    try {
		// MBean Datasource registering
		((JmxService)ServiceManager.getJmxService()).getJmxServer()
		    .registerMBean(ds,JonasObjectName.datasource(datasourceName));
	    } catch (ServiceException se) {
		// Jmx Service not available, do nothing
	    }
	    
	} else {
	    // JonasXAPoolDataSource (LUTRIS DataSource)
	    logger.log(BasicLevel.DEBUG, "create Lutris DataSource");
	    JonasXAPoolDataSource ds = new JonasXAPoolDataSource();
	    boolean debug = logger.isLoggable(BasicLevel.DEBUG);
	    ds.setDebug(debug);
	    ds.setLogWriter(Log.getLogWriter("org.objectweb.jonas.dbm"));
	    
	    ds.setCheckLevelObject((new Integer(connCheckLevel)).intValue());
	    ds.setLifeTime((new Integer(connMaxAge)).intValue()*60*1000);
	    ds.setJdbcTestStmt(defaultStatement);
	    try {
		ds.setMinSize((new Integer(dsd.getProperty(MINCONPOOL, DEF_MINCONPOOL).trim())).intValue());
		ds.setMaxSize((new Integer(dsd.getProperty(MAXCONPOOL, DEF_MAXCONPOOL).trim())).intValue());
	    } catch (Exception e) {
		logger.log(BasicLevel.ERROR, " Cannot set min and max size of the pool");
	    }
	    
	    ds.setDataSourceName(xaName);// set XA name
	    ds.setDSName(dsName);// set jndi name
	    ds.setDatasourceName(datasourceName);// set datasource name
	    ds.setUser(user);
	    ds.setPassword(password);
	    ds.setClassName(className);
	    ds.setUrl(url);
	    ds.setXAFactory(xaFactory);
	    ds.setDataSourceDescription(description);
	    ds.setTransactionManager(transactionService.getTransactionManager());
	    ds.setSleepTime((new Integer(dsd.getProperty(SLEEPTIMEPOOL, DEF_SLEEPTIMEPOOL).trim())).intValue());
	    String gc = dsd.getProperty(GCPOOL, DEF_GCPOOL);
	    ds.setGC(gc.equals("true"));

	    if (!GENERATIONPOOL.equals(DEF_GENERATIONPOOL))
		ds.setGeneration((new Integer(dsd.getProperty(GENERATIONPOOL, DEF_GENERATIONPOOL).trim())).intValue());

	    ds.setDeadLockMaxWait((new Integer(dsd.getProperty(DEADLOCKPOOL, DEF_DEADLOCKPOOL).trim())).intValue());
	// 10 seconds
	    ds.setDeadLockRetryWait((new Integer(dsd.getProperty(LOOPWAITPOOL, DEF_LOOPWAITPOOL).trim())).intValue());

	    ldsList.addElement(ds);

	    // rebind in JNDI.
	    // Set it in the poolList : lookup must retrieve this object later.
	    poolList.put(xaName, ds);
	    ictx.rebind(dsName, ds);
	    // allows for getting the data source name from the jndi name
	    bindedDatasources.put(dsName, datasourceName);

	    // bind all information from the XA factory
	    try {
		CommonDataSource connect = (CommonDataSource)(Class.forName(xaFactory).newInstance());
		connect.setProperties(url, className, user, password);
		connect.setTransactionManager(transactionService.getTransactionManager());
		connect.setMinCon((new Integer(dsd.getProperty(MINCON, DEF_MINCON).trim())).intValue());
		connect.setMaxCon((new Integer(dsd.getProperty(MAXCON, DEF_MAXCON).trim())).intValue());
		connect.setDebug(debug);
		connect.setLogWriter(Log.getLogWriter("org.objectweb.jonas.dbm"));

		ds.xads = (XADataSource)connect;
		ictx.rebind(xaName, connect);
	    } catch (ClassNotFoundException e) {
		logger.log(BasicLevel.ERROR, " Class Not Found Exception");
		throw e;
	    } catch (InstantiationException e) {
		logger.log(BasicLevel.ERROR, " Cannot create instance of XAFactory");
		throw e;
	    } catch (IllegalAccessException e) {
		logger.log(BasicLevel.ERROR, " Illegal Access Exception");
		throw e;
	    } catch(SQLException e) {
		logger.log(BasicLevel.ERROR, " SQLException:"+e);
		throw e;
	    }
	    try {
		// MBean Datasource registering
		((JmxService)ServiceManager.getJmxService()).getJmxServer().
		    registerMBean(ds, JonasObjectName.datasource(datasourceName));
	    } catch (ServiceException se) {
		// Jmx Service not available, do nothing
	    } 

	    logger.log(BasicLevel.WARN, "Mapping JonasXAPoolDataSource "+url+" on "+dsName);
            logger.log(BasicLevel.WARN, "XADataSource = "+xaName);
	}
    }

    /**
     * Propagation of the new configuration of the trace functionnalities
     * in the driver
    */
    public void traceJdbc() {

	if (ldsList.size() > 0) {
	    PrintWriter pr = Log.getLogWriter("org.objectweb.jonas.dbm");
	    for (Enumeration lk = ldsList.elements(); lk.hasMoreElements(); ) {
		JonasXAPoolDataSource cm = (JonasXAPoolDataSource)lk.nextElement();
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		cm.setDebug(debug);
		cm.setLogWriter(pr);
		CommonDataSource connect = (CommonDataSource)cm.xads;
		connect.setDebug(debug);
		connect.setLogWriter(pr);
	    }
	}
    }
    
    /*
     * Unbind dataSource names from the Naming
     */
    public void unbindDataSources() throws NamingException {
       
	logger.log(BasicLevel.DEBUG, "entering");

	try {
	    if (cmList.size() > 0) {
		String dsn = null;
		for (Enumeration lk = cmList.elements(); lk.hasMoreElements(); ) {
		    ConnectionManager cm = (ConnectionManager) lk.nextElement();
		    cm.closeAllConnection();
		    dsn = cm._getDSName();
		    ictx.unbind(dsn);
		    bindedDatasources.remove(dsn);
		}
	    }
	    if (ldsList.size() > 0) {
		for (Enumeration lk = ldsList.elements(); lk.hasMoreElements(); ) {
		    StandardPoolDataSource cm = (StandardPoolDataSource) lk.nextElement();
		    cm.stopPool();
		    String dsn = cm.getDataSourceName();
		    ictx.unbind(dsn);
		    bindedDatasources.remove(dsn);
		}
	    }
	} catch (NamingException e) {
	    logger.log(BasicLevel.ERROR, "cannot unbind DataSources : " + e);
	    throw e;
	}
	
	// unbind all mbean 
	try {
	    MBeanServer mbeanServer = ((JmxService)ServiceManager.getJmxService()).getJmxServer();
	    // unregister all Datasource
	    Set objectNames = mbeanServer.queryNames(JonasObjectName.allDatasources(), null) ;
	    Iterator it  = objectNames.iterator();
	    while (it.hasNext()) {
		mbeanServer.unregisterMBean((ObjectName)it.next());
	    }
	} catch (ServiceException se) {
	    // Jmx Service not available, do nothing
	} catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "DataBase Service : Cannot unbind the DataSource:\n"+e);
	}	
    }

    /**
     * get the XAPoolDataSource matching the DataSource name
     */
    public static JonasXAPoolDataSource getXAPoolDataSource(String dsname) {
        JonasXAPoolDataSource cm = (JonasXAPoolDataSource) poolList.get(dsname);
	return cm;
    }

    /**
     * get ConnectionManager for the datasource having this JNDI name.
     */
    public ConnectionManager getConnectionManager(String dsname) {
	ConnectionManager cm = null;
	if (cmList.size() > 0) {
	    for (Enumeration lk = cmList.elements(); lk.hasMoreElements(); ) {
		cm = (ConnectionManager)lk.nextElement();
		if (cm._getDSName().equals(dsname)) {
		    return cm;
		}
	    }
	}
	return null;
    }

    //------------------------------------------------------------------------------------------

    /**
     * MBean method: 
     * @return Integer Total Number of Datasource available in JOnAS
     */ 
    public Integer getCurrentNumberOfDataSource() {
	return new Integer(cmList.size() + ldsList.size());
    }
    
    /**
     * MBean method:
     * @return Integer Total Number of JDBC connection open
     */ 
    public Integer getTotalCurrentNumberOfJDBCConnectionOpen(){
	int result = 0;
	if (cmList.size() > 0) {
	    for (Enumeration lk = cmList.elements(); lk.hasMoreElements(); ) {
		ConnectionManager cm = (ConnectionManager)lk.nextElement();
		result += cm.getCurrentNumberOfJDBCConnectionOpen().intValue();
	    }
	}
	if (ldsList.size() > 0) {
	    for (Enumeration lk = ldsList.elements(); lk.hasMoreElements(); ) {
		JonasXAPoolDataSource cm = (JonasXAPoolDataSource)lk.nextElement();
		result += cm.getCurrentNumberOfJDBCConnectionOpen().intValue();
	    }
	}
	return new Integer(result);
    }
	
    /**
     * MBean method:
     * @return true if the datasource having thid jndi name is loaded
     */
    public boolean isLoadedDataSource(String dsName) {
	logger.log(BasicLevel.DEBUG, "entering");
	boolean result = false;
	if (cmList.size() > 0) {
	    for (Enumeration lk = cmList.elements(); lk.hasMoreElements(); ) {
		ConnectionManager cm = (ConnectionManager)lk.nextElement();
		if (cm._getDSName().equals(dsName)) 
		    return true;
	    }
	}
	if (ldsList.size() > 0) {
	    for (Enumeration lk = ldsList.elements(); lk.hasMoreElements(); ) {
		JonasXAPoolDataSource cm = (JonasXAPoolDataSource)lk.nextElement();
		if (cm.getDSName().equals(dsName)) 
		    return true;
	    }
	}
	return result;
    }

    /**
     * MBean method:
     * @param name of the data source to unload
     */
    public void unloadDataSource(String name) {
	try {
	    if (cmList.size() > 0) {
		for (Enumeration lk = cmList.elements(); lk.hasMoreElements(); ) {
		    ConnectionManager cm = (ConnectionManager)lk.nextElement();
		    String dsName = cm.getDatasourceName();
		    String jndiName = cm.getDSName();
		    if (dsName.equals(name)) {
			// unbind data source 
			cm.closeAllConnection();
			ictx.unbind(jndiName);
			// remove datasource 
			cmList.remove(cm);
			try {
			    // unregistering MBean
			    ((JmxService)ServiceManager.getJmxService()).getJmxServer().
				unregisterMBean(JonasObjectName.datasource(dsName));
			} catch  (ServiceException se) {
			    // Jmx Service not available, do nothing
			}
			return;
		    }
		}
	    }
	    if (ldsList.size() > 0) {
		for (Enumeration lk = ldsList.elements(); lk.hasMoreElements(); ) {
		    JonasXAPoolDataSource cm = (JonasXAPoolDataSource)lk.nextElement();
		    String dsName = cm.getDatasourceName();
		    String jndiName = cm.getDSName();
		    if (dsName.equals(name)) {
			// unbind data source
			cm.stopPool();
			ictx.unbind(jndiName);
			// remove datasource
			ldsList.remove(cm);
			try {
			    // unregistering MBean
			    ((JmxService)ServiceManager.getJmxService()).getJmxServer().
				unregisterMBean(JonasObjectName.datasource(dsName));
			}  catch  (ServiceException se) {
			    // Jmx Service not available, do nothing
			}
			return;
		    }
		}
	    }
	}  catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "cannot unload DataSources : " + e);
	}
    }

    /**      
     * MBean method: 
     * @return datasource properties from a local file
     */
    public Properties getDataSourcePropertiesFile(String dsFile) throws Exception {
	try {
	    return JProp.getInstance(dsFile).getConfigFileEnv();

	} catch (Exception e) {
	    if (e instanceof FileNotFoundException) {
		logger.log(BasicLevel.ERROR, "Please check if "+dsFile+".properties is available in JONAS_BASE/conf/ directory");
	    } else {
		logger.log(BasicLevel.ERROR, "Error occured when reading file " + dsFile);
	    }
	    throw e;
	}
    }

    /**
     * MBean method:
     * load a new datasource
     * @param name datasource name
     * @param prop datasource properties
     */
    public void loadDataSource(String name, Properties prop) throws RemoteException {
	logger.log(BasicLevel.DEBUG, "JOnAS: Load data source named " + name);
	String dsName = prop.getProperty("datasource.name").trim();
	try {
	    if (isLoadedDataSource(dsName)) {
		unloadDataSource(dsName);
	    }
	    logger.log(BasicLevel.DEBUG, "JOnAS: call method to create a data source");
	    createDataSource(name, prop);
	    logger.log(BasicLevel.DEBUG, "JOnAS: new data source created");
	} catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "JOnAS: Cannot create datasource: "+name);
	    throw new RemoteException("JOnAS: Cannot create datasource: "+name+",\n"+e);
	}	
    }
    
    /**
     * MBean method allowing to determine the datasource name from its jndi name
     * @param jndiName The jndi name of a datasource
     * @return The datasource name
     */
    public String getDatasourceName(String jndiName) {
	return (String)bindedDatasources.get(jndiName);
    }  
    
}
	
