/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
  

package org.objectweb.jonas.resource;

import javax.resource.ResourceException;
import javax.resource.spi.LocalTransaction;

import javax.transaction.xa.XAResource;
import javax.transaction.xa.XAException;
import javax.transaction.xa.Xid;

import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 *  A LocalXAWrapper that intercepts the XA calls for an RAR that only
 *   supports LocalTransactions and translates them to the appropriate
 *   Local Transaction methods.
 *
 * @author     Eric.Hardesty@bull.com
 */
public final class LocalXAWrapper
    implements XAResource
{

    /**
     *  The Logger instance where messages are written.
     */
    static private Logger logger = null;

    /**
     *  The boolean to determine if in a current transaction
     */
    boolean isInTransaction;

    /**
     *  The LocalTransaction object to make the begin(), commit(), 
     *    and rollback() calls.
     */
    LocalTransaction localTrans = null;

    LocalXAWrapper(LocalTransaction _localTrans, Logger _logger)
    {
        isInTransaction = false;
        localTrans = _localTrans;
        logger = _logger;
    }

    /**
     *  Commit the localTransaction, the params aren't used for a 
     *    local transaction.
     *
     *@exception  XAException  Exception trying to commit local transaction
     */
    public void commit(Xid xid, boolean flag)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        try
        {
            localTrans.commit();
        }
        catch(ResourceException res)
        {
            throw new XAException(res.toString());
        }
    }

    /**
     *  No method to map for a local transaction.
     */
    public void end(Xid xid, int i)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
    }

    /**
     *  No method to map for a local transaction.
     */
    public void forget(Xid xid)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
    }

    /**
     *  No method to map for a local transaction, just 
     *   return no timeout.
     */
    public int getTransactionTimeout()
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        return -1;
    }

    /**
     *  No method to map for a local transaction, just return false4.
     */
    public boolean isSameRM(XAResource xaresource)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        return false;
    }

    /**
     *  No method to map for a local transaction, just return XA_OK.
     */
    public int prepare(Xid xid)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        return XAResource.XA_OK;
    }

    /**
     *  No method to map for a local transaction.
     */
    public Xid[] recover(int i)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        return null;
    }

    /**
     *  Rollback the localTransaction, the param isn't used for a 
     *    local transaction.
     *
     *@exception  XAException  Exception trying to rollback local transaction
     */
    public void rollback(Xid xid)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        try
        {
            localTrans.rollback();
        }
        catch(ResourceException res)
        {
            throw new XAException("Rollback failed, " + res.getMessage());
        }
    }

    /**
     *  No method to map for a local transaction, just 
     *   return no timeout.
     */
    public boolean setTransactionTimeout(int i)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        return false;
    }

    /**
     *  Only start a local transaction if a new transaction is being
     *   attempted, just return if joining or resuming.  
     *
     *@exception  XAException  Transaction already started or error
     *                          starting a new local transaction
     */
    public void start(Xid xid, int i)
        throws XAException
    {
        logger.log(BasicLevel.DEBUG, "");
        try
        {
            if(i != XAResource.TMJOIN && i != XAResource.TMRESUME)
            {
                if(isInTransaction)
                {
                    throw new XAException("LocalXAWrapper.start:  Local transaction already started");
                }
                localTrans.begin();
                isInTransaction = true;
            }
        }
        catch(ResourceException res)
        {
            throw new XAException("LocalTransaction.begin failed, " + res.toString());
        }
    }

}
