/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.server;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.StringTokenizer;
import org.objectweb.common.Env;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.tools.CheckEnv;


/**
 * This class load all the jars needed to start JOnAS and after this, it 
 * launch the JOnAS server (Server class).
 * @author Ludovic Bert   (initial developper)
 * @author Florent Benoit (initial developper)
 * @author Philippe Durieux (new jonas jar files architecture)
 */
public class Bootstrap {

    private static boolean jdk14;
    private static String orb;
    private static String jonasLib;
    private static String commonLib;
    private static String javaLib;
    private static String jonasBase;
    private static JProp props = null;
    private static String[] services;
    private static boolean service_web = false;
    private static boolean service_web_catalina = false;
    private static boolean service_web_jetty = false;

    /**
     * Return an array containing all the resources declared in the classpath
     * environment variable.
     * @return a list containing all the resources declared in the classpath
     * environment variable.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getClassPath(boolean xtracp) throws Exception {
        String cpprop = xtracp ? "jonas.classpath" : "java.class.path";
        String cpValue = System.getProperty(cpprop);
        //System.out.println("classpath="+cpValue); // DEBUG
        String sep = File.pathSeparator;
        JURLs jurls = new JURLs();
        StringTokenizer tokenizer = new StringTokenizer(cpValue, sep);
        while (tokenizer.hasMoreTokens()) {
            File file = new File(tokenizer.nextToken());
            if (file.isDirectory()) {
                jurls.addDir(file);
            } else {
                jurls.add(file);
            }
        }
        return jurls;
    }


    /**
     * Return an array containing all the resources used for the integration
     * of Catalina.
     * @param catalinaHomeDir the home directory of Catalina.
     * @return a list containing all the resources used for the integration
     * of Catalina.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getCatalinaResources(File catalinaHomeDir) throws Exception {
                    
        JURLs jurls = new JURLs();
        File jarf = new File(jonasLib, "web_catalina.jar");

        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            jurls.add(new File(catalinaHomeDir, "bin" + File.separator + "bootstrap.jar"));
            jurls.add(new File(catalinaHomeDir, "server" + File.separator + "lib"));
            jurls.add(new File(catalinaHomeDir, "common" + File.separator + "lib"));
            jurls.add(new File(commonLib, "ow_jonas_tomcat4.jar"));
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Jetty.
     * @param jettyHomeDir the home directory of Jetty.
     * @return a list containing all the resources used for the integration
     * of Jetty.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJettyResources(File jettyHomeDir) throws Exception {

        JURLs jurls = new JURLs();

        File jarf = new File(jonasLib, "web_jetty.jar");

        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            String jettyLib = jettyHomeDir + File.separator + "lib";
            if (jdk14){
                jurls.add(new File(jettyLib, "org.mortbay.jetty.jar"));
            } else {
                jurls.add(new File(jettyLib, "org.mortbay.jetty-jdk1.2.jar"));
            }
            String jettyExt = jettyHomeDir + File.separator + "ext";
            jurls.add(new File(jettyExt, "ant.jar"));
            jurls.add(new File(jettyExt, "jasper-compiler.jar"));
            jurls.add(new File(jettyExt, "jasper-runtime.jar"));
            jurls.add(new File(jettyLib, "org.mortbay.jmx.jar"));
            jurls.add(new File(jettyLib, "javax.servlet.jar"));
            jurls.add(new File(commonLib, "ow_jonas_jetty.jar"));
        }
        return jurls;
    }    

    /**
     * Return an array containing all the resources used for JOnAS
     * @return a list containing all the resources.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJonasResources() throws Exception {

        JURLs jurls = new JURLs();
	
        File jarf = new File(jonasLib, orb+"_jonas.jar");

        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            System.out.println("INFO : Please take the time to do the 'ant install' task in the JONAS_ROOT directory. This will improve the performances, remove some warnings and allow you to build the examples.");
            // add resources for RMI
            if (Env.isSunRMI()) {
                jurls.add(new File(jonasLib, "rmi"));
            }

            // add resources for JEREMIE
            if (Env.isJeremie()) {
                jurls.add(new File(jonasLib, "jeremie"));
                jurls.add(new File(jonasLib, "jonathan"));
            }

            // add resources for DAVID
            if (Env.isDavid()) {
                jurls.add(new File(jonasLib, "david"));
                jurls.add(new File(jonasLib, "jonathan"));
            }

            // add Orb Independant JOnAS resources
            jurls.add(new File(commonLib, "ow_jonas.jar"));

            // add external jars needed for JOnAS
            jurls.add(new File(commonLib, "j2ee"));
            jurls.add(new File(commonLib, "acme"));
            jurls.add(new File(commonLib, "connector"));
            jurls.add(new File(commonLib, "joram"));
            jurls.add(new File(commonLib, "jorm"));
            jurls.add(new File(commonLib, "log"));
            jurls.add(new File(commonLib, "medor"));
            jurls.add(new File(commonLib, "velocity"));
            jurls.add(new File(commonLib, "xml"));

            // add ext
            jurls.add(new File(jonasLib, "ext"), ".jar");

        }

        // Add resources for jmx, depending on the choosen implementation
        String myjmx = props.getValue("jonas.service.jmx.class");
        String jmxLib = commonLib + File.separator + "jmx";
        if (myjmx.indexOf("sunri") == -1) {
            jurls.add(new File(jmxLib, "mx4j-jmx.jar"));
            jurls.add(new File(jmxLib, "mx4j-tools.jar"));
        } else {
            jurls.add(new File(jmxLib, "jmxri.jar"));
            jurls.add(new File(jmxLib, "jmxtools.jar"));
        }

        // add some methods missing in JDK 1.3
        if (!jdk14) {
            jurls.add(new File(commonLib, "ow_jonas_jdk13.jar"));
        }

        return jurls;
    }

    /**
     * Get the list of the resources to load when launching JOnAS.
     * @return the list of the resources to load when launching JOnAS.
     * @throws Exception if a resource cannot be found.
     */
    public static URL[] getResources(boolean defcl) throws Exception {
        JURLs jurls = new JURLs();

        // Start with resources from the classpath
        jurls.merge(getClassPath(defcl));

        // Load now mail.jar because it must be loaded before activation.jar
        jurls.add(new File(commonLib + File.separator + "j2ee", "mail.jar"));

        // activation.jar needed at least for GenIC
        jurls.add(new File(commonLib + File.separator + "j2ee", "activation.jar"));

        // add JOnAS resources
        jurls.merge(getJonasResources());

        String catalinaHome = System.getProperty("catalina.home");
        String jettyHome = System.getProperty("jetty.home");

        // Set catalina.home to $JONAS_BASE if library is found
        // and if catalina.home isn't set.
        if (catalinaHome == null) {
            File jarCatalina = new File(jonasLib, "web_catalina.jar");
            if (jarCatalina.exists()) {
                System.setProperty("catalina.home", jonasBase);
                catalinaHome = jonasBase;
            }
        }

        // Set jetty.home to $JONAS_BASE if library is found
        // and if jetty.home isn't set.
        if (jettyHome == null) {
            File jarJetty = new File(jonasLib, "web_jetty.jar");
            if (jarJetty.exists()) {
                System.setProperty("jetty.home", jonasBase);
                jettyHome = jonasBase;
            }
        }

        // Add the specific resource of the web_container (tomcat 4.1.x or jetty) service
        if (service_web) {
            if (catalinaHome != null && service_web_catalina) {
                jurls.merge(getCatalinaResources(new File(catalinaHome)));
            } else if (jettyHome != null && service_web_jetty) {
                jurls.merge(getJettyResources(new File(jettyHome)));
            } 
            // else user has not set the webcontainer HOME var but no error message.
        }

        // add resources for catalina (tomcat 4.1.x) if required
        if (catalinaHome != null ) {
            jurls.merge(getCatalinaResources(new File(catalinaHome)));
        } 

        // add resources for jetty if required
        if (jettyHome != null ) {
            jurls.merge(getJettyResources(new File(jettyHome)));
        }

        return jurls.toURLs();
    }


    /**
     * Server main routine (Load all the jars needed to start jonas)
     * @param args the list of the args to give to the bootstrap class.
     */
    public static void main(String args[]) {
        String classToRun = args[0];
        try {
            // Check JDK version
            jdk14 = Env.isJAVA4();

            // Check orb
            orb = System.getProperty("objectweb.orb");

            // Get jonas properties
            props = JProp.getInstance();
            services = props.getValueAsArray("jonas.services");
            for (int i = 0; i < services.length; i++) {
                if (services[i].equals("web")) {
                    service_web = true;
                }
            }

            // Get $JONAS_BASE
            jonasBase = props.getJonasBase();


            String implWebContServiceClass = props.getValue("jonas.service.web.class");
            if (implWebContServiceClass != null) {
                if (implWebContServiceClass.equals("org.objectweb.jonas.web.jetty.JettyJWebContainerServiceImpl")) {
                    service_web_jetty = true;
                } else if (implWebContServiceClass.equals("org.objectweb.jonas.web.catalina41.CatalinaJWebContainerServiceImpl")
                           || implWebContServiceClass.equals("org.objectweb.jonas.web.catalina40.CatalinaJWebContainerServiceImpl")) {
                    service_web_catalina = true;
                }
            }

            // Get the $JONAS_ROOT/lib directories
            String jonasRoot = System.getProperty("install.root");
            if (jonasRoot == null) {
                System.err.println("JONAS_ROOT is not set");
                System.exit(2);
            }
            jonasLib = jonasRoot + File.separator + "lib";
            commonLib = jonasLib + File.separator + "common";

            // Create a ClassLoader and set it for the current thread.
            boolean dlr = props.getValueAsBoolean("jonas.default.classloader", false);
            JClassLoader jonasLoader = null;
            if (dlr) {
                jonasLoader = new JClassLoader(getResources(dlr));
            } else {
                jonasLoader = new JClassLoader(getResources(dlr), null);
            }
            Thread.currentThread().setContextClassLoader(jonasLoader);

            //jonasLoader.printURLs(); // DEBUG

            // Launch the "class_to_run" by using our classloader.
            Class clazz = jonasLoader.loadClass(classToRun);
            Class[] argList = new Class[] { args.getClass() };
            Method meth = clazz.getMethod("main", argList);
            String[] newArgs = new String [args.length - 1];
            System.arraycopy(args, 1, newArgs, 0, newArgs.length);
            Object appli = meth.invoke(null, new Object[]{newArgs});
        } catch (Exception e) {
            System.err.println(classToRun+" error: " + e);
            e.printStackTrace();
            System.exit(2);
        }
    }
}
