/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.server;

import java.io.FileInputStream;
import java.util.Properties;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import javax.naming.NamingException;
import javax.management.MBeanServer;

import org.objectweb.common.Env;
import org.objectweb.common.RemoteObject;
import org.objectweb.jonas.adm.Adm;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.registry.RegistryManager;
import org.objectweb.jonas.service.Service;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas_ejb.lib.Version;
import org.objectweb.util.monolog.api.HandlerFactory;
import org.objectweb.util.monolog.api.LevelFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.file.monolog.PropertiesConfAccess;

/**
 * This class implements an JOnAS Server.
 * @author Philippe Coq, Philippe Durieux
 * 00/06/15: Jeff Mesnil
 * 01/06/15: Regis Le Brettevillois - Libelis / JOnAS team - evidian
 *           Service as org.objectweb.jonas.service.Service
 * 02/04/25: Sebastien Chassande-Barrioz sebastien.chassande@inrialpes.fr
 *           Monolog for traces
 *
 *           Adriana Danes
 * 03/01/20:  -Highlight configuration properties
 * 03/03/..:  -Force jmx service startup
 -Create and register the JonAS configuration MBean. This MBean manages JOnAS persistent re-configuration (see JonasConfig class)
*/
public class Server {

    static private Logger logger = null;
    static private String SERVERNAME = "JOnAS"; // J2EE Application Server Name
    private String serverName = null; // Current server name (the server instance name)

    // The Server singleton value
    private static Server unique = null;

    // The associated ServiceManager for JOnAS
    ServiceManager sm = null;

    // Associated Adm object
    Adm adm = null;

    // Number of active threads in the initial thread's group
    int threadCount = 0;
    
    // Estimated number of active groups in the initial thread's group
    int groupCount = 0;

    // Server configuration properties names
    static final String NAME = "jonas.name";
    static final String DEF_NAME = "jonas";
    static final String ORB_PORT = "jonas.orb.port";
    static final String DEF_ORB_PORT = "0";
    static final String LOG_CONFIGFILE = "jonas.log.configfile";
    static final String DEF_LOG_CONFIGFILE = "trace.properties";
    static final String CONFIGFILE_USECLASSPATH = "jonas.log.configfile.useclasspath";
    static final String DEF_CONFIGFILE_USECLASSPATH = "true";

    // Server configuration properties
    int orbPort;
    String logConfigFile;
 
    /**
     * Private constructor to force only 1 instance.
     */
    private Server() throws Exception {
        sm = ServiceManager.getInstance();
    }

    /**
     * Get the unique instance.
     * Create it at first call.
     */
    public static Server getInstance() throws Exception {
        if (unique == null) {
            unique = new Server();
        }
        return unique;
    }

    /**
     * Start the server
     */
    public void start() throws Exception {

        // Print jonas version and orb
        String orbs = null;
        if (Env.isJeremie()) {
            orbs = "Jeremie";
        } else if (Env.isSunRMI()){
            orbs = "rmi";
        } else if (Env.isDavid()) {
            orbs = "David RMI";
        }

        // Read the JONAS properties
        JProp props = null;
        try {
            props = JProp.getInstance();
        } catch (Exception e) {
            throw new Exception("JOnAS: Cannot read the JOnAS properties ("+e+")");
        }

        // Server name
        serverName = props.getValue(NAME, DEF_NAME);

        // RMI Security Manager
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        // Set configuration info to the Log System (monolog)
        // Default is to use trace.properties found in the classpath.
        logConfigFile = props.getValue(LOG_CONFIGFILE, DEF_LOG_CONFIGFILE);
        String usecp = props.getValue(CONFIGFILE_USECLASSPATH, DEF_CONFIGFILE_USECLASSPATH);
        Log.configure(logConfigFile, usecp.equalsIgnoreCase("true"));
        // get a logger for server traces
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);

        // Init the ORB
        try {
            // Port value zero: an anonymous port is choosen
            orbPort = Integer.parseInt(props.getValue(ORB_PORT, DEF_ORB_PORT));
            RemoteObject.setPort(orbPort);
            logger.log(BasicLevel.DEBUG, "The " + ORB_PORT + " value is " + orbPort);

        } catch (NumberFormatException e) {
            logger.log(BasicLevel.ERROR, "JOnAS: Invalid " + ORB_PORT + " property value (default value used!)");
        }

        // First of all, start registry service 
        sm.startRegistry();

        // then, start JMX service (start an MBean server and create RMI connector for it registered in JNDI)

        sm.startJmx();

        // get the JMX server
        JmxService jmxService = null;
        MBeanServer jmxServer = null;
        try {
            jmxService = (JmxService) ServiceManager.getJmxService();
            jmxServer = jmxService.getJmxServer();
        } catch (ServiceException se) {
            // this could mean that jmx service not available
            // this is not possible, as we started it even if it does not apper in the services list configuration parameter
            logger.log(BasicLevel.ERROR, "Could not access the jmx service");	
            throw se;
        }

        // create an MBean allowing server configuration management
        try {
            JonasConfig configMBean = new JonasConfig(props.getConfigFileEnv(), jmxServer);
            jmxServer.registerMBean(configMBean, JonasObjectName.serverConfig());
            // TODO change JmxServer class name !!! 
            jmxServer.registerMBean(new JmxServer(this, serverName, Version.NUMBER, orbs), JonasObjectName.ejbServer());
            jmxServer.registerMBean(Log.getInstance(), JonasObjectName.logService());
        } catch (Exception e) {
            // Cannot administrate this server throu JMX
            jmxService.stop();	    
        }

        // Creates the Adm instance for remote administration
        try {
            adm = new Adm(props);
        } catch (Exception e) {
            e.printStackTrace();
            throw new Exception("Cannot create Adm : "+e);
        }

        // start all other services
        sm.startServices();

        // TM Server or EJB Server ?
        boolean isEJB = true;
        try {
            Service ejbservice = ServiceManager.getEjbService();
        } catch (ServiceException se) {
            isEJB = false;
        }
        adm.serverReady(isEJB);
        System.out.println("The JOnAS Server '"+serverName+"' version "+Version.NUMBER+" is ready");
        logger.log(BasicLevel.INFO, "Jonas server "+serverName+" started on "+orbs);

        // Periodically count the number of active threads in this thread's group
        Thread thisThread = Thread.currentThread();
        ThreadGroup thGroup = thisThread.getThreadGroup();
        while(true) {	    
            threadCount = Thread.activeCount();
            groupCount = thGroup.activeGroupCount();
            thisThread.sleep(10000);
        }
    }

    /**
     * get the number of active threads in the JOnAS server's initial thread group
     */
    protected int getActiveThreadsCount() {
        return threadCount;
    }
    /** 
     * get the estimated number of active groups in the JOnAS server's initial thread group
     */
    public int getActiveGroupCount() {
        return groupCount;
    }    
    /**
     * Stop the server without stopping the JVM
     */
    public void stop() throws Exception {
        if (adm != null) {
            System.out.println("The JOnAS Server '"+serverName+"' has been stopped");
            logger.log(BasicLevel.INFO, "Jonas server "+serverName+" stopped");
            adm.stopServer();
        }
    }

    /**
     * Stop the server and stop the JVM (MBean method)
     */
    public void kill() throws Exception {
        if (adm != null) {
            adm.killServer();
        }
    }


    /**
     * Server main routine.
     * No args.
     */
    public static void main(String args[]) {
        Server server = null;
        try {
            server = Server.getInstance();
            server.start();
        }  catch (JonasAlreadyStartedException e) {
            // if we get this exception this means that
            // we are trying to restart the same JOnAS
            System.err.println("JOnAS error: "+SERVERNAME+" already started");
            System.exit(2);
        } catch (Exception e) {
            // We don't use here the trace system.
            System.err.println("JOnAS error: "+e.getMessage());
            System.err.println(SERVERNAME+" halting");
            try {
                if (server!=null) server.stop();
            } catch (Exception ignore) {
            }
            System.exit(2);
        }
    }

}
