/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Ludovic BERT & Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.catalina41;

//import java
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

//import javax
import javax.naming.LinkRef;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.StringRefAddr;

//import objectweb
import org.objectweb.util.monolog.api.BasicLevel;

//import jonas
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImpl;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas.web.lib.JarTools;

//import apache
import org.apache.commons.digester.Digester;
import org.apache.commons.digester.Rule;
import org.apache.commons.modeler.Registry;

//import catalina
import org.apache.catalina.Container;
import org.apache.catalina.Context;
import org.apache.catalina.Engine;
import org.apache.catalina.Host;
import org.apache.catalina.Lifecycle;
import org.apache.catalina.Server;
import org.apache.catalina.Service;
import org.apache.catalina.Valve;
import org.apache.catalina.authenticator.AuthenticatorBase;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.core.StandardServer;
import org.apache.catalina.deploy.ContextResource;
import org.apache.catalina.deploy.NamingResources;
import org.apache.catalina.deploy.ResourceParams;
import org.apache.catalina.loader.WebappClassLoader;
import org.apache.catalina.startup.ContextConfig;
import org.apache.catalina.startup.EngineRuleSet;
import org.apache.catalina.startup.NamingRuleSet;
import org.apache.catalina.valves.ValveBase;

// import sax
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;


/**
 * This class provides an implementation of the Catalina service (as 
 * web container service).
 * 02/03 : Change this service for catalina version 4.1.x
 * 03/03 : Embedded is no more used. Use Catalina classes directly.
 *         Works with the catalina MBeans.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class CatalinaJWebContainerServiceImpl extends AbsJWebContainerServiceImpl {

    /**
     * Relative path of the configuration file
     */
    protected static final String CONFIG_FILE = "conf/server.xml";

    /**
     * The reference on the catalina server.
     */
    private Server server = null;

    /**
     * The shared extensions class loader for this server.
     */
    private ClassLoader parentClassLoader =
        Thread.currentThread().getContextClassLoader();

    /**
     * The name of the current jonas server
     */
    private static String nameOfServer = null;

    /**
     * Web container started ?
     */
    private boolean started = false;


    /**
     * Initialize the Catalina service.
     * @param ctx the configuration context of the service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(javax.naming.Context ctx) throws ServiceException {
        super.doInit(ctx);
        boolean useSecurity = false;
        try {
            ServiceManager.getSecurityService();
            useSecurity = true;
        } catch (Exception e) {
            // nothing to do 
        }
        
        // if the user do not specify catalina.base,
        // catalina.base = catalina.home.
        String catalinaHome = System.getProperty("catalina.home");
        if (catalinaHome != null) {
            if (System.getProperty("catalina.base") == null) {
                System.setProperty("catalina.base", catalinaHome);
            }
        }

        JProp jp = null;
        try {
            jp = JProp.getInstance();
        } catch (Exception e) {
            String err = "Error when trying to get jonas name ";
            logger.log(BasicLevel.ERROR, err + " " + e.getMessage());
            throw new ServiceException(err, e);  
        }
        CatalinaJWebContainerServiceImpl.nameOfServer = jp.getValue("jonas.name", "jonas");
        System.setProperty("catalina.useNaming", "false");

        //Set the MBeanServer instead of creating a new one
        //But only if we have the jmx service which is started.
        JmxService srvcJMX = null;
        try {
            srvcJMX = (JmxService) ServiceManager.getJmxService();
            if (srvcJMX != null) {
                logger.log(BasicLevel.DEBUG, "Set the MBean server to the JOnAS jmx server");
                Registry.setServer(srvcJMX.getJmxServer());
            }
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Can't set JMX server for Catalina " + e.getMessage());
        }

        //Add JOnAS/Tomcat mbean
        logger.log(BasicLevel.DEBUG, "Add mbeans descriptions to the JMX Registry");
        InputStream stream = this.getClass().getResourceAsStream("/org/objectweb/jonas/web/catalina41/jonas-mbeans-descriptors.xml");
        try {
            Registry.loadRegistry(stream);
            stream.close();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Can't add the mbeans description into Tomcat" + e.getMessage());
        }

    }

    /**
     * Start the Catalina service in a new thread
     * @throws ServiceException if the startup failed.
     */
    public void doStart() throws ServiceException {

        // Create the digester for the parsing of the server.xml.
        Digester digester = createStartDigester();
     
        // Execute the digester for the parsing of the server.xml.
        // And configure the catalina server.
        File configFile = null;

        try {
            configFile = getConfigFile();
        } catch (FileNotFoundException e) {
            logger.log(BasicLevel.ERROR, "Cannot find config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }

        try {
            InputSource is = new InputSource("file://" + configFile.getAbsolutePath());
            FileInputStream fis = new FileInputStream(configFile);
            is.setByteStream(fis);
            digester.push(this);
            digester.parse(is);
            fis.close();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Error when reading config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }

        // Set the global context of the server
        NamingResources namingResources = server.getGlobalNamingResources();
        javax.naming.Context globalCtx = null;
        javax.naming.Context ctxold = null;
        try {
            globalCtx = naming.createEnvironmentContext("catalina_global");
            ctxold = naming.setComponentContext(globalCtx);

            int i;
            // Resources
            ContextResource[] resources = namingResources.findResources();
            for (i = 0; i < resources.length; i++) {
                addResource(globalCtx, namingResources, resources[i]);
            }
        } catch (NamingException e) {
            String err = "Error while populating environment of the server";
            logger.log(BasicLevel.ERROR, err + "  :" + e.getMessage());
            throw new ServiceException(err, e);
        }
        // Set the global naming context of the server
        ((StandardServer) server).setGlobalNamingContext(globalCtx);

        // Finaly start catalina ...
        if (server instanceof Lifecycle) {
            try {
                server.initialize();
                ((Lifecycle) server).start();
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Cannot start catalina");
                throw new ServiceException(e.getMessage(), e);
            }
        }
        // Reset server context
        naming.setComponentContext(ctxold);

        started = true;
        // ... and deploy wars of the jonas.properties
        super.doStart();
    }

    /**
     * Stop the Catalina service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {
        // Undeploy the wars ...
        super.doStop();
        
        // ... and shut down embedded catalina 
        if (server instanceof Lifecycle) {
            try {
                ((Lifecycle) server).stop();
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "Cannot stop catalina");
                throw new ServiceException(e.getMessage(), e);
            }
        }

        started = false;
    }

    /**
     * Deploy a specific WAR file specified in the context.
     * @param ctx the context which contains the configuration in order 
     * to deploy the WAR.
     * @throws JWebContainerServiceException if the registration of the WAR 
     * failed.
     */
    protected void doRegisterWar(javax.naming.Context ctx) 
        throws JWebContainerServiceException {

        // Get the 5 parameters :
        //   - warURL is the URL of the war to register (required param).
        //   - contextRoot is the context root to which this application
        //     should be installed (must be unique) (required param).
        //   - hostName is the name of the host on which deploy the war
        //     (optional param taken into account only if no <context> element
        //      was declared in server.xml for this web application) .
        //   - java2DelegationModel the compliance to java2 delegation model
        //   - parentCL the war classloader of this war.

        URL warURL = null;
        String contextRoot = null;
        boolean java2DelegationModel = true;
        try {
            warURL = (URL) ctx.lookup("warURL");
            contextRoot = (String) ctx.lookup("contextRoot");
            Boolean bool = (Boolean) ctx.lookup("java2DelegationModel");
            java2DelegationModel = bool.booleanValue();
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }

        ClassLoader webClassLoader = null;

        try {
            webClassLoader = (ClassLoader) ctx.lookup("parentCL");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
        
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
            hostName = "";
        }

        String earAppName = null;
        try {
            earAppName = (String) ctx.lookup("earAppName");
        } catch (NamingException e) {
        }

        // Install a new web application, whose web application archive is 
        // at the specified URL, into this container with the specified 
        // context root.
        // A context root of "" (the empty string) should be used for the root
        // application for this container.  Otherwise, the context root must
        // start with a slash.

        if (contextRoot.equals("/")) {
            contextRoot = "";
        } else {
            contextRoot = "/" + contextRoot;
        }

        // install the war.
        File fWar = new File(warURL.getFile());
        String fileName = fWar.getName();
        String destDir = null;
        if (fWar.isDirectory()) {
            destDir = fWar.getPath();
        } else {
            destDir = WEBAPPS_DIR + File.separator + nameOfServer + File.separator;

            // Two cases :
            // war is alone --> unpack it in : WEBAPPS_DIR/servername/filename_without_extension_dir/
            // war come from an ear --> unpack it in WEBAPPS_DIR/ear_AppName/filename_without_extension_dir/


            // Note : if the war is alone, earAppName is null.
            // ear case
            if (earAppName != null) {
                destDir += earAppName + File.separator;
            } else {
                //war alone, no need of appName in the destDir
            }
            
            destDir += fileName.substring(0, fileName.length() - 4);
            // Unpack the war into a directory
            JarTools.unpack(warURL.getFile(), destDir);
        }

        // Check contexts of all services to see if there is a configured context.
        // that we must use instead of creating a new one.
        // If a context is matching, we set its docbase and start it.
        Service[] services = server.findServices();
        boolean deployed = false;

        // Get contexts of each engine of each service
        for (int s = 0; s < services.length; s++) {
            Container cont = services[s].getContainer();
            if (!(cont instanceof Engine)) {
                String err = "The container of the service must be an engine (server.xml)";
                throw new JWebContainerServiceException(err);  
            }
            Engine engine = (Engine) cont;
            Container hosts[] = engine.findChildren();

            for (int j = 0; j < hosts.length; j++) {
                Container contexts[] = hosts[j].findChildren();
                for (int k = 0; k < contexts.length; k++) {
                    Context sctx = (Context) contexts[k];
                    if (sctx != null) {
                        // The context was configured ?
                        String serverCtxRoot = sctx.getPath();
                        if (serverCtxRoot.equals(contextRoot)) {
                            try {
                                if (sctx instanceof JOnASStandardContext) {
                                    JOnASStandardContext stdCtx = (JOnASStandardContext) sctx;
                                    if (sctx.getPrivileged()) {
                                        // Can deploy a privileged context only if it's a .war file
                                        if (fWar.isDirectory()) {
                                            String err = "You are trying to deploy a privileged context with a webapp directory.";
                                            err += "Privileged contexts can be deployed by JOnAS only if there is a JOnASStandardContext and if it's a .war archive.";
                                            logger.log(BasicLevel.ERROR, err);
                                        }
                                        sctx.setDocBase(warURL.getFile());
                                        sctx.setParentClassLoader(webClassLoader);
                                    } else {
                                        sctx.setDocBase(destDir);
                                        sctx.setParentClassLoader(webClassLoader);
                                        sctx.setLoader(new JWebappLoader(webClassLoader));
                                    }
                                    stdCtx.start();
                                    logger.log(BasicLevel.DEBUG, "Customized context " + contextRoot + " started");
                                } else {
                                    String err = "You have a matching context " + contextRoot + ", but this context is not a JOnASStandardContext. Don't start it ! Maybe it was started by Tomcat";
                                    logger.log(BasicLevel.ERROR, err);
                                    throw new JWebContainerServiceException(err);
                                }
                                
                                // Check if the context was configured
                                if (!sctx.getConfigured()) {
                                    logger.log(BasicLevel.ERROR, "context " + contextRoot + " was not configured properly. See the traces for more information.");
                                } else {
                                    WebappClassLoader cl = (WebappClassLoader) sctx.getLoader().getClassLoader();
                                    if (cl != null) {
                                        cl.setDelegate(java2DelegationModel);
                                    }
                                    logger.log(BasicLevel.DEBUG, "Webapp class loader java 2 delegation model set to " + java2DelegationModel);
                                    unsetAuthenticationCaching(sctx);
                                }
                                deployed = true;

                            } catch (Exception e) {
                                String err = "Cannot start this web application " + e;
                                logger.log(BasicLevel.ERROR, err);
                                throw new JWebContainerServiceException(err);
                            }
                        }
                    }
                }
            }
        }

        // The context was not configured in server.xml, a new context will be created.
        if (!deployed) {
            // No context created
            Host deployer = findHost(hostName);
            
            // Create context
            StandardContext context = new StandardContext();
            int debug = 0;
            if (server instanceof StandardServer) {
                debug =  ((StandardServer) server).getDebug();
            }
            context.setDebug(debug);
            context.setDocBase(destDir);
            context.setPath(contextRoot);
            ContextConfig config = new ContextConfig();
            config.setDebug(debug);
            ((Lifecycle) context).addLifecycleListener(config);
            
            context.setParentClassLoader(webClassLoader);
            // Define our own loader
            context.setLoader(new JWebappLoader(webClassLoader));
            deployer.addChild(context);
            
            // Check if the context was configured
            if (!context.getConfigured()) {
                logger.log(BasicLevel.ERROR, "context " + contextRoot + " was not configured properly. See the traces for more information.");
            } else {
                WebappClassLoader cl = (WebappClassLoader) context.getLoader().getClassLoader();
                if (cl != null) {
                    cl.setDelegate(java2DelegationModel);
                }
                logger.log(BasicLevel.DEBUG, "Webapp class loader java 2 delegation model set to " + java2DelegationModel);
                unsetAuthenticationCaching(context);
                logger.log(BasicLevel.DEBUG, "context " + contextRoot + " started");
            }
        }
    }

    /** 
     * Undeploy a specific WAR file specified in the context.
     * @param ctx the context which contains the configuration in order
     * to undeploy a WAR.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    protected void doUnRegisterWar(javax.naming.Context ctx) 
        throws JWebContainerServiceException {

        // Get the 2 parameters :
        //   - contextRoot is the context root to be removed (required param).
        //   - hostName is the name of the host to remove the war (optional).
        String contextRoot = null;
        try {
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
    
        if (contextRoot.equals("/")) {
            contextRoot = "";
        } else {
            contextRoot = "/" + contextRoot;
        }

        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
           
        }

        // Undeploy all matching context if there is no hostname provided
        if (hostName == null) {
            Service[] services = server.findServices();
            boolean found = false;

            // Get contexts of each engine of each service
            for (int s = 0; s < services.length; s++) {
                Container cont = services[s].getContainer();
                if (!(cont instanceof Engine)) {
                    String err = "The container of the service must be an engine (server.xml)";
                    throw new JWebContainerServiceException(err);  
                }
                Engine engine = (Engine) cont;
                Container hosts[] = engine.findChildren();
                
                for (int j = 0; j < hosts.length; j++) {
                    Container contexts[] = hosts[j].findChildren();
                    for (int k = 0; k < contexts.length; k++) {
                        Context sctx = (Context) contexts[k];
                        if (sctx != null) {
                            //The context was configured ?
                            String serverCtxRoot = sctx.getPath();
                            if (serverCtxRoot.equals(contextRoot)) {
                                removeContext(sctx);
                                found = true;
                            }
                        }
                    }
                }
            }
            if (!found) {
                String err = "Cannot undeploy war: ";
                err = err + "Context does not exist for named path " + contextRoot;
                logger.log(BasicLevel.ERROR, err);
                throw new JWebContainerServiceException(err);
            }
        } else {
            // Remove the specified Context from the set of defined Contexts 
            // for its associated Host.
            Host host = findHost(hostName);
            org.apache.catalina.Context context = host.map(contextRoot);
            if (context != null) {
                removeContext(context);
            } else {
                String err = "Cannot undeploy war: ";
                err = err + "Context does not exist for named path " + contextRoot;
                logger.log(BasicLevel.ERROR, err);
                throw new JWebContainerServiceException(err);
            }
        }
    }


    /**
     * Remove the specified Context from the set of defined Contexts for its
     * associated Host.  If this is the last Context for this Host, the Host
     * will also be removed.
     *
     * @param context The Context to be removed
     */
    public synchronized void removeContext(Context context) {

        // Is this Context actually among those that are defined?
        boolean found = false;
        
        Service[] services = server.findServices();
        //get contexts of each engine of each service
        for (int s = 0; s < services.length; s++) {
            Container cont = services[s].getContainer();
            if (!(cont instanceof Engine)) {
                String err = "The container of the service must be an engine";
                throw new JWebContainerServiceException(err);  
            }
            Engine engine = (Engine) cont;
            Container hosts[] = engine.findChildren();
            for (int j = 0; j < hosts.length; j++) {
                Container contexts[] = hosts[j].findChildren();
                for (int k = 0; k < contexts.length; k++) {
                    if (context == (Context) contexts[k]) {
                        found = true;
                        break;
                    }
                }
                if (found) {
                    break;
                }
            }
        }
        if (!found) {
            return;
        }
        // Remove this Context from the associated Host
        context.getParent().removeChild(context);
    }


    /**
     * Set the server instance we are configuring.
     *
     * @param server The new server
     */
    public void setServer(Server server) {
        this.server = server;
    }

    /**
     * The server is started ?
     * @return boolean true if the catalina container is running.
     */
    public boolean isStarted() {
        return started;
    }



    /**
     * Return a File object representing the server.xml configuration file.
     * @return a File object representing the server.xml configuration file.
     * @throws FileNotFoundException if the configuration file is not found.
     */
    protected File getConfigFile() throws FileNotFoundException {
        String fileName = System.getProperty("catalina.base");
        fileName = fileName + File.separator + CONFIG_FILE;
        File file = new File(fileName);
        if (!file.exists()) {
            String err = "Can't find configuration file : " + fileName;
            throw new FileNotFoundException(err);
        }
        return (file);
    }

    /**
     * Find the specified host.
     * @param hostName the name of the host to find.
     * @return the host found.
     * @throws JWebContainerServiceException if the specified host cannot be
     * found.
     */
    public Host findHost(String hostName) 
        throws JWebContainerServiceException {

        Service[] services = server.findServices();
        // Check number of services
        if (services.length < 1) {
            String err = "At least one service must be define in the server.xml of Tomcat";
            throw new JWebContainerServiceException(err);  
        }

        // Two cases :
        //   1/ No host specified
        //      -> defaulthost of the engine of the first service.
        //   2) A host is specified in the jonas-web.xml
        //      -> find the host and deploy on this host


        if (hostName == null || hostName.equals("")) {
            // First case
            
            //Take first service
            Service service = services[0];
            
            Container cont = service.getContainer();
            if (!(cont instanceof Engine)) {
                String err = "The container of the service must be an engine";
                throw new JWebContainerServiceException(err);  
            }
        
            Engine engine = (Engine) cont;
            String defaultHost = engine.getDefaultHost();
            if (defaultHost == null) {
                String err = "Default host must be specified in server.xml or host must be specified in jonas-web.xml";
                throw new JWebContainerServiceException(err);
            }
            Container child = engine.findChild(defaultHost);
            if (child instanceof Host) {
                return (Host) child;
            } else {
                String err = "Default host " + defaultHost + " not found";
                throw new JWebContainerServiceException(err);
            }         
        } else {
            // Second case.
            // Get all hosts.
            Vector hosts = new Vector();

            for (int s = 0; s < services.length; s++) {
                Container cont = services[s].getContainer();
                if (!(cont instanceof Engine)) {
                    String err = "The container of a service must be an engine";
                    throw new JWebContainerServiceException(err);  
                }
                Engine engine = (Engine) cont;
                Container child = engine.findChild(hostName);
                if (child instanceof Host) {
                    hosts.addElement(child);
                }
            }

            if (hosts.size() == 0) {
                // No host found.
                String err = "Host " + hostName + " not found in all services/Engine of server.xml";
                throw new JWebContainerServiceException(err);
            } else {
                return (Host) hosts.elementAt(0);
            }
        } 
    }


    /**
     * Set the specified resources in the naming context.
     * @param ctx the context where the bindings are set.
     * @param resources all the naming resources.
     * @param resource the resource to bind.
     * @throws JWebContainerServiceException if the bind failed.
     */
    private void addResource(javax.naming.Context ctx, NamingResources resources, ContextResource resource) throws JWebContainerServiceException {

        String name = resource.getName();
        String type = resource.getType();
        String registryName = name;

        // Add only UserDatabase resources
        if (!type.equals("org.apache.catalina.UserDatabase")) {
            return;
        }

        // Create a reference to the resource.
        Reference ref = new Reference("org.apache.catalina.UserDatabase", "org.apache.catalina.users.MemoryUserDatabaseFactory", null);

        // Add extra parameters
        ResourceParams resourceParameters = resources.findResourceParams(name);
        if (resourceParameters != null) {
            Hashtable params = resourceParameters.getParameters();
            Enumeration enum = params.keys();
            while (enum.hasMoreElements()) {
                String paramName = (String) enum.nextElement();
                String paramValue = (String) params.get(paramName);
                StringRefAddr refAddr = new StringRefAddr(paramName, paramValue);
                ref.add(refAddr);
            }
        }
        try {
            javax.naming.Context ictx = NamingManager.getInstance().getInitialContext();

            // Bind into registry
            ictx.bind(registryName, ref);

            // Bind LinkRef
            LinkRef lref = new LinkRef(registryName);
            ctx.bind(name, lref);

        } catch (Exception e) {
            e.printStackTrace();
            String err = "Can't add the resource " + name;
            throw new JWebContainerServiceException(err, e);
        }

    }


    /** 
     * Disable cache of the authenticator of the specified context
     * @param ctx the context on which to disable authentication caching
     */
    protected void unsetAuthenticationCaching(Context ctx) {
        Valve[] valves = ((StandardContext) ctx).getValves();
        for(int i = 0; i < valves.length; i++) {
            Valve valve = valves[i];
            if (valve instanceof AuthenticatorBase) {
                AuthenticatorBase authBase = (AuthenticatorBase) valve;
                authBase.setCache(false);
            }
        }
    }

    /**
     * Create and configure the Digester that will be used for the xml
     * parsing of the configuration file.
     * @return Digester the digester containing the rules for the xml parsing
     * of the server.xml.
     */
    protected Digester createStartDigester() {

        // Initialize the digester
        Digester digester = new Digester();
        digester.setDebug(0);
        digester.setValidating(false);

        // Configure the actions we will be using
        digester.addObjectCreate("Server",
                                 "org.apache.catalina.core.StandardServer",
                                 "className");
        digester.addSetProperties("Server");
        digester.addSetNext("Server",
                            "setServer",
                            "org.apache.catalina.Server");

        digester.addObjectCreate("Server/GlobalNamingResources",
                                 "org.apache.catalina.deploy.NamingResources");
        digester.addSetProperties("Server/GlobalNamingResources");
        digester.addSetNext("Server/GlobalNamingResources",
                            "setGlobalNamingResources",
                            "org.apache.catalina.deploy.NamingResources");


        digester.addObjectCreate("Server/Listener",
                                 null, // MUST be specified in the element
                                 "className");
        digester.addSetProperties("Server/Listener");
        digester.addSetNext("Server/Listener",
                            "addLifecycleListener",
                            "org.apache.catalina.LifecycleListener");

        digester.addObjectCreate("Server/Service",
                                 "org.apache.catalina.core.StandardService",
                                 "className");
        digester.addSetProperties("Server/Service");
        digester.addSetNext("Server/Service",
                            "addService",
                            "org.apache.catalina.Service");
        
        digester.addObjectCreate("Server/Service/Listener",
                                 null, // MUST be specified in the element
                                 "className");
        digester.addSetProperties("Server/Service/Listener");
        digester.addSetNext("Server/Service/Listener",
                            "addLifecycleListener",
                            "org.apache.catalina.LifecycleListener");

        digester.addObjectCreate("Server/Service/Connector",
                                 "org.apache.catalina.connector.http.HttpConnector",
                                 "className");
        digester.addSetProperties("Server/Service/Connector");
        digester.addSetNext("Server/Service/Connector",
                            "addConnector",
                            "org.apache.catalina.Connector");

        digester.addObjectCreate("Server/Service/Connector/Factory",
                                 "org.apache.catalina.net.DefaultServerSocketFactory",
                                 "className");
        digester.addSetProperties("Server/Service/Connector/Factory");
        digester.addSetNext("Server/Service/Connector/Factory",
                            "setFactory",
                            "org.apache.catalina.net.ServerSocketFactory");

        digester.addObjectCreate("Server/Service/Connector/Listener",
                                 null, // MUST be specified in the element
                                 "className");
        digester.addSetProperties("Server/Service/Connector/Listener");
        digester.addSetNext("Server/Service/Connector/Listener",
                            "addLifecycleListener",
                            "org.apache.catalina.LifecycleListener");

        // Add RuleSets for nested elements
        digester.addRuleSet(new NamingRuleSet("Server/GlobalNamingResources/"));
        digester.addRuleSet(new EngineRuleSet("Server/Service/"));
        digester.addRuleSet(new JHostRuleSet("Server/Service/Engine/"));
        digester.addRuleSet(new JContextRuleSet("Server/Service/Engine/Default", this));
        //digester.addRuleSet(new NamingRuleSet("Server/Service/Engine/DefaultContext/"));
        digester.addRuleSet(new JContextRuleSet("Server/Service/Engine/Host/Default", this));
        //digester.addRuleSet(new NamingRuleSet("Server/Service/Engine/Host/DefaultContext/"));
        digester.addRuleSet(new JContextRuleSet("Server/Service/Engine/Host/", this));
        //digester.addRuleSet(new NamingRuleSet("Server/Service/Engine/Host/Context/"));
        digester.addRule("Server/Service/Engine", new SetParentClassLoaderRule(digester, parentClassLoader));

        return (digester);

    }

    /**
     * Class that sets the parent class loader for the top class on the stack.
     * @author take from Catalina.
     */
    final class SetParentClassLoaderRule extends Rule {

        /**
         * The parent class loader to be set.
         */
        private ClassLoader parentClassLoader = null;


        /**
         * Construct a new action.
         * @param digester the digester need for the rule
         * @param parentClassLoader The new parent class loader
         */
        public SetParentClassLoaderRule(Digester digester,
                                        ClassLoader parentClassLoader) {
            super(digester);
            this.parentClassLoader = parentClassLoader;
        }

        /**
         * Add the requested parent class loader.
         * @param attributes the attributes.
         * @throws Exception if an error occurs.
         */
        public void begin(Attributes attributes) throws Exception {
            Container top = (Container) digester.peek();
            top.setParentClassLoader(parentClassLoader);
            
        }
    }

}





