/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Miroslav Halas based on Catalina service
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.jetty;

//import java
import java.util.Stack;
import java.util.Vector;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.lang.reflect.Constructor;

//import javax
import javax.servlet.ServletContext;
import javax.naming.NamingException;
import javax.naming.Context;
import javax.management.ObjectName;

//import objectweb
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

//import jonas
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImpl;
import org.objectweb.jonas.web.JWebContainerServiceException;


//import jetty
import org.mortbay.http.*;
import org.mortbay.jetty.*;
import org.mortbay.jetty.jmx.*;
import org.mortbay.util.*;
import org.mortbay.util.jmx.*;
import org.mortbay.jetty.servlet.*;

/**
 * This class provides an implementation of the Jetty service (as 
 * web container service).
 * This works only for Jetty version 2.4.2 or higher.
 * @author Miroslav Halas
 * Contributor(s):
 * Florent Benoit 
 *   Now works with the JSPs.(like alarm example).
 *     -> Need to specify a classpath required by jasper compiler.
 *   Can deploy directories.
 */
public class JettyJWebContainerServiceImpl extends AbsJWebContainerServiceImpl {

    /**
     * Private logger for this class.
     */
    private static Logger m_lgLogger = null;

    /**
     * Configuration used to configure Jetty.
     */
    private static String m_strConfig = null;
      
    /**
     * Name of the variable specifying configuration file for Jetty
     */
    private static final String JETTY_CONFIG = "config";

    /**
     * Our own instance of Jetty server.
     */
    private Server m_srvJettyServer = null;

    /**
     * The shared extensions class loader for this server.
     */
    protected ClassLoader parentClassLoader =
        Thread.currentThread().getContextClassLoader();

    /**
     * Initialize the Jetty service.
     * @param ctx the configuration context of the service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(javax.naming.Context ctx) throws ServiceException {

        super.doInit(ctx);
     
        /*
          TODO: Jonas and Jetty security are not yet integrated
          boolean useSecurity = false;
          try 
          {
          ServiceManager.getSecurityService();
          useSecurity = true;
          } catch (Exception e) 
          {
          // nothing to do 
          }
        */
        
        String strJettyHome = System.getProperty("jetty.home");
        if (strJettyHome != null) {
            if (m_lgLogger == null) {
                m_lgLogger = Log.getLogger(Log.JONAS_WEB_PREFIX);
            }
      
            m_lgLogger.log(BasicLevel.DEBUG, "");
      
	    try {
		m_strConfig = (String)ctx.lookup(JETTY_CONFIG);          
	    } catch (NamingException neExc) {
		// by default we take jetty.xml under JONAS_BASE/conf
		String jonasbase = System.getProperties().getProperty("jonas.base");
		m_strConfig = jonasbase + "/conf/jetty.xml";           
	    }
	    m_lgLogger.log(BasicLevel.LEVEL_DEBUG, "using configuration file "+ m_strConfig);

            m_srvJettyServer = new Server();
        }
    }

    /**
     * Start the Jetty service in a new thread
     * @throws ServiceException if the startup failed.
     */
    public void doStart() throws ServiceException {       
        m_lgLogger.log(BasicLevel.DEBUG, "");

        if (m_srvJettyServer != null) {
            JmxService srvcJMX = null;

            // Start Jetty directly and just pass it to JMX if available
            if (m_strConfig != null) {
                try {
                    m_srvJettyServer.configure(m_strConfig);
                    m_srvJettyServer.start();
                } catch (Exception eExc) {
                    m_lgLogger.log(BasicLevel.LEVEL_ERROR, "error has occured while starting Jetty server using configuration file " 
                                   + m_strConfig, eExc);
                }
            }

            try {
                srvcJMX = (JmxService) ServiceManager.getJmxService();
                if (srvcJMX != null) {
                    // Register JettyService MBean : JettyService 
                    Object obj;
                    ObjectName objname;
                    /*
                      obj = new LogMBean();
                      objname = new ObjectName("jonas:type=jetty,name=jettylog");
                      srvcJMX.getJmxServer().registerMBean(obj,objname);
                    */
                    // Once we register this bean, it will start Jetty using given configuration 
                    // file. That's why we have to do it in run() instead in doStart()
                    obj = new JettyJonasServerMBean(m_srvJettyServer);
                    objname = JonasObjectName.wwwService();
                    srvcJMX.getJmxServer().registerMBean(obj, objname);
                    obj = new CodeMBean();
                    objname = JonasObjectName.www("Code");
                    srvcJMX.getJmxServer().registerMBean(obj, objname);         
                }
            } catch (javax.management.MalformedObjectNameException moneExc) {
                // this should never occur
                m_lgLogger.log(BasicLevel.LEVEL_ERROR, 
                               "cannot start Jetty server using configuration file " 
                               + m_strConfig, moneExc);
            } catch (ServiceException seExc) {
                // Jmx Service not available, do nothing
                m_lgLogger.log(BasicLevel.LEVEL_DEBUG, 
                               "cannot start Jetty server using configuration file " 
                               + m_strConfig + " using JMX. Will start without JMX.", seExc);
            } catch (Exception eExc) {
                m_lgLogger.log(BasicLevel.LEVEL_ERROR, 
                               "cannot start Jetty server using configuration file " 
                               + m_strConfig, eExc);
                throw new ServiceException("Cannot start Jetty server using configuration file " 
                                           + m_strConfig, eExc);
            }

            // ... and deploy wars of the jonas.properties
            super.doStart();
        } else {
            throw new ServiceException("Cannot start Jetty server.");
        }
    }

    /**
     * Stop the Jetty service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {
        // Undeploy the wars ...
        super.doStop();
        
        // ... and shut down embedded jetty 
        m_lgLogger.log(BasicLevel.DEBUG, "");
        if (isStarted()) {
            JmxService srvcJMX = null;

            try {
                srvcJMX = (JmxService) ServiceManager.getJmxService();
                if (srvcJMX != null) {
                    ObjectName objname;
            
                    objname = new ObjectName("jonas:type=service,name=jetty");
                    srvcJMX.getJmxServer().unregisterMBean(objname);
                    m_srvJettyServer = null;
                    objname = new ObjectName("jonas:type=jetty,name=jettylog");
                    srvcJMX.getJmxServer().unregisterMBean(objname);
                    objname = new ObjectName("jonas:type=jetty,name=jettycode");
                    srvcJMX.getJmxServer().unregisterMBean(objname);            
                }
            } catch (ServiceException seExc) {
                // Jetty Service not available, do nothing
            } catch (Exception eExc) {
                m_lgLogger.log(BasicLevel.LEVEL_ERROR, "Cannot stop Jetty server", eExc);
                throw new ServiceException("Cannot stop Jetty server", eExc);
            }

            if (m_srvJettyServer != null) {
                try {
                    m_srvJettyServer.stop();
                    m_srvJettyServer.destroy();
                    m_srvJettyServer = null;
                } catch (Exception eExc) {
                    m_lgLogger.log(BasicLevel.LEVEL_ERROR, 
                                   "error has occured while stopping Jetty server using configuration file " 
                                   + m_strConfig, eExc);
                }
            }
        }        
    }

    /**
     * Create the environment and delegate the operation to the 
     * implementation of the web container.
     * @param ctx the context which contains the configuration in order 
     * to deploy a WAR.
     * @throws JWebContainerServiceException if the registration of the 
     * WAR failed.
     */
    protected void doRegisterWar(Context ctx) throws JWebContainerServiceException {
        // Get the 5 parameters :
        //   - warURL is the URL of the war to register (required param).
        //   - contextRoot is the context root to which this application
        //     should be installed (must be unique) (required param).
        //   - hostName is the name of the host on which deploy the war
        //     (optional param taken into account only if no <context> element
        //      was declared in server.xml for this web application) .
        //   - java2DelegationModel the compliance to java2 delegation model
        //   - parentCL the war classloader of this war.
        URL warURL = null;
        String contextRoot = null;
        boolean java2DelegationModel = true;
        try {
            warURL = (URL) ctx.lookup("warURL");
            contextRoot = (String) ctx.lookup("contextRoot");
            Boolean bool = (Boolean) ctx.lookup("java2DelegationModel");
            java2DelegationModel = bool.booleanValue();
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            m_lgLogger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
        
        ClassLoader webClassLoader = null;
        try {
            webClassLoader = (ClassLoader) ctx.lookup("parentCL");
        } catch (NamingException e) {
            String err = "error while getting parameter from context param ";
            m_lgLogger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
        
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        }  catch (NamingException e) {
            hostName = "";
        }
        
        // Install a new web application, whose web application archive is 
        // at the specified URL, into this container with the specified 
        // context root.
        // A context root of "" (the empty string) should be used for the root
        // application for this container.  Otherwise, the context root must
        // start with a slash.
        
        if (contextRoot.equals("/")) {
            contextRoot = "";
        } else if (contextRoot.equalsIgnoreCase("ROOT")) {
            // Jetty uses ROOT.war and ROOT directory to as root context
            contextRoot = "";
        }

        // install the war.
        File fWar = new File(warURL.getFile());
        String fileName = fWar.getAbsolutePath();

        if (m_srvJettyServer != null) {
            try {
                WebApplicationContext contextWebApp;

                if ((hostName == null) || (hostName.length() == 0)) {
                    // There is no host
                    m_lgLogger.log(BasicLevel.DEBUG, 
                                   "Jetty server installing web app " + fileName
                                   + " and context " + contextRoot);
                    contextWebApp = m_srvJettyServer.addWebApplication(contextRoot, fileName);
                } else {
                    // Host was specified
                    m_lgLogger.log(BasicLevel.DEBUG, 
                                   "Jetty server installing web app " + fileName
                                   + " on host " + hostName + " and context " + contextRoot);
                    contextWebApp = m_srvJettyServer.addWebApplication(hostName, contextRoot, fileName);
                }

                // Specifying the jsp class path used by jasper
                contextWebApp.setAttribute("org.apache.catalina.jsp_classpath", getJOnASClassPath(webClassLoader));

                //Don't unpack a directory (if we deploy a directory, not a war)
                if (!fWar.isDirectory()) {
                    contextWebApp.setExtractWAR(true);
                }
                // Set the parent class loader
                contextWebApp.setParentClassLoader(webClassLoader);
                
                // Set this classloader to the Java2 compliant mode ?
                contextWebApp.setClassLoaderJava2Compliant(java2DelegationModel);

                logger.log(BasicLevel.DEBUG, "Webapp class loader java 2 delegation model set to " + java2DelegationModel);

                m_lgLogger.log(BasicLevel.DEBUG, "Jetty server starting web app " + fileName);
                contextWebApp.start();
                m_lgLogger.log(BasicLevel.DEBUG, "Jetty server is running web app " + fileName);
            } catch (IOException ioeExc) {
                String err = "Cannot install this web application " + ioeExc;
                m_lgLogger.log(BasicLevel.ERROR, err);
                throw new JWebContainerServiceException(err, ioeExc);
            } catch (Exception eExc) {
                String err = "Cannot start this web application " + eExc;
                m_lgLogger.log(BasicLevel.ERROR, err);
                throw new JWebContainerServiceException(err, eExc);
            }
        } else {
            m_lgLogger.log(BasicLevel.DEBUG, "No Jetty server to install web app " + fileName);
        }
    }


    /**
     * Return the classpath which can be used for jsp compiling by Jasper.
     * This classpath is extracted from the web classloader.
     * @param webClassLoader the ClassLoader used for extract URLs.
     * @return the jonas classpath which is useful for JSP compiling.
     */
    public String getJOnASClassPath(ClassLoader webClassLoader) {

        StringBuffer classpath = new StringBuffer();
        int n = 0;
        while (webClassLoader != null) {
            if (!(webClassLoader instanceof URLClassLoader)) {
                break;
            }
            URL repositories[] = ((URLClassLoader) webClassLoader).getURLs();
            for (int i = 0; i < repositories.length; i++) {
                String repository = repositories[i].toString();
                if (repository.startsWith("file://")) {
                    repository = repository.substring(7);
                } else if (repository.startsWith("file:")) {
                    repository = repository.substring(5);
                } else {
                    continue;
                }
                if (repository == null) {
                    continue;
                }
                if (n > 0) {
                    classpath.append(File.pathSeparator);
                }
                classpath.append(repository);
                n++;
            }
            webClassLoader = webClassLoader.getParent();
        }

        return classpath.toString();
    }

    /**
     * Delegate the unregistration to the implementation of the web container.
     * @param ctx the context which contains the configuration in order to 
     * undeploy a WAR.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    protected void doUnRegisterWar(Context ctx) throws JWebContainerServiceException {
        // Get the 2 parameters :
        //   - contextRoot is the context root to be removed (required param).
        //   - hostName is the name of the host to remove the war (optional).
        String contextRoot = null;
        try {
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
    
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
            hostName = "";
        }
        // A context root of "" (the empty string) should be used for the root
        // application for this container.  Otherwise, the context root must
        // start with a slash.
        
        if (contextRoot.equals("/")) {
            contextRoot = "";
        } else if (contextRoot.equalsIgnoreCase("ROOT")) {
            // Jetty uses ROOT.war and ROOT directory to as root context
            contextRoot = "";
        }

        if (m_srvJettyServer != null) {
            HttpContext contextWebApp;

            if ((hostName == null) || (hostName.length() == 0)) {
                // There is no host
                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server looking upweb app "
                               + " from context " + contextRoot);
                contextWebApp = m_srvJettyServer.getContext(contextRoot);
            } else {
                // Host was specified
                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server looking up web app "
                               + " on host " + hostName + " and context " + contextRoot);
                contextWebApp = m_srvJettyServer.getContext(hostName, contextRoot);
            }

            if (contextWebApp != null) {
                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server found and is stopping web app at context " 
                               + contextRoot);
              

                // Stop it gracefully
                try {
                    contextWebApp.stop(true);
                } catch (InterruptedException ieExc) {
                    m_lgLogger.log(BasicLevel.LEVEL_DEBUG, 
                                   "Jetty server encoutered exception while stopping web application ", ieExc);
                }
              
                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server stopped and is removing web app at context " 
                               + contextRoot);

                m_srvJettyServer.removeContext(contextWebApp);

                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server removed and is destroying web app at context " 
                               + contextRoot);

                contextWebApp.destroy();

                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server unloaded web app at context " + contextRoot);
            } else {
                m_lgLogger.log(BasicLevel.DEBUG, 
                               "Jetty server didn't find web app at context " + contextRoot);
            }

        } else {
            m_lgLogger.log(BasicLevel.DEBUG, 
                           "No Jetty server to install web app at context " + contextRoot);
        }
    }
}

