/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Philippe Coq
 * Contributor(s): 
 * Jeff Mesnil : security
 * Christophe Ney : PrincipalFactory
 * Philippe Durieux : add missing methods
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.container;

import java.security.Guard;
import javax.transaction.TransactionManager;
import org.objectweb.jonas_lib.naming.ContainerNaming;
import org.objectweb.jonas_jms.api.JmsManager;

/**
 * A Container is what a JOnAS Server can see of a JOnAS Container.
 * It holds all beans coming from a same ejbjar file.
 * There is no possibility to add or remove dynamically beans in 
 * a Container since it is build with all its beans.
 * JMX can use this interface, but no specific Jmx classes should
 * be inside the Container. Its management is done outside jonas_ejb.
 * @author Philippe Coq
 */
public interface Container {

    /**
     * @return List of beans hosted in this Container
     */
    String [] listBeanNames();

    /**
     * @return name of this Container
     */
    String getName();

    /**
     * @return the file name of the container (.xml or .jar)
     */
    String getFileName();

    /**
     * Set the MethodGuard
     * @param methodGuard The Guard to be set for methods.
     */
    void setMethodGuard(Guard methodGuard);

    /**
     * @return the Method Guard of the Container
     */
    Guard getMethodGuard();

    /**
     * Set the Role Guard
     * @param roleGuard The Guard to be set for roles.
     */
    void setRoleGuard(Guard roleGuard);

    /**
     * @return the Role Guard of the Container
     */
    Guard getRoleGuard();

    /**
     * @return the classloader used for this Container
     */
    ClassLoader getClassLoader();

    /**
     * Set the PrincipalFactory. This factory can be JOnAS Server dependant.
     * The Container makes no assumption on how to get the Principal.
     * @param pf the PrincipalFactory
     */
    void setPrincipalFactory(PrincipalFactory pf);

    /**
     * @return the PrincipalFactory of the Container
     */
    PrincipalFactory getPrincipalFactory();

    /**
     * set the ContainerNaming object
     * @param naming the ContainerNaming object
     */
    void setContainerNaming(ContainerNaming naming);

    /**
     * @return the ContainerNaming object
     */
    ContainerNaming getContainerNaming();

    /**
     * Set the thread pool size. This thread pool is currently used
     * only for message driven beans.
     * @param sz number of threads in the pool.
     */
    void setThreadPoolSize(int sz);

    /**
     * Synchronize all entity beans
     * @param passivate true if bean instances will be released after
     * having been written on storage.
     */
    void syncAll(boolean passivate);

    /**
     * set the Transaction Manager.
     * @param tm the Transaction Manager.
     */
    void setTransactionManager(TransactionManager tm);

    /**
     * @return the Transaction Manager
     */
    TransactionManager getTransactionManager();

    /**
     * set the JmsManager object
     * @param jms the JmsManager
     */
    void setJmsManager(JmsManager jms);

    /**
     * @return the JmsManager object
     */
    JmsManager getJmsManager();

    /**
     * set the name of the ear application containing this container.
     * @param fileName the name of the ear application containing this 
     * container.
     */
    void setEarFileName(String fileName);

    /**
     * get the name of the ear application containing this container.
     * @return the name of the ear application containing this 
     * container.
     */
    String getEarFileName();

    /**
     * Remove the JOnAS container and unregister all beans.
     */
    void remove();

}
