/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.security.Guard;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;
import java.util.Vector;
import java.util.Properties;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;
import javax.ejb.AccessLocalException;
import javax.ejb.EJBException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.StringRefAddr;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;

import org.objectweb.common.Env;
import org.objectweb.jonas.dbm.ConnectionManager;
import org.objectweb.jonas_ejb.container.jorm.RdbFactory;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbLocalRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityBmpDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmpDesc;
import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp2Desc;
import org.objectweb.jonas_jms.api.JmsManager;
import org.objectweb.jonas_lib.naming.ContainerNaming;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class represents an EJB container.
 *
 * A container is where an enterprise Bean object lives.
 * All beans from a same ejb-jar file are installed in a single container.
 * For each EJB installed, the container provides a factory and
 * makes it available in the JNDI name space.
 * The JContainer basically manages a set of BeanFactory objects.
 * @author Philippe Coq
 * @author Jeff Mesnil     (Security)
 * @author Christophe Ney   (Making easier Enhydra integration)
 * @author Philippe Durieux (New architecture for local interfaces)
 * @author Florent Benoit, Ludovic Bert (Ear service, ejb-link)
 */

public class JContainer implements Container {

    /**
     * This class is the default factory class name used for the Entity bean.
     */
    public static final String DEFAULT_FACTORY_CLASS_NAME =
        "org.objectweb.jonas_ejb.container.JEntityFactory";

    private Swapper swapper;
    private ClassLoader loader = null;
    private ContainerNaming naming = null;
    private TransactionManager tm = null;
    private JmsManager jms = null;
    private DeploymentDesc deploymentDesc = null;

    public String myname; // Container name
    public String fileName; // Container file name x.jar or x.xml
    private HashMap beanList = new HashMap(); // key=EJBName / Value=BeanFactory
    private Guard methodGuard = null;
    private Guard roleGuard = null;
    private PrincipalFactory principalFactory = null;

    // Name of the ear application containing this container.
    private String earFileName = null;

    public JStatefulPool statefulSessionPool = null;

    // Pool of threads for Message Driven Beans
    // We use only 1 pool for all containers.
    private static final int DEFAULT_THREAD_POOL_SIZE = 10;
    int threadpoolsize = DEFAULT_THREAD_POOL_SIZE;
    static ThreadPool threadpool = null;

    /**
     * constructor
     * @param name name of the container.
     * @param file file name (.jar or .xml)
     * @param ld the class loader to be used
     */
    public JContainer(String name, String file, ClassLoader ld) {
        myname = name;
        fileName = file;
        loader = ld;
        swapper = new Swapper(this);
        swapper.start();
    }

    // ---------------------------------------------------------------
    // Container Implementation
    // ---------------------------------------------------------------

    /**
     * @return List of beans hosted in this Container
     */
    public String [] listBeanNames() {

        String ret[] = new String[beanList.size()];
        Vector namelist = new Vector();
        BeanFactory bf = null;
        return (String[]) beanList.keySet().toArray(new String[0]);
    }

    /**
     * @return name of this Container
     */
    public String getName() {
        return myname;
    }

    /**
     * @return the file name of the container (.xml or .jar)
     */
    public String getFileName() {
        return fileName;
    }

    /**
     * Set a MethodGuard
     * @param methodGuard The Guard to be set for methods.
     */
    public void setMethodGuard(Guard methodGuard) {
        if (TraceEjb.isDebugSecurity()) {
            TraceEjb.security.log(BasicLevel.DEBUG, myname + ": methodGuard= " + methodGuard + ")");
        }
        this.methodGuard = methodGuard;
    }

    /**
     * @return the Method Guard of the Container
     */
    public Guard getMethodGuard() {
        return methodGuard;
    }

    /**
     * Set the Role Guard
     * @param roleGuard The Guard to be set for roles.
     */
    public void setRoleGuard (Guard roleGuard) {
        if (TraceEjb.isDebugSecurity()) {
            TraceEjb.security.log(BasicLevel.DEBUG, myname + ": roleGuard(" + roleGuard + ")");
        }
        this.roleGuard = roleGuard;
    }

    /**
     * @return the Role Guard of the Container
     */
    public Guard getRoleGuard() {
        return roleGuard;
    }

    /**
     * @return the classloader used for this Container
     */
    public ClassLoader getClassLoader() {
        if (loader == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "container has been removed");
            //return Thread.currentThread().getContextClassLoader();
            return null;
        } else {
            return loader;
        }
    }

    /**
     * Set the PrincipalFactory. This factory can be JOnAS Server dependant.
     * The Container makes no assumption on how to get the Principal.
     * @param pf the PrincipalFactory
     */
    public void setPrincipalFactory(PrincipalFactory pf) {
        principalFactory = pf;
    }

    /**
     * @return the PrincipalFactory of the Container
     */
    public PrincipalFactory getPrincipalFactory() {
        return principalFactory;
    }

    /**
     * set the ContainerNaming object
     * Called by the EJB Server when starting the service.
     * @param naming the ContainerNaming object
     */
    public void setContainerNaming(ContainerNaming naming) {
        this.naming = naming;
    }

    /**
     * used internally by all the EJB Container classes.
     * @return the ContainerNaming object
     */
    public ContainerNaming getContainerNaming() {
        return naming;
    }

    /**
     * set the Thread pool size.
     * @param sz number of threads in the pool.
     */
    public void setThreadPoolSize(int sz) {
        threadpoolsize = sz;
    }

    /**
     * Try to passivate all entity bean instances
     * @param passivate true if bean instances will be released after
     * having been written on storage.
     */
    public void syncAll(boolean passivate) {
        TraceEjb.interp.log(BasicLevel.DEBUG, myname);
        BeanFactory bf = null;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (passivate) {
                bf.reduceCache();
            } else {
                bf.sync();
            }
        }
    }

    /**
     * set the Transaction Manager.
     * Called by the EJB Server when starting the service.
     * @param tm the Transaction Manager.
     */
    public void setTransactionManager(TransactionManager tm) {
        this.tm = tm;
    }

    /**
     * return the Transaction Manager
     * used internally by all the EJB Container classes.
     * @return the Transaction Manager
     */
    public TransactionManager getTransactionManager() {
        return tm;
    }

    /**
     * set the JmsManager object
     * Called by the EJB Server when starting ths service.
     * @param jms the JmsManager
     */
    public void setJmsManager(JmsManager jms) {
        this.jms = jms;
    }

    /**
     * used internally by all the EJB Container classes.
     * @return the JmsManager object
     */
    public JmsManager getJmsManager() {
        return jms;
    }

    /**
     * set the name of the ear application containing this container.
     * @param fileName the name of the ear application containing this
     * container.
     */
    public void setEarFileName(String fileName) {
        earFileName = fileName;
    }

    /**
     * get the name of the ear application containing this container.
     * @return the name of the ear application containing this container.
     */
    public String getEarFileName() {
        return earFileName;
    }

    /**
     * Remove the JOnAS container and unregister all beans.
     */
    public void remove() {
        TraceEjb.interp.log(BasicLevel.DEBUG, myname);

        // Stop the swapper.
        swapper.stopIt();

        // synchronize all entity beans before removing them.
        // must be done here, not in the swapper.
        syncAll(false);

        // remove all factories
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            BeanFactory bf = (BeanFactory) it.next();
            try {
                bf.stop();
                TraceEjb.interp.log(BasicLevel.INFO, myname + ": " + bf.getEJBName() + " no longer available");
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, myname, e);
            }
        }
        beanList.clear();

        // to unload all classes.
        loader = null;
    }


    // -------------------------------------------------------------------
    // Methods used by JOnAS Server (but not part of JContainer interface)
    // -------------------------------------------------------------------

    /**
     * Adds beans in container.
     * This method is not part of JContainer interface, although it
     * is used by JOnAS Server, because we don't want to get all BeanDesc
     * classes (jonas_ejb.deployment.api)
     * LATER: Replace this by setDeploymentDesc ?
     * @param dd  The Bean Deployment Descriptor
     * @return The bean factory created for this bean.
     * @throws EJBException
     */
    public BeanFactory addBean(BeanDesc dd) {
        BeanFactory bf = null;
        String beanName = dd.getEjbName();
        if (dd instanceof SessionStatefulDesc) {
            TraceEjb.interp.log(BasicLevel.DEBUG, "add SessionStatefulBean " + beanName);
            if (statefulSessionPool == null) {
                statefulSessionPool = new JStatefulPool();
            }
            bf = new JStatefulFactory((SessionStatefulDesc) dd, this, statefulSessionPool);
        } else if (dd instanceof SessionStatelessDesc) {
            TraceEjb.interp.log(BasicLevel.DEBUG, "add SessionStatelessBean " + beanName);
            bf = new JStatelessFactory((SessionStatelessDesc) dd, this);
        } else if (dd instanceof MessageDrivenDesc) {
            TraceEjb.interp.log(BasicLevel.DEBUG, "add MessageDrivenBean " + beanName);
            if (threadpool == null) {
                threadpool = new ThreadPool(threadpoolsize);
            }
            bf = new JMdbFactory((MessageDrivenDesc) dd, this, threadpool);
        } else if (dd instanceof EntityJdbcCmp2Desc) {
            EntityJdbcCmp2Desc ecd = (EntityJdbcCmp2Desc) dd;
            // TODO: We only manage "Jdbc" CMP2 entity beans today.
            TraceEjb.interp.log(BasicLevel.DEBUG, "add CMP2 EntityBean " + beanName);
            // Instanciate the Factory depending on the mapper
            String cn = null;
            try {
                String dsn = ecd.getDatasourceJndiName();
                // XXX Ugly dependency on the DBM, but how can we do better ?
                // Pb: We need the mapper to instanciate the Factory.
                // The mapper is in the Datasource. No way to get it in a standard way
                // since Datasource interface is very simple (getConnection() only)
                InitialContext ictx = naming.getInitialContext();
                ConnectionManager datasource = (ConnectionManager) ictx.lookup(dsn);
                String mapperName = datasource.getMapperName();
                cn = ecd.getFactoryClassName(mapperName);
                bf = (JEntityFactory) loader.loadClass(cn).newInstance();
                ((RdbFactory) bf).init((EntityDesc) ecd, this, mapperName);
                // Set the passivation timeout
                setSwapTime(((EntityDesc) ecd).getPassivationTimeout());
                TraceEjb.interp.log(BasicLevel.INFO, beanName + " is loaded and using " + mapperName);
            } catch (Exception e) {
                throw new EJBException("Impossible to instanciate the entity factory: " + cn, e);
            }
        } else if (dd instanceof EntityDesc) {
            TraceEjb.interp.log(BasicLevel.DEBUG, "add EntityBean " + beanName);
            // Instanciate the default factory
            String cn = null;
            try {
                cn = DEFAULT_FACTORY_CLASS_NAME;
                bf = (JEntityFactory) loader.loadClass(cn).newInstance();
            } catch (Exception e) {
                throw new EJBException("Impossible to instanciate the specified entity factory: " + cn, e);
            }
            ((JEntityFactory) bf).init((EntityDesc) dd, this);
            // Set the passivation timeout
            setSwapTime(((EntityDesc) dd).getPassivationTimeout());
        } else {
            throw new EJBException("Bad Descriptor Type for " + beanName);
        }
        beanList.put(beanName, bf);
        TraceEjb.interp.log(BasicLevel.INFO, beanName + " available");
        return bf;
    }

    /**
     * Get the bean factory for the given bean.
     * @param ejbName the name of the bean
     * @return the bean factory for this bean
     */
    public BeanFactory getBeanFactory(String ejbName) {
        return (BeanFactory) beanList.get(ejbName);
    }

    /**
     * @return total Number of Beans
     */
    public int getBeanNb() {
        return beanList.size();
    }

    /**
     * @return int Number of BMP type currently in this container
     */
    public int getEntityBMPNb() {
        BeanFactory bf;
        int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (bf.getDeploymentDescriptor() instanceof EntityBmpDesc) {
                total++;
            }
        }
        return total;
    }

    /**
     * @return int Number of CMP type currently in this container
     */
    public int getEntityCMPNb() {
        BeanFactory bf;
        int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (bf.getDeploymentDescriptor() instanceof EntityCmpDesc) {
                total++;
            }
        }
        return total;
    }

    /**
     * @return int Number of SBF type currently in this container
     */
    public int getStatefulSessionNb() {
        BeanFactory bf;
        int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (bf.getDeploymentDescriptor() instanceof SessionStatefulDesc) {
                total++;
            }
        }
        return total;
    }

    /**
     * @return int Number of SBL type currently in this container
     */
    public int getStatelessSessionNb() {
        BeanFactory bf;
        int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (bf.getDeploymentDescriptor() instanceof SessionStatelessDesc) {
                total++;
            }
        }
        return total;
    }

    /**
     * @return int Number of MDB type currently in this container
     */
    public int getMessageDrivenNb() {
        BeanFactory bf;
        int total = 0;
        Iterator it = beanList.values().iterator();
        while (it.hasNext()) {
            bf = (BeanFactory) it.next();
            if (bf.getDeploymentDescriptor() instanceof MessageDrivenDesc) {
                total++;
            }
        }
        return total;
    }

    /**
     * @return passivation time out for this container
     */
    public long getSwapTime() {
        return swapper.getTimeout();
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the ejbs using a given data source.
     * @param dsName JNDI name of the data source
     * @return a set of Properties describing the beans that use the data source
     */
    public Set getDataSourceDependence(String dsName) {
        // the true param corresponds to a resource-ref
        return beansDependence(Collections.enumeration(beanList.values()), dsName, true);
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the beans using a JMS destination.
     * @param destName JNDI name of a JMS destination
     * @return a set of Properties describing the ejbs that use the
     * JMS destination.
     */
    public Set getJmsDestinationDependence(String destName) {
        // the true param corresponds to a resource-ref
        return beansDependence(Collections.enumeration(beanList.values()), destName, false);
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the ejbs using a JMS Connection Factory.
     * @param cfName JNDI name of a JMS Connection Factory.
     * @return a set of Properties describing the beans that use the
     * JMS Connection Factory.
     */
    public Set getJmsConnectionFactoryDependence(String cfName) {
        // the true param corresponds to a resource-ref
        return beansDependence(Collections.enumeration(beanList.values()), cfName, true);
    }

    /**
     * Management method used by the EJBServiceImpl MBean. Determine
     * which are the ejbs using a Mail Factory.
     * @param mfName JNDI name of a Mail Factory.
     * @return a set of Properties describing the beans that use the
     * given Mail Factory.
     */
    public Set getMailFactoryDependence(String mfName) {
        // the true param corresponds to a resource-ref
        return beansDependence(Collections.enumeration(beanList.values()), mfName, true);
    }

    /**
     * Used by the above getXXXDependence() methods.
     * @param enum the bean factories managed by this container.
     * @param rName jndi name of a resource (datasource, JMS destination or
     * connection factory, mail factory).
     * @param isResRef true for resource-ref, false for resource-env-ref.
     * @return set of Properties, where each Properties provides the bean
     * name, the container's file-name and the type of the bean which uses
     * the resource.
     */
    private Set beansDependence(Enumeration enum, String rName, boolean isResRef) {
        HashSet result = new HashSet();
        BeanFactory bf = null;
        while (enum.hasMoreElements()) {
            // for each bean,
            bf = (BeanFactory) enum.nextElement();
            BeanDesc ejbDesc = bf.getDeploymentDescriptor();

            boolean isDependent = false; // suppose its not dependent

            // get its type from the DeploymentDescriptor
            String ejbType = null;
            if (bf instanceof JEntityFactory) {
                if (ejbDesc instanceof EntityBmpDesc) {
                    ejbType = "ejbbmp";
                } 
                if (ejbDesc instanceof EntityCmpDesc) {
                    ejbType = "ejbcmp";
                } 
            } else if (bf instanceof JStatefulFactory) {
                ejbType = "ejbsbf";
            } else if (bf instanceof JStatelessFactory) {
                ejbType = "ejbsbl";
            } else if (bf instanceof JMdbFactory) {
                ejbType = "ejbmdb";
            }

            if (isResRef) {
                // if this is a cmp bean, directly compare rName with the data source JNDI name
                // from its deployment descriptor
                if (ejbType.equals("ejbcmp")) {
                    String jndiName = null;
                    if (ejbDesc instanceof EntityJdbcCmp1Desc) {
                        jndiName = ((EntityJdbcCmp1Desc) ejbDesc).getDatasourceJndiName();
                    } else if (ejbDesc instanceof EntityJdbcCmp2Desc) {
                        jndiName = ((EntityJdbcCmp2Desc) ejbDesc).getDatasourceJndiName();
                    }
                    isDependent = rName.equals(jndiName);
                }
                // get the description of the bean's resource-refs
                ResourceRefDesc[] rrDesc = ejbDesc.getResourceRefDesc();
                for (int i = 0; i < rrDesc.length; i++) {
                    // for each resource-ref compare its name with rName
                    if (rrDesc[i].getJndiName().equals(rName)) {
                        isDependent = true;
                        break;
                    }
                }
            } else {
                if (ejbType.equals("ejbmdb")) {
                    if (rName.equals(((MessageDrivenDesc) ejbDesc).getDestinationJndiName())) {
                        isDependent = true;
                    }
                }
                // get the description of the bean's  resource-env-refs
                ResourceEnvRefDesc[] rerDesc = ejbDesc.getResourceEnvRefDesc();
                for (int i = 0; i < rerDesc.length; i++) {
                    // for each resource-env-ref compare its name with rName
                    if (rerDesc[i].getJndiName().equals(rName)) {
                        isDependent = true;
                        break;
                    }
                }
            }

            if (isDependent) {
                Properties toAdd = new Properties();
                toAdd.setProperty("type", ejbType);
                toAdd.setProperty("fname", getFileName());
                toAdd.setProperty("name", ejbDesc.getEjbName());
                result.add(toAdd);
            }

        }
        return result;
    }

    // -------------------------------------------------------------------
    // Other public methods, used internally in this package.
    // -------------------------------------------------------------------

    /**
     * register an Entity that must be passivated
     * @param ejb The Entity Switch to be registered
     */
    public void registerEJB(JEntitySwitch ejb) {
        TraceEjb.interp.log(BasicLevel.DEBUG, myname);

        swapper.addEntity(ejb);
    }

    /**
     * register a BeanFactory
     * @param bf The Bean Factory to be registered
     */
    public void registerBF(BeanFactory bf) {
        TraceEjb.interp.log(BasicLevel.DEBUG, myname);
        swapper.addBeanFactory(bf);
    }

    /**
     * Take into account the swapping time for the bean.
     * @param t time in seconds  (t = 0 no time out)
     */
    public void setSwapTime(int t) {
        TraceEjb.interp.log(BasicLevel.DEBUG, myname);
        if (t > 0) {
            swapper.setTimeout(1000 * t);
        }
    }

    /**
     * Set the bean environment
     * @param ctx Context for this bean
     * @param dd Bean Deployment Descriptor
     * @throws NamingException if could not rebind objects
     */
    public void setBeanEnvironment(Context ctx, BeanDesc dd) throws NamingException {

        // Set bean context
        // Build comp/env
        Context ctxold = naming.setComponentContext(ctx);
        Context envCtx = ctx.createSubcontext("comp/env");

        // Bean Environment
        EnvEntryDesc [] envt = dd.getEnvEntryDesc();
        for (int i = 0; i < envt.length; i++) {
            // get information in descriptor
            String name = envt[i].getName();
            Object obj = envt[i].getValue();
            // register object in JNDI
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": Binding object " + name + " -> " + obj);
            }
            envCtx.rebind(name, obj);
        }
        
        // Resource References
        ResourceRefDesc [] resref = dd.getResourceRefDesc();
        for (int i = 0; i < resref.length; i++) {
            // get information in descriptor
            String name = resref[i].getName();
            String resname = resref[i].getJndiName();
            // build the LinkRef that will be registered:
            // FactoryClassName = null, size = 1, refAddr = resname.
            LinkRef lref = new LinkRef(resname);
            
            // if we are using David add a parameter to the link to tell that the
            // object must be found in rmiregistry (and not in CosNaming)
            // this info is used by CompNamingContext.lookup()
            if (Env.isDavid()) {
                lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
            }
            
            // register object in JNDI
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": Linking resource " + name + " -> " + resname);
            }
            envCtx.rebind(name, lref);
        }
        
        // Resource Environment References
        ResourceEnvRefDesc [] resEnvref = dd.getResourceEnvRefDesc();
        for (int i = 0; i < resEnvref.length; i++) {
            // get information in descriptor
            String name = resEnvref[i].getName();
            String resname = resEnvref[i].getJndiName();
            LinkRef lref = new LinkRef(resname);

            // if we are using David add a parameter to the link to tell that the
            // object must be found in rmiregistry (and not in CosNaming)
            // this info is used by CompNamingContext.lookup()
            if (Env.isDavid()) {
                lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
            }

            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": Linking resource environment " + name
                                    + " -> " + resname);
            }
            envCtx.rebind(name, lref);
        }

        // EJB References
        EjbRefDesc [] ejbref = dd.getEjbRefDesc();
        for (int i = 0; i < ejbref.length; i++) {
            // get information in descriptor
            String name = ejbref[i].getName();
            String ejbname = null;
            String ejblink = ejbref[i].getEjbLink();
            ejbname = ejbref[i].getJndiName();
            LinkRef lref = new LinkRef(ejbname);

            // if we are using David add a parameter to the link to tell that the
            // object must be found in the CosNaming (and not in rmiregistry)
            // this info is used by CompNamingContext.lookup()
            if (Env.isDavid()) {
                lref.add(new StringRefAddr("TargetNaming", "CosNaming"));
            }

            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": Linking ejb " + name + " -> " + ejbname);
            }
            envCtx.rebind(name, lref);
        }

        // EJB Local Refs
        // We use here ejb-link tag. This should be used also for ejb-ref when
        // we are able to manage references to another jar file.
        EjbLocalRefDesc [] ejblocalref = dd.getEjbLocalRefDesc();
        for (int i = 0; i < ejblocalref.length; i++) {
            String name = ejblocalref[i].getName();
            String ejblink = ejblocalref[i].getEjbLink();
            String ejbname = ejblocalref[i].getJndiLocalName();
            LinkRef lref = new LinkRef(ejbname);
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, myname + ": Linking ejb " + name + " -> " + ejbname);
            }
            envCtx.rebind(name, lref);
        }

        // Reset bean context
        naming.setComponentContext(ctxold);
    }

    /**
     * Check Security. No control for Message Driven Beans
     * @param methodSignature The method security signature
     */
    protected void checkSecurity(String methodSignature) {
        // A "null" string means : no security control.
        if (methodSignature.equals("null")) {
            return;
        }
        // methodGuard == null : no security service available
        if (methodGuard == null) {
            TraceEjb.security.log(BasicLevel.DEBUG, "no security service available");
            return;
        }
        TraceEjb.security.log(BasicLevel.DEBUG, "");
        try {
            methodGuard.checkGuard(methodSignature);
        } catch (SecurityException e) {
            throw new AccessLocalException("Access Denied");
        }
    }
}
