/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.ListIterator;
import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.naming.Context;
import javax.naming.NamingException;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for an Entity Bean. It is responsible for
 * - managing Home and LocalHome.
 * - managing a pool of instances/contexts
 * - keeping the list of PKs and the associated JEntitySwitch's
 * - keeping the JNDI context for this component (java:comp/env)
 * @author Philippe Coq, Philippe Durieux
 */
public class JEntityFactory extends JFactory {

    protected JEntityHome home = null;
    protected JEntityLocalHome localhome = null;
    protected boolean reentrant = false;
    protected boolean shared = false;

    // instance pool management (list of available JEntityContext objects)
    protected List bctxlist = Collections.synchronizedList(new ArrayList());

    // initial value for pool size
    protected int minPoolSize;

    protected int instanceCount = 0;
    protected int maxCacheSize;

    // PKs management
    // At each PK is associated a JEntitySwitch object
    protected Hashtable pklist = new Hashtable();

    // Datasource in case of CMP
    protected Object datasource = null;

    /**
     * This field is a pool of JEntitySwitch instance unused.
     */
    protected ArrayList entitySwitchUnused = null;

    /**
     * This field is used to generate an uid for unique automatic pk 
     */   
    private static int count = 1; 

    /**
     * constructor
     * must be without parameters (required by Jorm)
     */
    public JEntityFactory() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        entitySwitchUnused = new ArrayList();
    }

    /**
     * Init this object
     * @param dd the deployment descriptor
     * @param cont the Container
     */
    public void init(EntityDesc dd, JContainer cont) {
        super.init(dd, cont);
        reentrant = dd.isReentrant();
        shared = dd.isShared();
        maxCacheSize = dd.getCacheMax();
        minPoolSize = dd.getPoolMin();

        // Finds the DataSource associated to the bean
        // Useful for Entity with Container Managed Persistence.
        if (dd instanceof EntityJdbcCmp1Desc) {
            String dsname = ((EntityJdbcCmp1Desc) dd).getDatasourceJndiName();
            if (dsname != null) {
                try {
                    datasource = getInitialContext().lookup(dsname);
                } catch (NamingException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, dsname + " is not known by the EJB server.", e);
                    throw new EJBException(dsname + " unknown");
                }
            }
        }

        if (dd instanceof EntityJdbcCmp2Desc) {
            String dsname = ((EntityJdbcCmp2Desc) dd).getDatasourceJndiName();
            if (dsname != null) {
                try {
                    datasource = getInitialContext().lookup(dsname);
                } catch (NamingException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, dsname + " is not known by the EJB server.");
                    throw new EJBException(dsname + " unknown");
                }
            }
        }

        if (dd instanceof EntityCmp2Desc) {
            String cn = ((EntityCmp2Desc) dd).getJormAccessorClassName();
            if (!isClassAvailable(cn, true)) {
                throw new EJBException("Container has detected duplicated class " + cn);
            }
        }

        // Create the Home if defined in DD
        Class homeclass = null;
        String clname = dd.getFullWrpHomeName();
        if (clname != null) {
            try {
                homeclass = cont.getClassLoader().loadClass(clname);
            } catch (ClassNotFoundException e) {
                throw new EJBException(ejbname + "Cannot load " + clname);
            }
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, ejbname + ": " + clname + " loaded");
            }
            try {
                // new JEntityHome(dd, this)
                int nbp = 2;
                Class ptype[] = new Class[nbp];
                Object pobj[] = new Object[nbp];
                ptype[0] = org.objectweb.jonas_ejb.deployment.api.EntityDesc.class;
                pobj[0] = (Object) dd;
                ptype[1] = org.objectweb.jonas_ejb.container.JEntityFactory.class;
                pobj[1] = (Object) this;
                home = (JEntityHome) homeclass.getConstructor(ptype).newInstance(pobj);
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot create home ", e);
            }
            // register it in JNDI
            try {
                home.register();
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot register home ", e);
            }
        }

        // Create the LocalHome if defined in DD
        clname = dd.getFullWrpLocalHomeName();
        if (clname != null) {
            try {
                homeclass = cont.getClassLoader().loadClass(clname);
            } catch (ClassNotFoundException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot load " + clname);
                throw new EJBException("Cannot load " + clname);
            }
            if (TraceEjb.isDebugIc()) {
                TraceEjb.interp.log(BasicLevel.DEBUG, ejbname + ": " + clname + " loaded");
            }
            try {
                // new JEntityLocalHome(dd, this)
                int nbp = 2;
                Class ptype[] = new Class[nbp];
                Object pobj[] = new Object[nbp];
                ptype[0] = org.objectweb.jonas_ejb.deployment.api.EntityDesc.class;
                pobj[0] = (Object) dd;
                ptype[1] = org.objectweb.jonas_ejb.container.JEntityFactory.class;
                pobj[1] = (Object) this;
                localhome = (JEntityLocalHome) homeclass.getConstructor(ptype).newInstance(pobj);
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot create localhome ", e);
            }
            // register it in JNDI
            try {
                localhome.register();
            } catch (Exception e) {
                throw new EJBException(ejbname + " Cannot register localhome ", e);
            }
        }

        // pre-allocate a set of JEntityContext (bean instances)
        for (int i = 0; i < minPoolSize; i++) {
            JEntityContext ctx = null;
            try {
                ctx = createNewInstance();
                bctxlist.add(ctx);
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot create new instance", e);
                throw new EJBException(ejbname + " Cannot init pool of instances ", e);
            }
        }
    }

    /**
     * get EJB by its PK
     * Creates if not exist yet.
     * @param pk The Primary Key Object
     * @return The JEntitySwitch matching the PK.
     */
    public synchronized JEntitySwitch getEJB(Object pk) {

        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        JEntitySwitch bs = (JEntitySwitch) pklist.get(pk);
        if (bs == null) {
            bs = getJEntitySwitch();
            bs.init(this, pk);
            pklist.put(pk, bs);
        }
        return bs;
    }

    /**
     * get EJB by its PK
     * @param pk The Primary Key Object
     * @return The JEntitySwitch matching the PK, or null if none exist.
     */
    public synchronized JEntitySwitch existEJB(Object pk) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        return (JEntitySwitch) pklist.get(pk);
    }

    /**
     * Bind a PK with a JEntitySwitch
     * @param pk The Primary Key Object
     * @param bs The JEntitySwitch
     */
    public synchronized void bindEJB(Object pk, JEntitySwitch bs) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        bs.init(this, pk);
        pklist.put(pk, bs);
    }

    /**
     * This method allocates a new JEntitySwitch. But no association has been
     * done between the primary key and the new JEntitySwitch. Therefore the
     * initialisation is not done.
     * @return The JEntitySwitch.
     */
    public JEntitySwitch getJEntitySwitch() {
        return new JEntitySwitch();
    }

    /**
     * remove an EJB by its PK
     * @param pk The Primary Key Object
     */
    public synchronized void removeEJB(Object pk) {

        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        pklist.remove(pk);
    }


    // ---------------------------------------------------------------
    // BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the Instance pool size for this Ejb
     */
    public int getPoolSize() {
        return bctxlist.size();
    }

    /**
     * stop this EJB.
     * Mainly unregister it in JNDI.
     */
    public void stop() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        try {
            if (home != null) {
                home.unregister();
            }
            if (localhome != null) {
                localhome.unregister();
            }
        } catch (NamingException e) {
        }
    }

    /**
     * synchronize bean instances if needed
     */
    public void sync() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        passivateAllIH(false);
    }

    /**
     * Reduce number of instances in memory
     */
    public void reduceCache() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");
        // passivate all instances that are not busy
        // count returned is the number of busy instances (not freed)
        // other instances should be added to the pool (bctxlist)
        int count = passivateAllIH(true);

        // compute the new size we want for the pool
        // (if more busy instances than min size, we want an empty pool)
        int poolsz = minPoolSize - count;
        if (poolsz < 0) {
            poolsz = 0;
        }

        // reduce the pool to this new value
        synchronized (bctxlist) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "try to reduce " + bctxlist.size() + " to " + poolsz);
            while (bctxlist.size() > poolsz) {
                ListIterator i = bctxlist.listIterator();
                if (i.hasNext()) {
                    i.next();
                    i.remove();
                    instanceCount--;	// here ???
                }
            }
        }
        TraceEjb.swapper.log(BasicLevel.DEBUG, "cacheSize= " + getCacheSize());
    }

    /**
     * @return the home if it exists
     */
    public JHome getHome() {
        return home;
    }

    /**
     * @return the local home if it exists
     */
    public JLocalHome getLocalHome() {
        return localhome;
    }

    // ---------------------------------------------------------------
    // Context Pool
    // ---------------------------------------------------------------

    /**
     * Get a Context from the pool, or create a new one if no more
     * available in the pool.
     * This JContext must be initialized then by the caller.
     * @return a JEntityContext, not initialized.
     */
    public JEntityContext getJContext() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JEntityContext bctx = null;

        // try to find a free context in the pool
        synchronized (bctxlist) {
            try {
                ListIterator i = bctxlist.listIterator();
                if (i.hasNext()) {
                    bctx = (JEntityContext) i.next();
                    i.remove();
                }
            } catch (IndexOutOfBoundsException ex) {
                // pool is empty
            }
        }

        if (bctx == null) {
            // create a new one if pool empty
            try {
                bctx = createNewInstance();
            } catch (Exception e) {
                throw new EJBException("Cannot create a new instance: " + e.toString());
            }
        }
        return bctx;
    }


    /**
     * Release a Context
     * @param ctx - The JContext to release.
     */
    public void releaseJContext(JContext ctx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JEntityContext bctx = (JEntityContext) ctx;
        bctx.razEntityContext();
        bctxlist.add(bctx);
    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    /**
     * @return min pool size
     * for Jmx
     */
    public int getMinPoolSize() {
        return minPoolSize;
    }

    /**
     * @return max cache size
     * for Jmx
     */
    public int getMaxCacheSize() {
        return maxCacheSize;
    }

    /**
     * @return current cache size
     * for Jmx
     */
    public int getCacheSize() {
        return instanceCount;
    }

    /**
     * @return true if reentrant
     */
    public boolean isReentrant() {
        return reentrant;
    }

    /**
     * @return true if shared
     */
    public boolean isShared() {
        return shared;
    }

    /**
     * @return the Datasource used for CMP
     */
    public Object getDataSource() {
        if (datasource == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, "datasource not defined in JEntityFactory");
        }
        return datasource;
    }

    /**
     * Check Transaction before calling a method on a bean.
     * For Entity beans, the only possible case is "Container Managed Tx"
     * @param rctx The Request Context
     */
    public void checkTransaction(RequestCtx rctx) {
        checkTransactionContainer(rctx);
    }

    /**
     * Take a dump of current entity counters and return them
     * @return EntityCounters
     */
    public synchronized EntityCounters getEntityCounters() {
        EntityCounters ec = new EntityCounters();
        Enumeration enum = pklist.elements();
        while (enum.hasMoreElements()) {
            JEntitySwitch es = (JEntitySwitch) enum.nextElement();
            switch (es.getState()) {
            case 0:
                ec.inTx++;
                break;
            case 1:
                ec.outTx++;
                break;
            case 2:
                ec.idle++;
                break;
            case 3:
                ec.passive++;
                break;
            case 4:
                ec.removed++;
                break;
            }
        }
        return ec;
    }


    /** 
     * Calculate a new uid for automatic pk creation
     * Used by JEntityCmpJdbc.vm
     * @return int (unique pk)
     */
    public int calculateAutomaticPk() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        int uid;
        synchronized (getClass()) {
            uid = (int) (System.currentTimeMillis() % (java.lang.Integer.MAX_VALUE / 10));
            uid = uid * 10 + count++;
        }
        return uid;

    }

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * Synchronizes all instances that are used outside a transaction.
     * @param passivation true if passivation must occur
     * @return int count of instances that where busy.
     */
    private int passivateAllIH(boolean passivation) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        int busy = 0;

        // Set the JNDI component context (for ejbStore)
        Context bnctx = setComponentContext();

        Enumeration enum = null;
        synchronized (this) {
            enum = pklist.elements();
        }

        JEntitySwitch bs = null;
        while (enum.hasMoreElements()) {
            bs = (JEntitySwitch) enum.nextElement();
            if (!bs.passivateIH(passivation)) {
                busy++;
            }
        }

        // Reset old value for component context
        resetComponentContext(bnctx);

        return busy;
    }

    /**
     * Create a new instance of the bean and its EntityContext
     * In case of CMP, the bean class is derived to manage entity persistence.
     * @return JEntityContext
     * @throws InstantiationException cannot instantiate bean
     * @throws IllegalAccessException cannot instantiate bean
     */
    protected JEntityContext createNewInstance() throws Exception  {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        // create the bean instance
        EntityBean bean = null;
        try {
            bean = (EntityBean) beanclass.newInstance();
        } catch (InstantiationException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot instantiate bean");
            throw e;
        } catch (IllegalAccessException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname + " cannot instantiate bean");
            throw e;
        }

        // create a new EntityContext and bind it to the instance
        JEntityContext ec = new JEntityContext(this, bean);
        bean.setEntityContext(ec);

        // If too much instances, ask swapper to clean memory.
        instanceCount++;
        if (maxCacheSize != 0 && instanceCount > maxCacheSize) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "too much instances :" + instanceCount + ", max=" + maxCacheSize);
            cont.registerBF(this);
        }

        return ec;
    }

}
