/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import javax.ejb.EJBException;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * In the current version, we implement only the first. Others will be implemented later.
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public class JEntitySwitch {

    protected JEntityFactory bf;
    protected Object pk = null;
    protected JEntityLocal local = null;
    protected JEntityRemote remote = null;
    private JContainer cont;
    protected boolean shared;
    protected boolean reentrant;

    // nb of non transacted requests running
    protected int IHcount = 0;

    // unique EntityContext
    protected JEntityContext Icontext = null;

    // ident is used mainly for debugging
    private static int counter = 0;
    protected String ident;

    // nb of threads waiting for the Context
    private int waiters = 0;

    // True if ctx has been used outside transactions.
    private boolean dirty = false;

    // True if a TX need this instance currently used outside tx
    private boolean mustStore = false;

    // True if we must reload the bean state before use.
    private boolean mustLoad = false;

    // True when context/instance has been discarded (outside tx only)
    private boolean IHdiscarded = false;

    // current tx running on this instance and not yet committed.
    protected Transaction currtx = null;

    // max time to wait when in a transaction, before checking if
    // transaction has been set rollback only (deadlock detection)
    private static int maxtime = 5000; // 5 sec.

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitch() {
    }

    /**
     * constructor.
     * A new object is build when a new PK is known in the container,
     * either when a new bean is created, or when a find occurs.
     * For create(), PK is not known yet when this object is build.
     * @param bf The Entity Factory
     * @param pk The Primary Key
     */
    public void init(JEntityFactory bf, Object pk) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        this.bf = bf;
        this.pk = pk;

        // keep these in local, to be more efficient.
        cont = bf.getContainer();
        shared = bf.isShared();
        reentrant = bf.isReentrant();

        // generate new ident (debug)
        ident = "<" + counter + ">";
        counter++;
        Icontext = null;
        IHcount = 0;
        waiters = 0;
        dirty = false;
        mustStore = false;
        mustLoad = false;
        IHdiscarded = false;
        currtx = null;

        // Create EJBObject if bean has a Remote Interface
        if (bf.getHome() != null) {
            try {
                remote = ((JEntityHome) bf.getHome()).createRemoteObject();
                remote.setEntitySwitch(this);
            } catch (RemoteException e) {
                throw new EJBException("cannot create Remote Object", e);
            }
        }

        // Create EJBLocalObject if bean has a Local Interface
        if (bf.getLocalHome() != null) {
            local = ((JEntityLocalHome) bf.getLocalHome()).createLocalObject();
            local.setEntitySwitch(this);
        }
    }

    /**
     * @return the underlaying EJBLocalObject
     */
    public JEntityLocal getLocal() {
        return local;
    }

    /**
     * @return the underlaying EJBObject
     */
    public JEntityRemote getRemote() {
        return remote;
    }

    /**
     * @return the Primary Key Object for this instance.
     */
    public Object getPrimaryKey() {
        if (pk == null) {
            throw new EJBException("No Primary Key yet");
        }
        return pk;
    }

    /**
     * bind a JEntityContext for a create method.
     * No need to synchronize since object was unknown before.
     * Do not activate the bean instance since ejbCreate has been done.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind
     */
    public synchronized void bindICtx(Transaction tx, JEntityContext bctx) {

        // In some rare cases, a context can already exist.
        // Only if it is marked removed : just discard it before.
        if (Icontext != null) {
            bf.releaseJContext(Icontext);
        }

        if (tx == null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IH create");
            Icontext = bctx;
            bctx.initEntityContext(this, tx);
        } else {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IT create");
            Icontext = bctx;
            bctx.initEntityContext(this, tx);
            if (currtx == null) {
                currtx = tx;
                try {
                    tx.registerSynchronization(bctx);
                } catch (RollbackException e) {
                    TraceEjb.logger.log(BasicLevel.WARN, ident + "IT create: transaction has been marked rollbackOnly");
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "IT create: ", e);
                    currtx = null; // not registered -> no tx (avoids deadlocks)
                    throw new IllegalStateException("cannot register synchro");
                } catch (IllegalStateException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "IT create: Transaction is in an illegal state");
                    currtx = null; // not registered -> no tx (avoids deadlocks)
                    throw e;
                }
            }
        }
    }

    /**
     * Try to bind a JEntityContext if none already bound.
     * Called by finder methods
     * @param tx - the Transaction object
     * @param bctx The Entity Context
     */
    public synchronized void tryBindICtx(Transaction tx, JEntityContext bctx) {

        // Don't bind if already a context.
        if (Icontext != null) {
            bf.releaseJContext(bctx);
            return;
        }

        // No synchronization, since this is used for readonly methods (finder)
        if (tx == null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IH find");
            Icontext = bctx;
            bctx.initEntityContext(this, tx);
            bctx.activate(true);
        } else {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IT find");
            Icontext = bctx;
            bctx.initEntityContext(this, tx);
            bctx.activate(true);
            if (currtx == null) {
                currtx = tx;
                try {
                    tx.registerSynchronization(bctx);
                } catch (RollbackException e) {
                    TraceEjb.logger.log(BasicLevel.WARN, ident + "bindICtx IT: transaction has been marked rollbackOnly");
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "bindICtx IT: ", e);
                    currtx = null;
                    throw new IllegalStateException("cannot register synchro");
                } catch (IllegalStateException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "bindICtx IT: Transaction is in an illegal state");
                    currtx = null; // not registered -> no tx (avoids deadlocks)
                    throw e;
                }
            }
        }
    }

    /**
     * bind a JEntityContext for a remove method.
     * called in case of remove(pk) or remove(handle)
     * @param tx - the Transaction object
     * @param newctx - the JEntityContext to bind
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx, JEntityContext newctx) {
        return mapICtx(tx, newctx, false, false);
    }

    /**
     * Get a context/instance associated with this transaction
     * Called at each request on the bean (including remove)
     * @param tx - the Transaction object
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx) {
        return mapICtx(tx, null, false, true);
    }

    /**
     * Map a context and its instance. 
     * This version manages only 1 context (Icontext).
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit) {

        // Synchronization.
        if (tx == null) {
            // ------------- IH begin -------------
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH");

            // Must wait in case of TX or if non reentrant and already busy
            while (currtx != null || (reentrant && IHcount > 0)) {
                if (currtx != null) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: WAIT end IT");
                } else {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: WAIT end IH");
                }
                waiters++;
                try {
                    wait(maxtime);
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: NOTIFIED");
                } catch (InterruptedException e) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                }
            }
        } else {
            // ------------- IT begin -------------
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT");

            // Must wait in case of TX or if instance has been modified outside transaction.
            while (dirty || (currtx != null && !tx.equals(currtx))) {
                if (currtx != null) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: WAIT end IT");
                } else {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: WAIT end IH");
                }
                // instance is dirty : must write it before working on it again.
                // this is mandatory to be able to retrieve a good state in case of rollback.
                if (dirty) {
                    if (IHcount == 0) {
                        // Cannot passivate here, because must be done outside tx context
                        cont.registerEJB(this);
                    } else {
                        mustStore = true;
                    }
                }
                waiters++;
                try {
                    wait(maxtime);
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: NOTIFIED");
                } catch (InterruptedException e) {
                    TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                }
                // If transaction has been rolledback or set rollback only, give up.
                int status = Status.STATUS_ROLLEDBACK;
                try {
                    status = tx.getStatus();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "getICtx IT: unexpected exception getting transaction status");
                }
                switch (status) {
                case Status.STATUS_MARKED_ROLLBACK:
                case Status.STATUS_ROLLEDBACK:
                case Status.STATUS_ROLLING_BACK:
                    TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                    throw new TransactionRolledbackLocalException("rollback occured while waiting");
                }
            }
        }

        // Choose the context to use.
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (Icontext != null) {
                bf.releaseJContext(Icontext);
            }
            Icontext = bctx;
            Icontext.initEntityContext(this, tx);
        } else {
            if (Icontext != null) {
                // Reuse the Icontext for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    bf.releaseJContext(bctx);
                }
                Icontext.reuseEntityContext(tx);
                if (mustLoad) {
                    Icontext.activate(false);
                    mustLoad = false;
                }
            } else {
                if (bctx != null) {
                    Icontext = bctx;
                } else {
                    // no Context available : get one from the pool.
                    Icontext = (JEntityContext) bf.getJContext();
                }
                Icontext.initEntityContext(this, tx);
                Icontext.activate(true);
            }
        }

        if (tx != null) {
            if (currtx == null) {
                // New transaction on this instance.
                // Register this Context to release instance at commit
                TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IT: new tx, registerSynchronization");
                try {
                    tx.registerSynchronization(Icontext);
                } catch (RollbackException e) {
                    TraceEjb.logger.log(BasicLevel.WARN, ident + " mapICtx IT: transaction has been marked rollbackOnly");
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident, "mapICtx IT: ", e);
                    throw new IllegalStateException("mapICtx : cannot register synchro");
                } catch (IllegalStateException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "mapICtx IT: Transaction is in an illegal state");
                    throw e;
                }
                currtx = tx; // not if an exception is raised! (possible deadlock)
            }
        } else {
            if (holdit) {
                IHcount++;
                TraceEjb.interp.log(BasicLevel.DEBUG, ident + "mapICtx IH count=" + IHcount);
            }
            dirty = true;
        }

        return Icontext;
    }

    /**
     * Release a context/instance at end of request.
     * @param tx - transaction associated to this context
     */
    public synchronized void releaseICtx(Transaction tx) {

        JEntityContext bctx = null;

        if (tx == null) {
            // ------------- IH end -------------
            bctx = Icontext;
            IHcount--;
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "releaseICtx IH count=" + IHcount);
            if (IHcount == 0) {
                if (bctx.isMarkedRemoved()) {
                    bf.releaseJContext(bctx);
                    Icontext = null;
                    noLongerUsed();
                } else if (mustStore) {
                    try {
                        bctx.storeIfModified();
                    } catch (EJBException e) {
                        TraceEjb.logger.log(BasicLevel.WARN, ident + " ejbexception:" + e);
                    }
                    mustStore = false;
                    dirty = false;
                }
                if (waiters > 0) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                    notifyAll();
                }
            }
        } else {
            // ------------- IT end -------------
            // nothing to do now. Wait for tx completed.
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "releaseICtx IT");

        }
    }

    /**
     * Discard a context/instance at end of request. A problem occured
     * on this insance and it must be discarded.
     * @param tx - transaction associated to this context
     */
    public synchronized void discardICtx(Transaction tx) {
        if (tx == null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "discardICtx IH");
            JEntityContext bctx = Icontext;
            IHcount--;
            if (IHcount == 0) {
                bf.releaseJContext(bctx);
                Icontext = null;
                IHdiscarded = false;
                dirty = false;
            } else {
                IHdiscarded = true;
            }
        } else {
            // nothing to do now. Wait for tx completed.
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "discardICtx IT");
        }
    }

    /**
     * This transaction is now over. We can dispose of the instance
     * for another transaction or discard it.
     * @param tx the transaction object
     * @param committed true if transaction was committed.
     */
    public synchronized void txCompleted(Transaction tx, boolean committed) {
        JEntityContext bctx = Icontext;
        currtx = null;

        TraceEjb.interp.log(BasicLevel.DEBUG, ident + "txCompleted " + committed);

        if (waiters > 0) {
            // Somebody needs this instance : don't release it.
            // Just mark it to be reloaded if needed and notify waiters.
            if (shared || !committed) {
                mustLoad = true;
            }
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        } else {
            if (bctx.isMarkedRemoved()) {
                bf.releaseJContext(bctx);
                Icontext = null;
                IHdiscarded = false; // ?
                dirty = false; // ?
                if (committed) {
                    noLongerUsed(); // remove is effective !
                } else {
                    // remove operation has been rolled back
                    // we must clean memory instance now
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " remove has been rolled back");
                }
            } else {
                // If bean is shared : passivate and release this instance
                // because it can be used outside this server.
                // If bean is not shared : keep a ready version in cache, until
                // the swapper decide that there are too many instances in memory.
                if (shared || !committed) {
                    bctx.passivate();
                    bf.releaseJContext(bctx);
                    Icontext = null;
                    IHdiscarded = false;
                    dirty = false;
                }
            }
        }
    }

    /**
     * try to passivate IH (called from swapper)
     * @param passivation false if only store bean state.
     * @return true if OK, false if busy.
     */
    public synchronized boolean passivateIH(boolean passivation) {

        JEntityContext bctx = Icontext;

        // If no context mapped, nothing to do
        if (bctx == null) {
            return true;
        }
        // If the instance is busy, don't touch it.
        if (IHcount > 0 || currtx != null) {
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "busy");
            TraceEjb.interp.log(BasicLevel.DEBUG, ident + "IHcount=" + IHcount + " currtx=" + currtx + " waiters=" + waiters);
            return false;
        }
        TraceEjb.interp.log(BasicLevel.DEBUG, ident);
        // Store the instance state
        // -> only if instance not discarded
        if (!IHdiscarded) {
            try {
                bctx.storeIfModified();
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident, "error while storing bean state:", e);
            }
            mustStore = false;
            dirty = false;
        }

        // passivate this instance if required.
        if (passivation) {
            bctx.passivate();
            bf.releaseJContext(bctx);
            Icontext = null;
            IHdiscarded = false;
            dirty = false;
        }
        if (waiters > 0) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        }
        return true;
    }

    /**
     * Remove this EJB and
     * UnExport the remote object.
     */
    private void noLongerUsed() {
        TraceEjb.interp.log(BasicLevel.DEBUG, ident + "");
        if (remote != null) {
            try {
                remote.unexportObject();
            } catch (NoSuchObjectException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident + " unexport entity failed: " + e);
            }
        }
        bf.removeEJB(getPrimaryKey());
    }

    /**
     * @return ejb instance's state
     * state values are 0=in-tx, 1=out-tx, 2=idle, 3=passive, 4=removed.
     * we don't synchronize this method to avoid jadmin blocks
     */
    public int getState() {
        if (Icontext == null) {
            return 3;
        }
        if (currtx != null) {
            return 0;
        }
        if (Icontext.isMarkedRemoved()) {
            return 4;
        }
        if (dirty) {
            return 1;
        } else {
            return 2;
        }
    }

}
