/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.IOException;
import java.lang.reflect.Field;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.net.URLClassLoader;
import javax.ejb.EJBException;
import javax.ejb.TransactionRequiredLocalException;
import javax.ejb.TransactionRolledbackLocalException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import org.objectweb.jonas_lib.naming.ContainerNaming;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.lib.Version;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for beans.
 * @author Philippe Durieux
 */
public abstract class JFactory implements BeanFactory {

    protected JContainer cont;
    protected ContainerNaming naming = null;
    protected TransactionManager tm = null;
    protected Context JNDICtx = null; // JNDI Context for this component
    protected String ejbname = null;
    protected BeanDesc dd;
    protected Properties ejb10Env = null;
    protected boolean txbeanmanaged = false; // always false if entity bean
    protected Class beanclass = null; // class of the bean instance

    /**
     * constructor (for entity)
     * must be without parameters (required by Jorm)
     */
    public JFactory() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
    }

    /**
     * constructor (for session)
     * @param dd The bean deployment descriptor
     * @param cont the container for this bean
     */
    public JFactory(BeanDesc dd, JContainer cont) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        init(dd, cont);
    }

    /**
     * Init this object
     * @param dd the deployment descriptor
     * @param cont the Container
     */
    public void init(BeanDesc dd, JContainer cont) {
        this.cont = cont;
        this.dd = dd;
        naming = cont.getContainerNaming();
        tm = cont.getTransactionManager();
        ejbname = dd.getEjbName();
        ejb10Env = dd.getEjb10Environment();

        // Check the JOnAS version used for deployment in case of Session or Entity
        String cln = dd.getFullWrpHomeName() != null
            ? dd.getFullWrpHomeName()
            : dd.getFullWrpLocalHomeName();
        if (cln != null) {
            checkJonasVersion(cln);
        }

        // Create a JNDI context for this bean component
        try {
            JNDICtx = naming.createEnvironmentContext(ejbname);
            cont.setBeanEnvironment(JNDICtx, dd);
        } catch (NamingException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "cannot build naming for this component", e);
            throw new EJBException("Cannot build naming for this component" + e);
        }

        // Get the class of the bean instance (for ejbCreate, ejbRemove, ...)
        String cn = null;
        try {
            cn = dd.getFullDerivedBeanName();
            beanclass = cont.getClassLoader().loadClass(cn);
            // prevent from generated class name collision in a J2EE application
            if (!isClassAvailable(cn, true)) {
                throw new EJBException("Container has detected duplicated class " + cn);
            }
        } catch (ClassNotFoundException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "failed to find class " + cn + e);
            throw new EJBException("Container failed to find class " + cn);
        }

    }


    /**
     * Assess availability of a class in a given class loader
     * @param className the name of the class without the .class extension
     * @param unique flag indicating if the class should be found only once in the classloader
     * @return True if class is available in the current class loader.
     */
    protected boolean isClassAvailable(String className, boolean unique) {
        className = className.replace('.', '/') + ".class";
        Enumeration e = null;
        try {
            e = ((URLClassLoader) cont.getClassLoader()).findResources(className);
        } catch (IOException e1) {
            return false;
        }
        if (e.hasMoreElements()) {
            e.nextElement();
            return unique ? !e.hasMoreElements() : true;
        } else {
            return false;
        }
    }

    // ---------------------------------------------------------------
    // BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the bean name
     */
    public String getEJBName() {
        return ejbname;
    }

    /**
     * @return the Instance pool size for this Ejb
     */
    public abstract int getPoolSize();

    /**
     * Get the Deployement descriptor of this Ejb
     * @return BeanDesc The bean deployment descriptor
     */
    public BeanDesc getDeploymentDescriptor() {
        return dd;
    }

    /**
     * @return the TransactionManager
     */
    public TransactionManager getTransactionManager() {
        return tm;
    }

    /**
     * @return the JContainer object
     */
    public JContainer getContainer() {
        return cont;
    }

    /**
     * @return A Hashtable containing the JNDI Environment
     */
    public Hashtable getEnv() {
        return naming.getEnv();
    }

    /**
     * @return the InitialContext
     */
    public InitialContext getInitialContext() {
        return naming.getInitialContext();
    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    private static InitialContext corbaIctx = null;

    /**
     * @return an InitialContext to access CosNaming
     * @throws NamingException cannot get the JNDI initial context
     */
    public static InitialContext getCorbaInitialContext() throws NamingException {
        if (corbaIctx == null) {
            Properties properties = new Properties();
            properties.put("java.naming.factory.initial", "com.sun.jndi.cosnaming.CNCtxFactory");
            String host = System.getProperty("david.CosNaming.default_host", "localhost");
            String port = System.getProperty("david.CosNaming.default_port", "14727");
            properties.put("java.naming.provider.url", "iiop://" + host + ":" + port);
            corbaIctx =  new InitialContext(properties);
        }
        return corbaIctx;
    }

    /**
     * @return the EJB 1.0 style environment associated with the Bean
     */
    public Properties getEjb10Environment() {
        return ejb10Env;
    }

    /**
     * @return true if transactions are managed inside the bean
     *  false if transactions are managed by the container
     */
    public boolean isTxBeanManaged() {
        return txbeanmanaged;
    }

    /**
     * set the Component Context for JNDI environment
     * @return previous Context
     */
    public Context setComponentContext() {
        Context oldctx = naming.setComponentContext(JNDICtx);
        if (oldctx == null) {
            TraceEjb.loaderlog.log(BasicLevel.DEBUG, "previous ctx was null");
        }
        return oldctx;
    }

    /**
     * reset old Component Context for JNDI environment
     * @param oldctx previous Component Context to restore.
     */
    public void resetComponentContext(Context oldctx) {
        if (oldctx == null) {
            TraceEjb.loaderlog.log(BasicLevel.DEBUG, "to null");
        }
        naming.setComponentContext(oldctx);
    }

    /**
     * Common preInvoke
     * @param txa Transaction Attribute (Supports, Required, ...)
     * @param secu Security String that uniquely identifies the method.
     * @return A RequestCtx object
     * @throws EJBException
     */
    public RequestCtx preInvoke(int txa, String secu) {

        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        RequestCtx rctx = null;

        try {
            // security controls (except for Message Driven Beans)
            if (secu.length() > 0) {
                cont.checkSecurity(secu);
            }

            // Build a RequestCtx to save information about this request
            // until the postInvoke time.
            rctx = new RequestCtx(txa);

            // Set classLoader to the correct value and save the previous one
            // in the RequestCtx to restore it at the end of the Request.
            rctx.cloader = Thread.currentThread().getContextClassLoader();
            Thread.currentThread().setContextClassLoader(myClassLoader());

            // Set bean context for JNDI and save the previous one in the
            // RequestCtx to restore it at the end of the Request.
            rctx.jndiCtx = setComponentContext();

            // Check Transaction conformance to bean settings
            checkTransaction(rctx);

        } catch (RuntimeException e) {
            // Log RuntimeException before rethrowing it
            // -> Print Stack Trace to see where the problem is.
            TraceEjb.logger.log(BasicLevel.ERROR, "unexpected Runtime Exception", e);
            throw e;
        }
        return rctx;
    }

    /**
     * Common postInvoke
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @throws EJBException
     */
    public void postInvoke(RequestCtx rctx) {

        TraceEjb.interp.log(BasicLevel.DEBUG, "");

        try {
            // Commit the transaction started in preInvoke
            if (rctx.mustCommit) {
                if (rctx.sysExc != null) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "system exception raised by request:", rctx.sysExc);
                    // rollback the transaction
                    try {
                        tm.rollback();
                    } catch (Exception e) {
                        TraceEjb.logger.log(BasicLevel.ERROR, "exception during rollback:", e);
                    }
                } else {
                    try {
                        // Check status to avoid returning RollbackException to the client
                        // instead of the application exception
                        switch (tm.getStatus()) {
                        case Status.STATUS_ACTIVE:
                            TraceEjb.interp.log(BasicLevel.DEBUG, "committing transaction");
                            tm.commit();
                            break;
                        case Status.STATUS_MARKED_ROLLBACK:
                            TraceEjb.interp.log(BasicLevel.DEBUG, "rolling back transaction");
                            tm.rollback();
                            break;
                        default:
                            TraceEjb.logger.log(BasicLevel.ERROR, "unexpected transaction status " + tm.getStatus());
                            break;
                        }
                    } catch (Exception e) {
                        throw new EJBException("Container exception", e);
                    }
                }
            }

            // Reset class loader to the original value
            Thread.currentThread().setContextClassLoader(rctx.cloader);

            // We got a system Exception in business method:
            // - log exception
            // - discard instance
            // - set client transaction rollback only
            // - throw EJBException
            if (rctx.sysExc != null) {
                // Log system exception
                TraceEjb.logger.log(BasicLevel.ERROR, "system exception in business method:", rctx.sysExc);

                // If client transaction: set it rollback only and throw TransactionRolledBackLocalException
                Transaction currentTx = null;
                try {
                    currentTx = tm.getTransaction();
                    if (currentTx != null) {
                        currentTx.setRollbackOnly();
                        throw new TransactionRolledbackLocalException("System exception in business method");
                    }
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot set rollback only current tx:", e);
                }
                return;
            }

            // Reset bean context for JNDI
            resetComponentContext(rctx.jndiCtx);

            // Resume client transaction suspended in preInvoke
            Transaction tx = rctx.clientTx;
            if (tx != null) {
                try {
                    tm.resume(tx);
                } catch (Exception e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot resume transaction");
                }
            }

        } catch (TransactionRolledbackLocalException e) {
            throw e;
        } catch (EJBException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "ejbexception: ", e);
            throw e;
        } catch (RuntimeException e) {
            // Log RuntimeException before rethrowing it
            // -> Print Stack Trace to see where the problem is.
            TraceEjb.logger.log(BasicLevel.ERROR, "unexpected runtime exception: ", e);
            throw e;
        }
    }

    /**
     * preInvoke for Remote access
     * @param txa Transaction Attribute (Supports, Required, ...)
     * @param secu Security String that uniquely identifies the method.
     * @return A RequestCtx object
     * @throws TransactionRequiredException
     * @throws TransactionRolledbackException
     * @throws NoSuchObjectException
     * @throws RemoteException preinvoke raised an EJBException
     */
    public RequestCtx preInvokeRemote(int txa, String secu) throws RemoteException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        try {
            return preInvoke(txa, secu);
        } catch (javax.ejb.TransactionRequiredLocalException e) {
            throw new javax.transaction.TransactionRequiredException(e.getMessage());
        } catch (javax.ejb.TransactionRolledbackLocalException e) {
            throw new javax.transaction.TransactionRolledbackException(e.getMessage());
        } catch (javax.ejb.NoSuchObjectLocalException e) {
            throw new java.rmi.NoSuchObjectException(e.getMessage());
        } catch (javax.ejb.EJBException e) {
            throw new java.rmi.RemoteException(e.getMessage());
        }
    }

    /**
     * postInvoke for Remote access
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @throws TransactionRequiredException
     * @throws TransactionRolledbackException
     * @throws NoSuchObjectException
     * @throws RemoteException postinvoke failed
     */
    public void postInvokeRemote(RequestCtx rctx) throws RemoteException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        try {
            postInvoke(rctx);
        } catch (javax.ejb.TransactionRequiredLocalException e) {
            throw new javax.transaction.TransactionRequiredException(e.getMessage());
        } catch (javax.ejb.TransactionRolledbackLocalException e) {
            throw new javax.transaction.TransactionRolledbackException(e.getMessage());
        } catch (javax.ejb.NoSuchObjectLocalException e) {
            throw new java.rmi.NoSuchObjectException(e.getMessage());
        } catch (javax.ejb.EJBException e) {
            throw new java.rmi.RemoteException(e.getMessage());
        }
    }

    /**
     * check Transaction attribute
     * @param rctx the Request Context
     */
    abstract void checkTransaction(RequestCtx rctx);

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * Process Transaction Attribute before calling a business method
     * @param rctx the Request Context
     * @throws EJBException
     * @throws TransactionRequiredLocalException
     */
    protected void checkTransactionContainer(RequestCtx rctx) {

        int txa = rctx.txAttr;
        if(TraceEjb.isDebugTx())
            TraceEjb.tx.log(BasicLevel.DEBUG, "transaction attribute = "+ 
                            MethodDesc.getTxAttributeName(txa));

        if (txa == MethodDesc.TX_NOT_SET) {
            // No check to do (for example: session home)
            return;
        }

        rctx.mustCommit = false;
        Transaction cltx = null;

        // First of all, get the current transaction
        Transaction currtx = null;
        try {
            currtx = tm.getTransaction();
        } catch (SystemException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "system exception while getting transaction:", e);
        }

        // Raises exception if "Never" and we are in a transaction
        if (txa == MethodDesc.TX_NEVER && currtx != null) {
            TraceEjb.logger.log(BasicLevel.WARN,"never and transaction not null");
            throw new EJBException("Never attribute = caller must not be in a transaction");
        }

        // Raises exception if "Mandatory" and we are not in a transaction
        if (txa == MethodDesc.TX_MANDATORY && currtx == null) {
            TraceEjb.logger.log(BasicLevel.WARN,"mandatory and not in transaction");
            throw new TransactionRequiredLocalException("Mandatory attribute = caller must be in a transaction");
        }

        // Suspend transaction if "NotSupported" or "RequiresNew"
        if (currtx != null &&
            (txa == MethodDesc.TX_REQUIRES_NEW || txa == MethodDesc.TX_NOT_SUPPORTED)) {
            try {
                cltx = tm.suspend();
                currtx = null;
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot suspend transaction:\n", e);
                throw new EJBException("Cannot suspend transaction", e);
            }
        }

        // Start a new transaction in 2 cases:
        // - "RequiresNew" attribute
        // - "Required" and no current transaction
        // this transaction will be closed at postInvoke.
        if (txa == MethodDesc.TX_REQUIRES_NEW ||
            (txa == MethodDesc.TX_REQUIRED && currtx == null)) {
            try {
                tm.begin();
                rctx.mustCommit = true;
                currtx = tm.getTransaction();
            } catch (NotSupportedException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot start a transaction: NotSupportedException");
                throw new EJBException("Nested Transactions Not Supported");
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot start a transaction:\n", e);
                throw new EJBException("Cannot start a transaction: SystemException");
            }
        }

        // update RequestCtx
        rctx.currTx = currtx;
        rctx.clientTx = cltx;
    }

    /**
     * Check if the given class have been generated by GenIC tool with a correct version.
     * Trace an error message, if not.
     * @param clName class name
     */
    protected void checkJonasVersion(String clName) {
        /*
         * Check if the static 'JONAS_VERSION' variable defined in the given class
         * (ie GenIC's Version used to deploy the bean),
         * has the same value as the JOnAS Version.NUMBER variable.
         */
        String fdName = "JONAS_VERSION";
        String gVersion = null;
        try {
            // The name of the resource ends with .class and the '.' are replace with '/'
            String resourceName = clName.replace('.','/') + ".class";

            // get the list of the resources for this class
            Enumeration e = null;
            try {
                e = cont.getClassLoader().getResources(resourceName);
            } catch (IOException ioe) {
                TraceEjb.logger.log(BasicLevel.ERROR, "failed to find class " + clName + ioe);
                throw new EJBException("Container failed to find class " + clName);
            }

            // Count the resources
            int nbCls = 0;
            String urls = "";
            while (e.hasMoreElements()) {
                nbCls++;
                urls += e.nextElement() + "\n";
            }

            // More than one resource.
            if (nbCls > 1) {
                TraceEjb.logger.log(BasicLevel.WARN, "there are " + nbCls + 
                                    " resources for the class " +
                                    clName + 
                                    ". Some problems can occur because it's the first resource which will be loaded. The list of resources is : \n" + urls);
            }

            // Get the class and get the value of the "JONAS_VERSION" static field
            Class cl = cont.getClassLoader().loadClass(clName);
            Field fd = cl.getDeclaredField("JONAS_VERSION");
            gVersion = (String)fd.get(null);
        } catch (ClassNotFoundException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "failed to find class " +clName+ e);
            return;
        } catch (NoSuchFieldException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "failed to find field "+fdName
                                +" of class "+clName+": "+e);
            return;
        } catch (IllegalAccessException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "failed to get the value of the field "+fdName
                                +" of class "+clName+": "+e);
            return;
        }
        // Compare the current JOnAS version to the GenIC's Version used to deploy the bean
        if (!Version.NUMBER.equals(gVersion)) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname+ " bean not deployed with the same JOnAS version. "
                                +"You have to redeploy "+cont.getFileName()+".");
        }
    }

    /**
     * @return the classloader for this container.
     */
    public ClassLoader myClassLoader() {
        return cont.getClassLoader();
    }
}
