/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import javax.ejb.EJBException;
import javax.ejb.SessionBean;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;
import org.objectweb.jonas_ejb.lib.JavaType;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for a Session Stateful Bean.
 * @author Philippe Durieux
 */
public class JStatefulFactory extends JSessionFactory {

    JStatefulPool statefulPool;

    /**
     * constructor
     * @param dd Bean Deployment Descriptor
     * @param cont Container where the bean is defined
     * @param sfp Pool des sessions stateful
     */
    public JStatefulFactory(SessionStatefulDesc dd, JContainer cont, JStatefulPool sfp) {
        super((SessionDesc)dd, cont);
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        isSynchro = javax.ejb.SessionSynchronization.class.isAssignableFrom(beanclass);
        isStateful = true;
        statefulPool = sfp;
    }

    // ---------------------------------------------------------------
    // BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the Instance pool size for this Ejb
     */
    public int getPoolSize() {
        return 0;		// TODO
    }

    // ---------------------------------------------------------------
    // preInvoke / postInvoke
    // ---------------------------------------------------------------

    /**
     * preInvoke for Session beans stateful
     * @param txa Transaction Attribute (Supports, Required, ...)
     * @param secu Security String that uniquely identifies the method.
     * @return A RequestCtx object
     * @throws EJBException
     */
    public RequestCtx preInvoke(int txa, String secu) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        RequestCtx rctx = super.preInvoke(txa, secu);
        return rctx;
    }

    /**
     * postinvoke
     * @param rctx The RequestCtx that was returned at preInvoke()
     * @throws EJBException
     */
    public void postInvoke(RequestCtx rctx) {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        super.postInvoke(rctx);
    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    /**
     * Creates a new Session Stateful
     * called back from createEJB
     * @return The Session Switch object
     */
    public JSessionSwitch createNewSession() throws RemoteException {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatefulSwitch bs = new JStatefulSwitch(this, statefulPool);
        return bs;
    }

    /**
     * get a new session context
     * must call newInstance (EJB specs) => no pool should be used.
     * @return a new Session Context
     */
    public JSessionContext getJContext() {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        JStatefulContext bctx = null;
        try {
            bctx = createNewInstance();
        } catch (Exception e) {
            throw new EJBException("Cannot create a new instance "+e);
        }
        return bctx;
    }

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * Create a new instance of the bean and its StatefulContext
     */
    private JStatefulContext createNewInstance() throws Exception {
        TraceEjb.interp.log(BasicLevel.DEBUG, "");
        // create the bean instance
        SessionBean bean = null;
        try {
            bean = (SessionBean) beanclass.newInstance();
        } catch (InstantiationException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname+" cannot instantiate session bean");
            throw e;
        } catch (IllegalAccessException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ejbname+" cannot instantiate session bean");
            throw e;
        }
        // create a new StatefulContext and bind it to the instance
        JStatefulContext bctx = new JStatefulContext(this, bean, isSynchro);
        bean.setSessionContext(bctx);
        return bctx;
    }

}
