/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Sebastien Chassandre Barrioz
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.deployment.api;

import java.util.Iterator;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.ArrayList;
import java.util.List;
import java.lang.reflect.Method;

import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.EjbJar;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelationName;
import org.objectweb.jonas_ejb.deployment.xml.Entity;
import org.objectweb.jonas_ejb.deployment.xml.JonasEjbJar;
import org.objectweb.jonas_ejb.deployment.xml.JonasEntity;
import org.objectweb.jonas_ejb.deployment.xml.JonasMessageDriven;
import org.objectweb.jonas_ejb.deployment.xml.MessageDriven;
import org.objectweb.jonas_ejb.genic.GenICException;
import org.objectweb.jonas_ejb.genic.BeanSources;
import org.objectweb.jonas_ejb.lib.MappingBuilder;
import org.objectweb.jonas_ejb.lib.JormType;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.jorm.metainfo.lib.BasicManager;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * Class to hold meta-information related to the deployment of an ejb-jar
 * with JORM based CMP2 persistence
 * @author S.Chassande-Barrioz
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class DeploymentCmp2Desc extends DeploymentDesc {

    /**
     * This constant defines the default mapping name
     */
    public final static String PROJECT_NAME = "jonas";

    /**
     * hold the JORM meta-information manager
     */
    private BasicManager manager = null;


    /**
     * Build the Meta-Information from the XML data binding trees
     * containing the EJB and JOnAS deployment descriptors.
     */
    public DeploymentCmp2Desc(ClassLoader classLoader, EjbJar ejbJar, JonasEjbJar jonasEjbJar, Logger l)
        throws DeploymentDescException {
        super(classLoader, ejbJar, jonasEjbJar, l);

        // HashMap of jonas-message-driven
        HashMap jonasMessageDriven = new HashMap();
        for (Iterator i = jonasEjbJar.getJonasMessageDrivenList().iterator(); i.hasNext(); ) {
            JonasMessageDriven jMd = (JonasMessageDriven)i.next();
            jonasMessageDriven.put(jMd.getEjbName().getValue(),jMd);
        }

        // message-driven beans
        for (Iterator i = ejbJar.getEnterpriseBeans().getMessageDrivenList().iterator(); i.hasNext(); ) {
            BeanDesc bd = null;
            MessageDriven md = (MessageDriven)i.next();
            /*
             * The Standard DTD allows a message-driven-bean without ejb-name
             * Problem to associate infos to this bean (trans-attribute, jonas-specific infos)
             * So the ejb-name must be specify and we check this
             */
            if (md.getEjbName() == null) {
                throw new DeploymentDescException("ejb-name missing for a message driven bean");
            }
            // find corresponding jonas message-driven
            JonasMessageDriven jMd = (JonasMessageDriven)jonasMessageDriven.get(md.getEjbName().getValue());
            if (jMd == null)
                throw new DeploymentDescException("jonas-message-driven-bean missing for bean "+md.getEjbName().getValue());
            bd = new MessageDrivenDesc(classLoader, md, asd, jMd);
            bd.check();
            bd.setDeploymentDesc(this);
            beanDesc.put(bd.getEjbName(), bd);
        }

        // Entity beans CMP2
        if (ejbJar.getRelationships() != null) {

            // ejb-relation
            for (Iterator i = ejbJar.getRelationships().getEjbRelationList().iterator(); i.hasNext();) {
                EjbRelationDesc erd = null;
                EjbRelation er = (EjbRelation) i.next();
                EjbRelationName ern = er.getEjbRelationName();
                if (ern == null) {
                    throw new DeploymentDescException("ejb-relation-name missing in ejb-relation "+er.getDescription());
                }
                String name = ern.getValue();
                if (name.length() == 0) {
                    throw new DeploymentDescException("ejb-relation-name value missing in jonas-ejb-jar.xml");
                }
                erd = new EjbRelationDesc(er);
                // ejb-relationship-role (source,target)
                EjbRelationshipRoleDesc rsd1 = erd.getRelationshipRoleDesc1();
                EjbRelationshipRoleDesc rsd2 = erd.getRelationshipRoleDesc2();
                String ejbName1 = rsd1.getEjbName();
                String ejbName2 = rsd2.getEjbName();
                EntityCmp2Desc ed1 = (EntityCmp2Desc) beanDesc.get(ejbName1);
                EntityCmp2Desc ed2 = (EntityCmp2Desc) beanDesc.get(ejbName2);
                ed1.addEjbRelationshipRoleDesc(rsd1);
                ed2.addEjbRelationshipRoleDesc(rsd2);
                rsd1.setLinkedEntityCmp2Desc(ed2);
                rsd2.setLinkedEntityCmp2Desc(ed1);
            }
        }
    }

    protected BeanDesc newEntityBeanDesc(ClassLoader classLoader,
                                         Entity ent,
                                         AssemblyDescriptor asd,
                                         JonasEntity jEnt) throws DeploymentDescException {
        // Default is CMP2.x for entity beans with a EJB2.0 DD.
        if (ent.getCmpVersion() == null || ent.getCmpVersion().getValue().equals("2.x")) {
            return new EntityJdbcCmp2Desc(classLoader, ent, asd, jEnt, this);
        } else {
            return new EntityJdbcCmp1Desc(classLoader, ent, asd, jEnt);
        }
    }

    /**
     * get the JORM Manager
     * At first call, loads all the meta information for entity beans CMP2
     */
    public Manager getManager() throws DeploymentDescException {
        if (manager == null) {
            manager = new BasicManager();
            manager.setLogger(logger);
            manager.init();
            manager.addMappingFactory("rdb", "org.objectweb.jorm.metainfo.rdb.BasicRDBMappingFactory");
            BeanDesc[] beansDD = getBeanDesc();
            for (int i = 0; i < beansDD.length; i++) {
                if (beansDD[i] instanceof EntityCmp2Desc) {
                    // Load the jorm meta information of the class
                    createFromBeanDesc((EntityCmp2Desc)beansDD[i]);
                }
            }
        }
        return manager;
    }

    /**
     * Create JORM MetaInfo from a JOnAS CMP2 Entity Bean Descriptor
     */
    private void createFromBeanDesc(EntityCmp2Desc bd) throws DeploymentDescException {

        String asn = bd.getAbstractSchemaName();
        logger.log(BasicLevel.DEBUG, "Load Jorm MI for the bean: " + asn);

        // Create a new Schema only if it does not exist yet.
        Schema schema = manager.createSchema(bd.getJormSchemaName());

        MappingBuilder mb = bd.getMappingBuilder();

        // create persistent class of name abstract schema name.
        // Be careful: the skeleton of class can be already defined when this
        // class is referenced by another class (scheduling problem)
        Class clazz = schema.getHiddenClass(asn);
        if (clazz == null) {
            clazz = schema.createClass(asn, false);
        } else {
            schema.moveHiddenToClass(asn);
        }
        mb.createClassMapping(clazz, bd, PROJECT_NAME);

        //----------------------------- CMP fields ---------------------------//
        //--------------------------------------------------------------------//
        for (Iterator i = bd.getCmpFieldDescIterator(); i.hasNext();) {
            FieldDesc fd = (FieldDesc) i.next();
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                           "Create PrimitiveElement for the CMP field: " + fd.getFieldName());
            }
            PType pt = JormType.getPType(fd.getFieldType());
            PrimitiveElement pe = clazz.createPrimitiveElement(fd.getFieldName(), pt);
            mb.createValueMapping(pe, fd, PROJECT_NAME);
        }

        NameDef[] nds = {clazz.createNameDef()};

        //--------------------------- Class NameDef --------------------------//
        //--------------------------------------------------------------------//
        for (int i = 0; i < nds.length; i++) {
            nds[i].setMappNames(PROJECT_NAME);
        }
        if (bd.hasPrimaryKeyField()) {
            // one primary key field
            String fieldName = bd.getPrimaryKeyFieldName();
            for (int i = 0; i < nds.length; i++) {
                nds[i].setFieldName(fieldName);
            }
        } else {
            // Define a Composite Name and a NameRef
            // The name of the composite is built with the abstract schema
            // name and the suffix "Pk"
            String cnName = asn + "Pk";
            CompositeName cn = schema.getCompositeName(cnName);
            if (cn == null) {
                cn = schema.createCompositeName(cnName);
            }
            // Used the new CompositeName into the NameDef
            NameRef[] nrs = new NameRef[nds.length];
            for (int i = 0; i < nds.length; i++) {
                nrs[i] = nds[i].createNameRef(cn.getName(), cn);
            }
            // fill the composite name with pk fields and link these fields to
            // to the class fields
            for (Iterator it = bd.getCmpFieldDescIterator(); it.hasNext();) {
                FieldDesc fd = (FieldDesc) it.next();
                if (fd.isPrimaryKey()) {
                    String fieldName = fd.getFieldName();
                    if (cn.getScalarField(fieldName) == null)
                        cn.createCompositeNameField(fieldName, JormType.getPType(fd.getFieldType()));
                    for (int i = 0; i < nds.length; i++) {
                        nrs[i].addProjection(fieldName, fieldName);
                    }
                }
            }
        }

        //--------------------------- CMR fields -----------------------------//
        //--------------------------------------------------------------------//
        for (Iterator i = bd.getEjbRelationshipRoleDescIterator(); i.hasNext();) {
            EjbRelationshipRoleDesc rsrd = (EjbRelationshipRoleDesc) i.next();
            if (rsrd.hasCmrField()) {
                EntityCmp2Desc bd2 = rsrd.getLinkedEntityCmp2Desc();
                if (rsrd.isMultiValuedCMR()) {
                    createGenClassReference(bd, clazz, rsrd, bd2, mb);
                } else {
                    createClassReference(clazz, rsrd, bd2, mb, logger);
                }
            }
        }
    }

    /**
     * This method create the jorm meta information which matches with a
     * relation between one bean and several others.
     * @param bd is the descriptor of the first bean
     * @param clazz is the class (jorm meta object) of the first bean.
     * @param fd is the descriptor of the relation
     * @param bd2 is the descriptor of the second bean
     */
    private void createGenClassReference(EntityCmp2Desc bd,
                                         Class clazz,
                                         EjbRelationshipRoleDesc rsrd,
                                         EntityCmp2Desc bd2,
                                         MappingBuilder mb)
        throws DeploymentDescException {
        String cmrFieldName = rsrd.getCmrFieldName();
        if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                       "Create GenClassReference for the CMR field: " + cmrFieldName);

        Schema schema = (Schema) clazz.getParent();

        // create the gen class
        GenClassRef gcr = clazz.createGenClassRef(cmrFieldName, rsrd.getCmrFieldType().getName());

        mb.createClassMapping(gcr, rsrd, PROJECT_NAME);

        // is the nameDef which desribes how the current bean
        // references the gen class
        // = the identifier of the source class (bd)
        logger.log(BasicLevel.DEBUG, "Fill the name def of gen class reference");
        fillNameDef(gcr.createNameDef(), schema, rsrd,
                    EjbRelationshipRoleDesc.SOURCE, bd, mb, cmrFieldName, false, clazz, logger);


        // Used for the gen class identifier
        String baseFieldName1 = bd.getAbstractSchemaName();
        // Used for the gen class element identifier
        String baseFieldName2 = bd2.getAbstractSchemaName();
        if (bd == bd2) {
            // The source class is also the target class. Then the index of
            // the role is added at the of the base name in order to
            // differenciate the fields.
            if (rsrd == rsrd.getEjbRelationDesc().getRelationshipRoleDesc1()) {
                baseFieldName1 += "_" + 1;
                baseFieldName2 += "_" + 2;
            }
            else {
                baseFieldName1 += "_" + 2;
                baseFieldName2 += "_" + 1;
            }
        }

        // is the namedef which describes how the gen class is
        // identified (the gen class name)
        // = the identifier of the source class (bd)
        logger.log(BasicLevel.DEBUG, "Fill the name def of gen class identifier");
        fillNameDef(gcr.createRefNameDef(), schema, rsrd,
                    EjbRelationshipRoleDesc.SOURCE, bd, mb, baseFieldName1, true, gcr, logger);

        Class clazz2 = schema.getClass(bd2.getAbstractSchemaName());
        if (clazz2 == null) {
            // The class is not yet defined => fetch a hidden class
            clazz2 = schema.getHiddenClass(bd2.getAbstractSchemaName());
            if (clazz2 == null) {
                clazz2 = schema.createHiddenClass(bd2.getAbstractSchemaName(), false);
            }
        }

        // This namedef describes how the referenced class
        // is referenced from the gen class.
        // In a implicit mapping this NameDef is equals to the NameDef
        // of the referenced class
        logger.log(BasicLevel.DEBUG, "Fill the name def of gen class element");
        fillNameDef(gcr.createClassRef(clazz2).createNameDef(), schema, rsrd,
                    EjbRelationshipRoleDesc.TARGET, bd2, mb, baseFieldName2, true, gcr, logger);

    }

    /**
     * This method create the jorm meta information which matches with a
     * simple relation between two bean.
     * @param clazz is the class (jorm meta object) of the first bean.
     * @param rsrd is the descriptor of the relation role
     * @param bd is the descriptor of the second bean
     */
    private void createClassReference(Class clazz,
                                      EjbRelationshipRoleDesc rsrd,
                                      EntityCmp2Desc bd,
                                      MappingBuilder mb, Logger logger)
        throws DeploymentDescException {
        String cmrfieldName = rsrd.getCmrFieldName();
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                       "Create ClassReference for the CMR field: " + cmrfieldName);
        }

        Schema schema = (Schema) clazz.getParent();
        String className = bd.getAbstractSchemaName();
        Class c = schema.getClass(className);
        if (c==null) {
            c = schema.getHiddenClass(className);
            if (c == null)
                c = schema.createHiddenClass(className, false);
        }
        ClassRef cr = clazz.createClassRef(cmrfieldName, c);

        fillNameDef(cr.createNameDef(), schema, rsrd, EjbRelationshipRoleDesc.TARGET, bd,
                    mb, cmrfieldName, true,clazz, logger);
    }

    /**
     * This method declares a new hidden field in a class.
     * @param fieldName is the name of the new field
     * @param desc describes the new field. But the name specified inside must
     * not be used. Indeed in order to prevent conflict the use of this method
     * specifies a name (fieldName param).
     * @param clazz is the class (jorm meta object) inside which the field must
     * be declared.
     * @throws DeploymentDescException when an existing field (hidden or not)
     * has the same name.
     */
    private void declareHiddenField(String fieldName,
                                    FieldDesc desc,
                                    EjbRelationshipRoleDesc rsrd,
                                    byte type,
                                    MetaObject _clazz,
                                    MappingBuilder mb) throws DeploymentDescException {
        if (_clazz instanceof Class) {
            Class clazz = (Class) _clazz;
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Declare the hidden field "
                           + fieldName + " in the class " + clazz.getName());
            }
            if (clazz.getHiddenField(fieldName) != null
                || clazz.getTypedElement(fieldName) != null) {
                throw new DeploymentDescException("Bad implicit mapping choice: Two fields with same name: "
                                                  + fieldName);
            }
            PrimitiveElement pe = (PrimitiveElement)
                clazz.createHiddenField(fieldName, JormType.getPType(desc.getFieldType()));
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Create ValueMapping for a Class reference");
            }
            mb.createValueMapping(pe, desc, rsrd, type, PROJECT_NAME);
        } else if (_clazz instanceof GenClassRef) {
            GenClassRef clazz = (GenClassRef) _clazz;
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Declare the hidden field "
                           + fieldName + " in the genclass referenced by the field"
                           + clazz.getName());
            }
            if (clazz.getHiddenField(fieldName) != null) {
                throw new DeploymentDescException("Bad implicit mapping choice: Two fields with same name: "
                                                  + fieldName + " in a genclass referenced by the field: "
                                                  + clazz.getName());
            }
            PrimitiveElement pe = (PrimitiveElement)
                clazz.createHiddenField(fieldName, JormType.getPType(desc.getFieldType()));
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Create ValueMapping for a Class reference");
            }
            mb.createValueMapping(pe, desc, rsrd, type, PROJECT_NAME);
        } else {
            throw new DeploymentDescException("Bad implicit mapping case: impossible to declare hidden field into this jorm meta object: " + _clazz);
        }
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(super.toString());
        return ret.toString();
    }

    private void fillNameDef(NameDef nds,
                             Schema schema,
                             EjbRelationshipRoleDesc rsrd,
                             byte type,
                             EntityCmp2Desc _bd,
                             MappingBuilder mb,
                             String baseFieldName,
                             boolean declareField,
                             MetaObject clazz,
                             Logger logger) throws DeploymentDescException {

        logger.log(BasicLevel.DEBUG,"**** fillNameDef ****");
        logger.log(BasicLevel.DEBUG,
                   "rsrd="+rsrd.toString()+
                   "\ntype="+type+
                   "\n_bd.getEjbName()"+_bd.getEjbName()+
                   "\nbaseFieldName="+baseFieldName+
                   "\ndeclareField="+declareField+
                   "\nclass.type="+clazz.getClass().getName());

        // Fill the NameDef of the class either the primary key
        EntityCmp2Desc bd = (rsrd.isMultiValuedCMR && clazz instanceof Class
                             ? rsrd.getOppositeEjbRelationshipRoleDesc().getLinkedEntityCmp2Desc()
                             : _bd);
        nds.setMappNames(PROJECT_NAME);
        if (bd.hasPrimaryKeyField()) {
            // one primary key field
            if (declareField) {
                // This case appears in the gen class only (id  & elem) or in
                // a class reference.
                // The name is prefixed by the bean name in order to avoid collisions
                String clafn = baseFieldName + "_" + bd.getPrimaryKeyFieldName();
                nds.setFieldName(clafn);
                declareHiddenField(clafn,
                                   bd.getCmpFieldDesc(bd.getPrimaryKeyFieldName()),
                                   rsrd,
                                   type,
                                   clazz,
                                   mb);
            } else {
                String classfn = bd.getPrimaryKeyFieldName();
                // Check if the field is already declared
                if (clazz instanceof Class
                    && ((Class) clazz).getTypedElement(classfn) == null) {
                    throw new DeploymentDescException("Bad Implicit mapping: field " + classfn 
                                                      + " is not availlable on the class "
                                                      + ((Class) clazz).getFQName());
                }
                // The field does not need to be declared because it already
                // exist and it is not hidden.
                nds.setFieldName(classfn);
                logger.log(BasicLevel.DEBUG,
                           "No declaration of the single field: "+ classfn);
            }
        } else {
            //Define a Composite Name and a NameRef
            // The name of the composite is built with the abstract schema
            // name and the suffix "Pk"
            String str = bd.getAbstractSchemaName() + "Pk";
            CompositeName cn = schema.getCompositeName(str);
            if (cn == null)
                cn = schema.createCompositeName(str);


            // Used the new CompositeName into the NameDef
            NameRef nrs = nds.createNameRef(cn.getName(), cn);
            // fill the composite name with pk fields and link these field to
            // to the class field
            for (Iterator it = bd.getCmpFieldDescIterator(); it.hasNext();) {
                FieldDesc fd = (FieldDesc) it.next();
                if (fd.isPrimaryKey()) {
                    String cofn = fd.getFieldName();
                    String classfn = cofn;
                    if (declareField) {
                        // In this case it is possible to have a collision between pk
                        // fields
                        classfn = baseFieldName + "_" + classfn;
                        declareHiddenField(classfn, fd, rsrd, type, clazz, mb);

                    } else if (clazz instanceof Class
                               && ((Class) clazz).getTypedElement(classfn) == null) {
                        throw new DeploymentDescException("Bad Implicit mapping: field "
                                                          + classfn + " is not availlable on the class " + ((Class) clazz).getFQName());

                    } else if (clazz instanceof GenClassRef
                               && ((GenClassRef) clazz).getHiddenField(classfn) == null) {
                        throw new DeploymentDescException("Bad Implicit mapping: hidden field "
                                                          + classfn + " is not availlable on the GenClassRef " + ((GenClassRef) clazz).getName());
                    }

                    if (cn.getScalarField(cofn) == null)
                        cn.createCompositeNameField(cofn, JormType.getPType(fd.getFieldType()));
                    nrs.addProjection(cofn, classfn);
                }
            }
        }
    }

    private String upperFL(String word) {
        return Character.toUpperCase(word.charAt(0)) + word.substring(1);
    }

}
