/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.api;

import org.objectweb.jonas_ejb.deployment.xml.EjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelationshipRole;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelationshipRoleName;


/**
 * Created on Jul 7, 2002
 * Class to hold meta-information related to an ejb-relation.
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class EjbRelationDesc {

    private String name;
    private EjbRelationshipRoleDesc relationshipRoleDesc1;
    private EjbRelationshipRoleDesc relationshipRoleDesc2;

    /**
     * constructor to be used by parent node
     */
    public EjbRelationDesc(EjbRelation ejbRelation) throws DeploymentDescException {
        name = ejbRelation.getEjbRelationName().getValue();

        // first role
        EjbRelationshipRole role1 = ejbRelation.getEjbRelationshipRole();
        EjbRelationshipRoleName ersrn1 = role1.getEjbRelationshipRoleName();
        if (ersrn1 == null) {
            throw new DeploymentDescException("You must specify a ejb-relationship-role-name in the relation "
                                              + name + " (first role). It is required by JOnAS");
        }

        // second role
        EjbRelationshipRole role2 = ejbRelation.getEjbRelationshipRole2();
        EjbRelationshipRoleName ersrn2 = role2.getEjbRelationshipRoleName();
        if (ersrn2 == null) {
            throw new DeploymentDescException("You must specify a ejb-relationship-role-name in the relation "
                                              + name + " (second role). It is required by JOnAS");
        }

        relationshipRoleDesc1 = new EjbRelationshipRoleDesc(role1, role2.getCascadeDelete() != null, this, ejbRelation);
        relationshipRoleDesc2 = new EjbRelationshipRoleDesc(role2, role1.getCascadeDelete() != null, this, ejbRelation);

        // Add the opposite CMR field for the relation XXu
        // in order to implement the coherence.
        boolean r1hf = relationshipRoleDesc1.hasCmrField();
        boolean r2hf = relationshipRoleDesc2.hasCmrField();
        EjbRelationshipRoleDesc nocmr = null;
        EjbRelationshipRoleDesc cmr = null;
        if (r1hf && !r2hf) {
            nocmr = relationshipRoleDesc2;
            cmr = relationshipRoleDesc1;
        } else if (!r1hf && r2hf) {
            nocmr = relationshipRoleDesc1;
            cmr = relationshipRoleDesc2;
        }
        if (nocmr != null) {
            // The relation is OXu, the role 'role' does not have cmr field
            String cmrName = name;
            // calculate a cmr field name with the relation name. The bad
            // character are replaced by the character '_'.
            for (int i = 0; i < cmrName.length(); i++) {
                char c = cmrName.charAt(i);
                if (!Character.isJavaIdentifierPart(c)) {
                    cmrName = cmrName.replace(c, '_');
                }
            }
            cmrName = "jonasCMR" + cmrName;
            // Add the cmr, no type is specified because the added cmr field
            // is mono valued.
            nocmr.setCMRFieldName(cmrName);
            nocmr.setIsJOnASCMR(true);
            if (nocmr.isMultiValuedCMR()) {
                if (cmr.isMultiValuedCMR()) {
                    nocmr.setCMRFieldType(cmr.cmrFieldType.getName());
                } else {
                    nocmr.setCMRFieldType("java.util.Collection");
                }
            }
        }
    }

    /**
     * get the name of the relationship.
     */
    public String getName() {
        return name;
    }

    /**
     * get the meta-information for the first relation-ship-role
     */
    public EjbRelationshipRoleDesc getRelationshipRoleDesc1() {
        return relationshipRoleDesc1;
    }

    /**
     * get the meta-information for the second relation-ship-role
     */
    public EjbRelationshipRoleDesc getRelationshipRoleDesc2() {
        return relationshipRoleDesc2;
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\ngetName()=" + getName());
        ret.append("\ngetRelationshipRoleDesc1() = " + getRelationshipRoleDesc1());
        ret.append("\ngetRelationshipRoleDesc2() = " + getRelationshipRoleDesc2());
        return ret.toString();
    }
}
