/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Christophe Ney
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas_ejb.deployment.api;

import org.objectweb.jonas_ejb.deployment.xml.CmrFieldType;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelation;
import org.objectweb.jonas_ejb.deployment.xml.EjbRelationshipRole;


/**
 * Class to hold meta-information related to an ejb-relation-role
 * Created on Jul 7, 2002
 * @author Christophe Ney [cney@batisseurs.com]
 */
public class EjbRelationshipRoleDesc {

    /**
     * This constant can be used to represent an index of a gen class.
     */
    public static final byte INDEX = 1;

    /**
     * This constant can be used to represent the source of the relation, ie the
     * element which references the other.
     */
    public static final byte SOURCE = 2;

    /**
     * This constant can be used to represent the target of the relation, ie the
     * element which is referenced by the other.
     */
    public static final byte TARGET = 4;

    public static final byte OOU = 0;
    public static final byte OOB = 1;
    public static final byte OMU = 2;
    public static final byte OMB = 3;
    public static final byte MOU = 4;
    public static final byte MOB = 5;
    public static final byte MMU = 6;
    public static final byte MMB = 7;

    boolean isMultiValuedCMR;
    String ejbName;
    EjbRelationDesc ejbRelationDesc;
    EntityCmp2Desc linkedEntityCmp2Desc;
    protected String cmrFieldName = null;
    protected Class cmrFieldType = null;
    protected boolean isJOnASCMR = false;
    private byte relationType = -1;
    private boolean mustCascade;


    /**
     * constructor to be used by parent node
     */
    public EjbRelationshipRoleDesc(EjbRelationshipRole role,
                                   boolean mustCascade,
                                   EjbRelationDesc ejbRelationDesc,
                                   EjbRelation ejbRelation) throws DeploymentDescException {
        this.ejbRelationDesc = ejbRelationDesc;
        this.mustCascade = mustCascade;
        boolean cascadeDelete = (role.getCascadeDelete() != null);
        ejbName = role.getRelationshipRoleSource().getEjbName().getValue();
        // mutiplicity is One or Many
        EjbRelationshipRole opposite = ejbRelation.getEjbRelationshipRole();
        if (opposite == role) {
            opposite = ejbRelation.getEjbRelationshipRole2();
        }
        if (opposite.getMultiplicity().getValue().equals("Many")) {
            isMultiValuedCMR = true;
            if (cascadeDelete) {
                throw new DeploymentDescException("Cascade delete not allowed because " 
                                                  + opposite.getEjbRelationshipRoleName().getValue() 
                                                  + " has a multiplicity of : MANY");
            }
        } else if (opposite.getMultiplicity().getValue().equals("One")) {
            isMultiValuedCMR = false;
        } else {
            throw new DeploymentDescException(
                                              "Invalid multiplicity value for relationshipRole \""
                                              + role.getEjbRelationshipRoleName().getValue()
                                              + "\". (must be One or Many)");
        }
        // store cmr field if any
        if (role.getCmrField() != null) {
            cmrFieldName = role.getCmrField().getCmrFieldName().getValue();
            if (isMultiValuedCMR) {
                CmrFieldType type = role.getCmrField().getCmrFieldType();
                if (type == null) {
                    throw new DeploymentDescException("You must specify a cmr-field-type in case where the relation is 'Many' in the cmr-field '"
                                                      + cmrFieldName + "' of bean " + ejbName);
                }
                setCMRFieldType(type.getValue());
            }
        }
    }

    public void setCMRFieldName(String name) throws DeploymentDescException {
        cmrFieldName = name;
    }

    public void setCMRFieldType(String type) throws DeploymentDescException {
        try {
            cmrFieldType = Class.forName(type);
        } catch (ClassNotFoundException e) {
            throw new DeploymentDescException("class name not found for cmr-field " + cmrFieldName
                                              + " of bean " + ejbName, e);
        }
        if (!(cmrFieldType.getName().equals("java.util.Collection")
              || cmrFieldType.getName().equals("java.util.Set")))
            throw new DeploymentDescException("value of cmr-field-type "
                                              + cmrFieldName + " of bean " + ejbName
                                              + " should be java.util.Set or java.util.Collection if set");
    }

    /**
     * mark the cmr as added by JOnAS
     */
    public void setIsJOnASCMR(boolean b) {
        isJOnASCMR = b;
    }

    /**
     * It returns a boolean value which indicates if the cmr has been added by
     * JOnAS (true) or if the user has specified a cmr field in the descriptor.
     * A CMR field is be added to manage the coherence of the relation OXu
     */
    public boolean isJOnASCMR() {
        return isJOnASCMR;
    }

    /**
     * return true if this bean must cascade delete the other bean in this relation.
     */
    public boolean mustCascade() {
        return mustCascade;
    }

    /**
     * get state of field is relation multiple
     */
    public boolean isMultiValuedCMR() {
        return isMultiValuedCMR;
    }

    /**
     * get name of the ejb involved in this relation-ship-role
     */
    public String getEjbName() {
        return ejbName;
    }


    /**
     * get the parent ejb relation
     */
    public EjbRelationDesc getEjbRelationDesc() {
        return ejbRelationDesc;
    }

    public EjbRelationshipRoleDesc getOppositeEjbRelationshipRoleDesc() {
        EjbRelationshipRoleDesc res = ejbRelationDesc.getRelationshipRoleDesc1();
        if (res == this) {
            return ejbRelationDesc.getRelationshipRoleDesc2();
        } else {
            return res;
        }
    }

    /**
     * It retrieves the EntityCmp2Desc which is linked to the EntityCmp2Desc
     * associated to this EjbRelationshipRoleDesc.
     */
    public EntityCmp2Desc getLinkedEntityCmp2Desc() {
        return linkedEntityCmp2Desc;
    }

    /**
     * set the linkedEntityCmp2Desc field.
     */
    public void setLinkedEntityCmp2Desc(EntityCmp2Desc linkedEntityCmp2Desc) {
        this.linkedEntityCmp2Desc = linkedEntityCmp2Desc;
        if (cmrFieldType == null) {
            cmrFieldType = linkedEntityCmp2Desc.getLocalClass();
        }
    }

    /**
     * It retrieves true if the EntityCmp2Desc associated to this
     * EjbRelationshipRoleDesc has a cmr field to the linked EntityCmp2Desc
     */
    public boolean hasCmrField() {
        return cmrFieldName != null;
    }

    /**
     * get the name of the cmr-field
     */
    public String getCmrFieldName() {
        return cmrFieldName;
    }

    /**
     * get the type of the cmr-field when set in the deployment descriptor.
     * @return Collection or Set for multiple rel. and null for non multiple rel.
     */
    public Class getCmrFieldType() {
        return cmrFieldType;
    }

    public byte getRelationType() {
        if (relationType == -1) {
            relationType = 0;
            EjbRelationshipRoleDesc rsr2 = getOppositeEjbRelationshipRoleDesc();
            if (rsr2.hasCmrField() && hasCmrField()) {
                relationType += 1;
            }
            if (isMultiValuedCMR()) {
                relationType += 2;
            }
            if (rsr2.isMultiValuedCMR()) {
                relationType += 4;
            }
        }
        return relationType;
    }

    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\nmustCascade() = " + mustCascade());
        ret.append("\nisMultiValuedCMR() = " + isMultiValuedCMR());
        ret.append("\ngetEjbName() = " + getEjbName());
        ret.append("\ngetEjbRelationDesc().getName() = " + getEjbRelationDesc().getName());
        ret.append("\ngetLinkedEntityCmp2Desc().getEjbName() = " + getLinkedEntityCmp2Desc().getEjbName());
        ret.append("\ngetCmrFieldName() = " + getCmrFieldName());
        ret.append("\ngetCmrFieldType() = " + getCmrFieldType());
        return ret.toString();
    }

}
