/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Laurent Chauvirey, Frederic Maistre, Nicolas Tachker
 * Contributor(s): 
 * Philippe Durieux
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_jms;

import java.io.Serializable;
import javax.jms.BytesMessage;
import javax.jms.JMSException;
import javax.jms.MapMessage;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.StreamMessage;
import javax.jms.TextMessage;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.objectweb.util.monolog.api.BasicLevel;

public abstract class JSession implements Session, XAResource {

    protected XAResource xares = null;	// The underlaying XAResource 
    protected boolean txover = true;
    protected Transaction currtx = null;
    protected boolean closed = false;
    protected JConnection jconn;
    protected static TransactionManager tm = null;

    /**
     * Constructor
     */
    public JSession(JConnection jconn) {
	this.jconn = jconn;
	if (tm == null) {
	    tm = JmsManagerImpl.getTransactionManager();
	}
    }

    // -----------------------------------------------------------------------
    // Internal Methods
    // -----------------------------------------------------------------------

    /**
     * Get the underlaying MOM Session.
     */
    abstract protected Session getMOMSession() throws JMSException;

    abstract protected void MOMSessionClose();

    /**
     * 
     */
    protected void PhysicalClose() {
	MOMSessionClose();
    }

    // -----------------------------------------------------------------------
    // Session Implementation
    // -----------------------------------------------------------------------

    /**
     * 
     */
    public void close() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	closed = true;
	if (txover) {
	    PhysicalClose();
	} else {
	    // delist XAResource now, that will lead to an XA-end call
	    if (currtx == null) {
		TraceJms.logger.log(BasicLevel.ERROR, "should be in a tx");
	    } else {
		try {
		    currtx.delistResource(this, XAResource.TMSUCCESS);
		} catch (SystemException e) {
		    TraceJms.logger.log(BasicLevel.ERROR, "cannot delist session:"+e);
		    throw new JMSException(e.toString());
		}
	    }
	}
    }

    /**
     * 
     */
    public void commit() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	throw new JMSException("JSession: commit Operation Not Allowed");
    }

    /**
     * 
     */
    public BytesMessage createBytesMessage() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createBytesMessage();
    }

    /**
     * 
     */
    public MapMessage createMapMessage() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createMapMessage();
    }

    /**
     * 
     */
    public Message createMessage() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createMessage();
    }

    /**
     * 
     */
    public ObjectMessage createObjectMessage() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createObjectMessage();
    }

    /**
     * 
     */
    public ObjectMessage createObjectMessage(Serializable object) throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createObjectMessage(object);
    }

    /**
     * 
     */
    public StreamMessage createStreamMessage() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createStreamMessage();
    }

    /**
     * 
     */
    public TextMessage createTextMessage() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createTextMessage();
    }

    /**
     * 
     */
    public TextMessage createTextMessage(String text) throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().createTextMessage(text);
    }

    /**
     * 
     */
    public MessageListener getMessageListener() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().getMessageListener();
    }
    
    /**
     * 
     */
    public boolean getTransacted() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return getMOMSession().getTransacted();
    }

    /**
     * 
     */
    public void recover() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	throw new JMSException("JSession: recover Operation Not Allowed");
    }

    /**
     * 
     */
    public void rollback() throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	throw new JMSException("JSession: rollback Operation Not Allowed");
    }

    /**
     * 
     */
    public void run() {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	try {
	    getMOMSession().run();
	} catch (JMSException e) {
	    TraceJms.logger.log(BasicLevel.ERROR, "exception: "+e);
	}
    }

    /**
     * 
     */
    public void setMessageListener(MessageListener listener) throws JMSException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	getMOMSession().setMessageListener(listener);
    }

    // -----------------------------------------------------------------------
    // XAResource Implementation
    // -----------------------------------------------------------------------

    /**
     *
     */	
    public void commit(Xid xid, boolean onePhase) throws XAException {
	if(TraceJms.isDebug()){
	    TraceJms.logger.log(BasicLevel.DEBUG, "xid = "+xid);
	}
	xares.commit(xid, onePhase);
	if (txover) {
	    TraceJms.logger.log(BasicLevel.ERROR, "tx over ???");
	}
	txover = true;
	if (closed) {
	    PhysicalClose();
	}
    }

    /**
     *
     */	
    public void end(Xid xid, int flags) throws XAException {
	if(TraceJms.isDebug()){
	    TraceJms.logger.log(BasicLevel.DEBUG, "xid = "+xid);
	}
	if (txover) {
	    TraceJms.logger.log(BasicLevel.ERROR, "tx over ???");
	}
	xares.end(xid, flags);
    }

    /**
     *
     */	
    public void forget(Xid xid) throws XAException {
	if(TraceJms.isDebug()){
	    TraceJms.logger.log(BasicLevel.DEBUG, "xid = "+xid);
	}
	xares.forget(xid);
    }

    /**
     *
     */	
    public int getTransactionTimeout() throws XAException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return xares.getTransactionTimeout();
    }

    /**
     *
     */	
    public boolean isSameRM(XAResource xar) throws XAException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return xares.isSameRM(xar);
    }

    /**
     *
     */	
    public int prepare(Xid xid) throws XAException {
	if(TraceJms.isDebug()){
	    TraceJms.logger.log(BasicLevel.DEBUG, "xid = "+xid);
	}
	return xares.prepare(xid);
    }

    /**
     *
     */	
    public Xid[] recover(int flag) throws XAException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return xares.recover(flag);
    }

    /**
     *
     */	
    public void rollback(Xid xid) throws XAException {
	if(TraceJms.isDebug()){
	    TraceJms.logger.log(BasicLevel.DEBUG, "xid = "+xid);
	}
	xares.rollback(xid);
	if (txover) {
	    TraceJms.logger.log(BasicLevel.ERROR, "tx over ???");
	}
	txover = true;
	if (closed) {
	    PhysicalClose();
	}
    }

    /**
     *
     */	
    public boolean setTransactionTimeout(int seconds) throws XAException {
	TraceJms.logger.log(BasicLevel.DEBUG, "");
	return xares.setTransactionTimeout(seconds);
    }

    /**
     *
     */	
    public void start(Xid xid, int flags) throws XAException {
	if(TraceJms.isDebug()){	
	    TraceJms.logger.log(BasicLevel.DEBUG, "xid = "+xid);
	}
	txover = false;
	xares.start(xid, flags);
    }


}

