/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Eric HARDESTY
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_rar.deployment.api;

//import jonas_lib.deployment
import org.objectweb.jonas_lib.deployment.AbsDeploymentDesc;
import org.objectweb.jonas_lib.deployment.SAXParserHelper;

//xml of connector.dtd and jonas-connector.dtd
import org.objectweb.jonas_rar.deployment.xml.Connector;
import org.objectweb.jonas_rar.deployment.xml.ConnectorUnmarshaller;
import org.objectweb.jonas_rar.deployment.xml.VendorName;
import org.objectweb.jonas_rar.deployment.xml.Version;
import org.objectweb.jonas_rar.deployment.xml.SpecVersion;
import org.objectweb.jonas_rar.deployment.xml.Resourceadapter;

import org.objectweb.jonas_rar.deployment.xml.JonasResource;
import org.objectweb.jonas_rar.deployment.xml.JonasResourceUnmarshaller;
import org.objectweb.jonas_rar.deployment.xml.Jndiname;
import org.objectweb.jonas_rar.deployment.xml.Rarlink;
import org.objectweb.jonas_rar.deployment.xml.PoolParams;
import org.objectweb.jonas_rar.deployment.xml.ConfigProperty;
import org.objectweb.jonas_rar.deployment.xml.JonasConfigProperty;

// SAX imports
import org.xml.sax.ErrorHandler;
import org.xml.sax.EntityResolver;

import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/**
 * This class extends the AbsDeploymentDescriptor class of JOnAS
 * It provides a description of the specific RAR desployment descriptor
 * @author Eric Hardesty
 */

public class RarDeploymentDesc extends AbsDeploymentDesc {

    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = true;

    private static Connector connector = null;
    
    private static JonasResource jonasConnector = null;
 
    /**
     * Vector of raConfigPropTags (config-property objects).
     */
    private Vector raConfigPropTags = null;

    /**
     * Vector of jonasConfigPropTags (config-property objects).
     */
    private Vector jonasConfigPropTags = null;


    /**
     * Path of the ra.xml deploymnet descriptor file
     */
    private static final String RA_FILE_NAME = "META-INF/ra.xml";

    /**
     * Path of the jonas-ra.xml deploymnet descriptor file
     */
    private static final String JONAS_RA_FILE_NAME = "META-INF/jonas-ra.xml";



    /**
     * Construct an instance of a RarDeploymentDesc.
     * Called by the getInstance() method.
     * @param classLoader ClassLoader of the classes .
     * @param connector ra.xml parsed file.
     * @param jonasConnector jonas-ra.xml parsed file.
     * @throws RarDeploymentDescException if we can't build an instance.
     */
    private RarDeploymentDesc(ClassLoader classLoader, Connector connector,
                              JonasResource jonasConnector) throws RarDeploymentDescException {
 

        // test classloader
        if (classLoader == null) {
            throw new RarDeploymentDescException("DeploymentDesc: Classloader is null");
        }

        if (connector != null) {
            // display name
            displayName = (connector.getDisplayName() != null) ? connector.getDisplayName().getValue() : null;

            raConfigPropTags = new Vector();
            //Tags
            for (Iterator i = connector.getResourceadapter().getConfigPropertyList().iterator(); i.hasNext();) {
                raConfigPropTags.add((ConfigProperty) i.next());
            }
        }
        
        if (jonasConnector != null) {
            jonasConfigPropTags = new Vector();
            //Tags
            for (Iterator i = jonasConnector.getJonasConfigPropertyList().iterator(); i.hasNext();) {
                jonasConfigPropTags.add((JonasConfigProperty) i.next());
            }
        }

    }
    
    /**
     * Set the config-property tags of the ra.xml file.
     * @param the Connector objeect with the config-property tags of the ra.xml file.
     */
    public void setRaConfigPropTags(Connector conn) {
        //Tags
        for (Iterator i = conn.getResourceadapter().getConfigPropertyList().iterator(); i.hasNext();) {
            if (raConfigPropTags == null)
                raConfigPropTags = new Vector();
            raConfigPropTags.add((ConfigProperty) i.next());
        }
        if (raConfigPropTags == null) return;
        ConfigProperty[] tmp = new ConfigProperty[raConfigPropTags.size()];
        raConfigPropTags.copyInto(tmp);
    }

    /**
     * Get the config-property tags of the ra.xml file.
     * @return the config-property tags of the ra.xml file.
     */
    public ConfigProperty[] getRaConfigPropTags() {
        if (raConfigPropTags == null) return null;
        ConfigProperty[] tmp = new ConfigProperty[raConfigPropTags.size()];
        raConfigPropTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the config-property tags of the jonas-ra.xml file.
     * @return the config-property tags of the jonas-ra.xml file.
     */
    public JonasConfigProperty[] getJonasConfigPropTags() {
        if (jonasConfigPropTags == null) return null;
        if (jonasConfigPropTags != null) {
            JonasConfigProperty[] tmp = new JonasConfigProperty[jonasConfigPropTags.size()];
            jonasConfigPropTags.copyInto(tmp);
            return tmp;
        }
        return null;
    }

    /**
     * Get an instance of an RAR deployment descriptor by parsing the 
     * ra.xml and jonas-ra.xml deployment descriptors.
     * @param rarFileName the fileName of the rar file for the deployment descriptors.
     * @param classLoader the classloader for the classes.
     * @return an RAR deployment descriptor by parsing the ra.xml & jonas-ra.xml
     * deployment descriptors.
     * @throws RarDeploymentDescException if the deployment descriptors are 
     * corrupted.
     */
    public static RarDeploymentDesc getInstance(String rarFileName,
                                                ClassLoader classLoader)
        throws RarDeploymentDescException {
        
        // want both ra.xml & jonas-ra.xml
        return getInstance(rarFileName, classLoader, true, true);
    }
    
    /**
     * Get an instance of an RAR deployment descriptor by parsing the 
     * ra.xml and jonas-ra.xml deployment descriptors.
     * @param rarFileName the fileName of the rar file for the deployment descriptors.
     * @param classLoader the classloader for the classes.
     * @param setupJonasRa a boolean to determine if the ra.xml should be processed
     * @param setupJonasRa a boolean to determine if the jonas-ra.xml should be processed
     * @return an RAR deployment descriptor by parsing the ra.xml & jonas-ra.xml
     * deployment descriptors.
     * @throws RarDeploymentDescException if the deployment descriptors are 
     * corrupted.
     */
    public static RarDeploymentDesc getInstance(String rarFileName,
                                                ClassLoader classLoader,
                                                boolean setupRa,
                                                boolean setupJonasRa)
        throws RarDeploymentDescException {

        //ear file
        JarFile rarFile = null;

        //Input Stream
        InputStream raInputStream = null;
        InputStream jonasRaInputStream = null;

        //ZipEntry
        ZipEntry raZipEntry = null;
        ZipEntry jonasRaZipEntry = null;
        
        connector = null;
        jonasConnector = null;

        //Check if the file exists.
        if (!new File(rarFileName).exists()) {
            throw new RarDeploymentDescException("The file '" + rarFileName + "' was not found.");
        }

        // load deployment descriptor data (META-INF/ra.xml)
        try {
            rarFile = new JarFile(rarFileName);;

            if (setupRa) {
                //Check the ra entry
                raZipEntry = rarFile.getEntry(RA_FILE_NAME);
                if (raZipEntry == null) {
                    connector = null;
                    setupRa = false;
                }
                else
                    //Get the stream
                    raInputStream = rarFile.getInputStream(raZipEntry);
            }

            if (setupJonasRa) {
                //Check the jonas-ra entry
                jonasRaZipEntry = rarFile.getEntry(JONAS_RA_FILE_NAME);
                if (jonasRaZipEntry == null) {
                    jonasConnector = null;
                    setupJonasRa = false;
                }
                else
                    //Get the stream
                    jonasRaInputStream = rarFile.getInputStream(jonasRaZipEntry);
            }

        } catch (Exception e) {
            if (rarFile != null) {
                try {
                    rarFile.close();
                } catch (IOException ioe) {
                    //We can't close the file
                }
            }
            throw new RarDeploymentDescException("Cannot read the XML deployment descriptors of the rar file '" + rarFileName + "'.", e);
        }

        if (setupRa) {
            connector = loadConnector(raInputStream, RA_FILE_NAME);
            try {
                raInputStream.close();
            } catch (IOException e) {
            }
        }

        if (setupJonasRa) {
            jonasConnector = loadJonasConnector(jonasRaInputStream, JONAS_RA_FILE_NAME);
            try {
                jonasRaInputStream.close();
            } catch (IOException e) {
            }
        }
        
        // instantiate deployment descriptor
        return new RarDeploymentDesc(classLoader, connector, jonasConnector);
    }

    /**
     * Load the ra.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (ra.xml).
     * @throws RarDeploymentDescException if the deployment descriptor is 
     * corrupted.
     * @return a connector object.
     */
    private static Connector loadConnector(InputStream is, String fileName)
        throws RarDeploymentDescException {

        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);

        try {
            ConnectorUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            ConnectorUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());

            // Unmarshal the XML, and validate as it is unmarshalled
            return ConnectorUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new RarDeploymentDescException(detailMsg, e);
            } else {
                throw new RarDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        }
    }

    /**
     * Load the jonas-ra.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (jonas-ra.xml).
     * @throws RarDeploymentDescException if the deployment descriptor is 
     * corrupted.
     * @return a JonasResource object.
     */
    private static JonasResource loadJonasConnector(InputStream is, String fileName)
        throws RarDeploymentDescException {

        Reader reader = new InputStreamReader(is);
        SAXParserHelper helper = new SAXParserHelper(errorHandler, entityResolver);

        try {
            JonasResourceUnmarshaller.setErrorHandler(helper.getWrappedErrorHandler());
            JonasResourceUnmarshaller.setEntityResolver(helper.getWrappedEntityResolver());

            // Unmarshal the XML, and validate as it is unmarshalled
            return JonasResourceUnmarshaller.unmarshal(reader, parsingWithValidation);
        } catch (IOException e) {
            if (helper.getSAXParseException() != null) {
                String detailMsg = getSAXMsg(fileName, helper.getSAXParseException(), "Invalid XML");
                throw new RarDeploymentDescException(detailMsg, e);
            } else {
                throw new RarDeploymentDescException("Error reading XML document " + fileName, e);
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignored) { }
            }
        }
    }

    /** 
     * Return a String representation of the RarDeploymentDesc.
     * @return a String representation of the RarDeploymentDesc.
     */
    public String toString() {
        String ret = "Connector :";

        return ret;
    }

    /**
     * Get the current Connector.
     * @return the current Connector.
     */
    public static Connector getConnector() {
        return connector;
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Get the current JonasResource.
     * @return the current JonasResource.
     */
    public static JonasResource getJonasResource() {
        return jonasConnector;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return RarDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        RarDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        RarDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        RarDeploymentDesc.parsingWithValidation = validation;
    }



}
