/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Craig R. McClanahan
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.security.catalina41.realm;

//import java
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.security.Principal;

//import apache-commons
import org.apache.commons.digester.Digester;
import org.apache.commons.digester.Rule;
import org.apache.commons.digester.RuleSetBase;

//import catalina
import org.apache.catalina.Container;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.realm.RealmBase;
import org.apache.catalina.realm.GenericPrincipal;
import org.apache.catalina.realm.Constants;
import org.apache.catalina.util.StringManager;

//import objectweb
import org.objectweb.security.context.SecurityCurrent;
import org.objectweb.security.context.SecurityContext;
import org.objectweb.jonas.common.Log;

//import monolog
import org.objectweb.util.monolog.api.BasicLevel;

//import xml
import org.xml.sax.Attributes;

/**
 * This class is used as security interceptor for tomcat 4.1.x
 * for propagating security context
 * The original code came from org/apache/catalina/realm/MemoryRealm
 * The only difference is in authenticate method where a JOnAS Security context
 * is associated to the current thread
 *
 * @author Craig R. McClanahan
 * @author Philippe Coq : Tomcat/JOnAS 4.0.x
 * @author Florent Benoit : Tomcat/JOnAS 4.1.x
 *
 */
public final class MemoryRealm extends RealmBase {

    /**
     * The Container with which this Realm is associated.
     */
    private Container container = null;


    /**
     * The Digester we will use to process in-memory database files.
     */
    private static Digester digester = null;


    /**
     * Descriptive information about this Realm implementation.
     */
    protected final String info = "org.objectweb.security.catalina41.realm.MemoryRealm/1.0";


    /**
     * Descriptive information about this Realm implementation.
     */
    protected static final String name = "MemoryRealm";


    /**
     * The pathname (absolute or relative to Catalina's current working
     * directory) of the XML file containing our database information.
     */
    private String pathname = "conf/tomcat-users.xml";


    /**
     * The set of valid Principals for this Realm, keyed by user name.
     */
    private HashMap principals = new HashMap();


    /**
     * The string manager for this package.
     */
    private static StringManager sm = StringManager.getManager(Constants.Package);


    /**
     * Has this component been started?
     */
    private boolean started = false;


    /**
     * Return descriptive information about this Realm implementation and
     * the corresponding version number, in the format
     * <code>&lt;description&gt;/&lt;version&gt;</code>.
     * @return the info
     */
    public String getInfo() {
        return info;
    }


    /**
     * Return the pathname of our XML file containing user definitions.
     * @return the path name
     */
    public String getPathname() {
        return pathname;
    }


    /**
     * Set the pathname of our XML file containing user definitions.  If a
     * relative pathname is specified, it is resolved against "catalina.base".
     *
     * @param pathname The new pathname
     */
    public void setPathname(String pathname) {
        this.pathname = pathname;
    }


    /**
     * Return the Principal associated with the specified username and
     * credentials, if there is one; otherwise return <code>null</code>.
     *
     * @param username Username of the Principal to look up
     * @param credentials Password or other credentials to use in
     *  authenticating this username
     * @return the principal associated
     */
    public Principal authenticate(String username, String credentials) {
        GenericPrincipal principal = (GenericPrincipal) principals.get(username);

        boolean validated = false;
        if (principal != null) {
            if (hasMessageDigest()) {
                // Hex hashes should be compared case-insensitive
                validated = (digest(credentials).equalsIgnoreCase(principal.getPassword()));
            } else {
                validated = (digest(credentials).equals(principal.getPassword()));
            }
        }

        if (validated) {
            if (debug >= 2) {
                log(sm.getString("memoryRealm.authenticateSuccess", username));
            }
            SecurityContext ctx = new SecurityContext (principal.getName());
            SecurityCurrent current = SecurityCurrent.getCurrent();
            current.setSecurityContext(ctx);
            return (principal);
        } else {
            if (debug >= 2) {
                log(sm.getString("memoryRealm.authenticateFailure", username));
            }
            return (null);
        }

    }


    /**
     * Add a new user to the in-memory database.
     *
     * @param username User's username
     * @param password User's password (clear text)
     * @param roles Comma-delimited set of roles associated with this user
     */
    void addUser(String username, String password, String roles) {
        log("++++ addUser username= " + username + " password = " + password + " roles = " + roles);
        // Accumulate the list of roles for this user
        ArrayList list = new ArrayList();
        roles += ",";
        while (true) {
            int comma = roles.indexOf(',');
            if (comma < 0) {
                break;
            }
            String role = roles.substring(0, comma).trim();
            list.add(role);
            roles = roles.substring(comma + 1);
        }

        // Construct and cache the Principal for this user
        GenericPrincipal principal = new GenericPrincipal(this, username, password, list);
        principals.put(username, principal);
    }

    /**
     * Return a configured <code>Digester</code> to use for processing
     * the XML input file, creating a new one if necessary.
     * @return the proper digester and configured.
     */
    protected synchronized Digester getDigester() {

        if (digester == null) {
            digester = new Digester();
            digester.setDebug(this.debug);
            digester.setValidating(false);
            digester.addRuleSet(new MemoryRuleSet());
        }
        return (digester);
    }

    /**
     * Return a short name for this Realm implementation.
     * @return short name for this Realm implementation.
     */
    protected String getName() {
        return (this.name);
    }


    /**
     * Return the password associated with the given principal's user name.
     * @param username the given principal's user name.
     * @return the password associated.
     */
    protected String getPassword(String username) {
        GenericPrincipal principal =
            (GenericPrincipal) principals.get(username);
        if (principal != null) {
            return (principal.getPassword());
        } else {
            return (null);
        }
    }


    /**
     * Return the Principal associated with the given user name.
     * @param username the given principal's user name.
     * @return the Principal associated
     */
    protected Principal getPrincipal(String username) {
        return (Principal) principals.get(username);
    }


    /**
     * Prepare for active use of the public methods of this Component.
     *
     * @exception LifecycleException if this component detects a fatal error
     *  that prevents it from being started
     */
    public synchronized void start() throws LifecycleException {
        // Validate the existence of our database file
        File file = new File(pathname);
        if (!file.isAbsolute()) {
            file = new File(System.getProperty("catalina.base"), pathname);
        }
        if (!file.exists() || !file.canRead()) {
            throw new LifecycleException
                (sm.getString("memoryRealm.loadExist",
                              file.getAbsolutePath()));
        }

        // Load the contents of the database file
        if (debug >= 1) {
            log(sm.getString("memoryRealm.loadPath",
                             file.getAbsolutePath()));
        }
        Digester digester = getDigester();
        try {
            synchronized (digester) {
                digester.push(this);
                digester.parse(file);
            }
        } catch (Exception e) {
            throw new LifecycleException("memoryRealm.readXml", e);
        }

        // Perform normal superclass initialization
        super.start();
    }


    /**
     * Gracefully shut down active use of the public methods of this Component.
     *
     * @exception LifecycleException if this component detects a fatal error
     *  that needs to be reported
     */
    public synchronized void stop() throws LifecycleException {

        // Perform normal superclass finalization
        super.stop();

        // No shutdown activities required

    }


   /**
     * Log a message on the Logger associated with our Container (if any)
     *
     * @param message Message to be logged
     */
    protected void log(String message) {
        Log.getLogger(Log.JONAS_SECURITY_PREFIX).log(BasicLevel.DEBUG, message);
    }


}


/**
 * <p><strong>RuleSet</strong> for recognizing the users defined in the
 * XML file processed by <code>MemoryRealm</code>.</p>
 * @author Craig R. McClanahan
 * @author Florent Benoit : use our JOnAS/Tomcat SecurityCtxInterceptor
 */

final class MemoryRuleSet extends RuleSetBase {

    /**
     * The matching pattern prefix to use for recognizing our elements.
     */
    private String prefix = null;


    /**
     * Construct an instance of this <code>RuleSet</code> with the default
     * matching pattern prefix.
     */
    public MemoryRuleSet() {
        this("tomcat-users/");
    }


    /**
     * Construct an instance of this <code>RuleSet</code> with the specified
     * matching pattern prefix.
     *
     * @param prefix Prefix for matching pattern rules (including the
     *  trailing slash character)
     */
    public MemoryRuleSet(String prefix) {
        super();
        this.namespaceURI = null;
        this.prefix = prefix;
    }


    /**
     * <p>Add the set of Rule instances defined in this RuleSet to the
     * specified <code>Digester</code> instance, associating them with
     * our namespace URI (if any).  This method should only be called
     * by a Digester instance.</p>
     *
     * @param digester Digester instance to which the new Rule instances
     *  should be added.
     */
    public void addRuleInstances(Digester digester) {
        digester.addRule(prefix + "user", new MemoryUserRule(digester));
    }
}


/**
 * Private class used when parsing the XML database file.
 * @author Craig R. McClanahan
 * @author Florent Benoit : use our JOnAS/Tomcat SecurityCtxInterceptor
 */
final class MemoryUserRule extends Rule {

    /**
     * Construct a new instance of this <code>Rule</code>.
     *
     * @param digester The <code>Digester</code> we are associated with.
     */
    public MemoryUserRule(Digester digester) {
        super(digester);
    }


    /**
     * Process a <code>&lt;user&gt;</code> element from the XML database file.
     *
     * @param attributes The attribute list for this element
     * @throws Exception if it fails.
     */
    public void begin(Attributes attributes) throws Exception {
        String username = attributes.getValue("name");
        if (username == null) {
            username = attributes.getValue("username");
        }
        String password = attributes.getValue("password");
        String roles = attributes.getValue("roles");

        MemoryRealm realm = (MemoryRealm) digester.peek(digester.getCount() - 1);
        realm.addUser(username, password, roles);
    }


}

